import Promise from 'bluebird';

const CIFS_URL_REGEX = /^cifs:\/\/([^\/]+)(?:(\/.*?)\/(\/.*?)|(\/.*?)(\/.*?)|(\/.*?))\/?$/;
const NFS_URL_REGEX = /^nfs:\/\/([^\/]+)(?:(\/.*?)\/(\/.*?)|(\/.*?)(\/.*?)|(\/.*?))\/?$/;
const S3_URL_REGEX = /^s3:\/\/([^\/]+)(\/.*?)?\/?$/;
const SGWS_URL_REGEX = /^sgws:\/\/([^\/:]+)(?::(\d*))?(?:\/([^\/+]+))(\/.*?)?\/?$/;


function getUrlElement(url, regex, flatten, index) {
    const matches = url.match(regex);
    const flattened = flatten ? matches.filter(e => e) : matches;
    if (flattened.length > index) {
        return flattened[index] || '';
    }
    return '';
}

function convertFromStringToObject(relationship, endPointString) {
    let result;
    const type = endPointString.split(':')[0];
    switch (type) {
        case 'cifs': {
            const host = getUrlElement(endPointString, CIFS_URL_REGEX, true, 1);
            const share = getUrlElement(endPointString, CIFS_URL_REGEX, true, 2);
            const path = getUrlElement(endPointString, CIFS_URL_REGEX, true, 3);
            result = {
                host,
                share: share.startsWith('/') ? share.substr(1) : share
            };
            if (path) {
                result.path = (path) ? (path.startsWith('/') ? path.substr(1) : path) : '';
            }
            break;
        }
        case 'nfs': {
            const host = getUrlElement(endPointString, NFS_URL_REGEX, true, 1);
            const nfsExport = getUrlElement(endPointString, NFS_URL_REGEX, true, 2);
            const path = getUrlElement(endPointString, NFS_URL_REGEX, true, 3);
            result = {
                host,
                export: nfsExport
            };
            if (path) {
                result.path = (path) ? (path.startsWith('/') ? path.substr(1) : path) : '';
            }
            break;
        }
        case 's3': {
            const bucket = getUrlElement(endPointString, S3_URL_REGEX, false, 1);
            const prefix = getUrlElement(endPointString, S3_URL_REGEX, false, 2);
            result = {
                bucket,
                prefix
            };
            if (prefix) {
                result.prefix = prefix.startsWith('/') ? prefix.substr(1) : prefix;
            }
            break;
        }
        case 'sgws': {
            const host = getUrlElement(endPointString, SGWS_URL_REGEX, false, 1);
            const port = getUrlElement(endPointString, SGWS_URL_REGEX, false, 2);
            const bucket = getUrlElement(endPointString, SGWS_URL_REGEX, false, 3);
            const prefix = getUrlElement(endPointString, SGWS_URL_REGEX, false, 4);
            result = {
                host,
                bucket,
                port
            };
            if (port) {
                result.port = port;
            }
            if (prefix) {
                result.prefix = prefix.startsWith('/') ? prefix.substr(1) : prefix;
            }
            break;
        }
    }
    return {
        protocol: type,
        [type]: result
    };
}
/**
 * Make any changes you need to make to the database here
 * and return promise
 * 'this' is the mongoose object
 */
export function up() {
    const db = this.connection.db;
    return Promise.resolve()
        .then(() => {
            const relationshipDb = db.collection('relationships');
            return relationshipDb.find({}).toArray()
                .then(relationships => Promise.map(relationships, relationship => {
                    const sourceObj = convertFromStringToObject(relationship, relationship.source);
                    const targetObj = convertFromStringToObject(relationship, relationship.target);
                    return relationshipDb.update({ _id: relationship._id }, { $set: { source: sourceObj, target: targetObj } });
                }));
        });
}

/**
 * Make any changes that UNDO the up function side effects here (if possible)
 * and return promise
 * 'this' is the mongoose object
 */
export function down() {
    return Promise.resolve();
}
