import swaggerTools from 'swagger-tools';
import express from 'express';
import compression from 'compression';
import path from 'path';
import http from 'http';
import https from 'https';
import helmet from 'helmet';
import SwaggerParser from 'swagger-parser';
import cors from 'cors';
import log4js from 'log4js';
import './globals';
import {UI_PATH, SERVER_HTTP_PORT, API_ROOT, SWAGGER_UI_PATH, SWAGGER_SPEC_PATH, SERVER_MAX_SOCKETS_PER_ORIGIN, SWAGGER_UI} from './utils/consts';
import {getRuntimeEnvironment} from '../../environment';
import bootstrap from './utils/bootstrap';
import healthRoutes from './routes/health';
import userRoutes from './routes/users';
import messagesRoutes from './routes/messages';
import storagesRoutes from './routes/storages';
import dataBrokersRoutes from './routes/data-brokers';
import relationshipsRoutes from './routes/relationships';
import groupsRoutes from './routes/groups';
import mpSaasRoutes from './routes/saas-registration';
import timelinesRoutes from './routes/timelines';
import configRoutes from './routes/config';
import authenticateMiddleware from './middlewares/authenticate';
import reactRouterFallbackMiddleware from './middlewares/react-router-fallback';
import continuationLocalStorageMiddleware from './middlewares/continuation-local-storage';
import errorHandlerMiddleware from './middlewares/error-handler';
import timelineMiddleware from './middlewares/timeline';
import { validateSwaggerSpec } from './utils/swagger';
import {getLogger} from './utils/logger';
import './utils/backup';


// patch max sockets per origin otherwise node starts to cough on requests
// see: http://docs.aws.amazon.com/sdk-for-javascript/v2/developer-guide/node-configuring-maxsockets.html
// see: https://gist.github.com/nscott/8735628
http.globalAgent.maxSockets = SERVER_MAX_SOCKETS_PER_ORIGIN;
https.globalAgent.maxSockets = SERVER_MAX_SOCKETS_PER_ORIGIN;

const logger = getLogger();

process.on('uncaughtException', err => logger.error('Uncaught exception', err));

process.on('unhandledRejection', (err, promise) => logger.error('Unhandled promise rejection', promise, err));

//TODO upgrade swagger
async function initSwagger(app) {
    try {
        const swaggerSpec = await new SwaggerParser().dereference(path.resolve(__dirname, SWAGGER_SPEC_PATH));
        // validate swagger spec
        const api = await validateSwaggerSpec(swaggerSpec);
        return new Promise(resolve => {
            swaggerTools.initializeMiddleware(api, middlewares => {
                // serve the Swagger documents and Swagger UI
                app.use(middlewares.swaggerUi({
                    swaggerUi: SWAGGER_UI,
                    swaggerUiDir: path.resolve(__dirname, SWAGGER_UI_PATH)
                }));
                // interpret Swagger resources and attach metadata to request
                app.use(middlewares.swaggerMetadata());
                //add timeline to all the relevant requests (not get and not messages)
                app.use(API_ROOT, timelineMiddleware);
                // validate Swagger requests
                app.use(middlewares.swaggerValidator({validateResponse: true}));
                resolve(app);
            });
        });
    } catch (err) {
        logger.error('Failed initializing swagger:', err);
        throw err;
    }
}

function initMiddlewares(app) {
    app.use(API_ROOT, express.json());
    app.use(API_ROOT, express.urlencoded({ extended: true }));
    // configure security helmet
    app.use(helmet());
    // configure continuation local storage
    app.use(API_ROOT, continuationLocalStorageMiddleware);

    // health check hidden (from swagger) route (and not logged in access log)
    app.use(API_ROOT, healthRoutes);
    // configure access logger
    app.use(API_ROOT, log4js.connectLogger(getLogger('access-log'), {
        level: log4js.levels.INFO,
        format: ':remote-addr :protocol :http-version :method :url :status :content-length :response-time'
    }));
    // always add response compression
    app.use(compression());
    // enable all CORS requests
    app.use(cors());

    // configure static assets middleware
    const staticAssetsMiddleware = express.static(UI_PATH);
    app.use(staticAssetsMiddleware);
    // always return index.html in case no resource found and it's not an API call
    app.use(reactRouterFallbackMiddleware(staticAssetsMiddleware));
    // authenticate request
    app.use(API_ROOT, authenticateMiddleware);
}

function initRoutes(app) {
    app.use(API_ROOT, userRoutes);
    app.use(API_ROOT, messagesRoutes);
    app.use(API_ROOT, storagesRoutes);
    app.use(API_ROOT, mpSaasRoutes);
    app.use(API_ROOT, dataBrokersRoutes);
    app.use(API_ROOT, relationshipsRoutes);
    app.use(API_ROOT, groupsRoutes);
    app.use(API_ROOT, timelinesRoutes);
    app.use(API_ROOT, configRoutes);
}

async function startServer() {
    try {
        const app = express();
        await bootstrap();
        initMiddlewares(app);
        await initSwagger(app);
        initRoutes(app);
        //init app exception handler
        app.use(errorHandlerMiddleware);
        app.listen(SERVER_HTTP_PORT);
        logger.info('======================================================================');
        logger.info(`Server version '${__VERSION__}' running in '${getRuntimeEnvironment()}' mode started listening.`);
    } catch (err) {
        logger.error('Failed staring server with error:', err);
        process.exit(0);
    }
}

process.once('SIGINT', () => {
    logger.info('CTRL-C caught');
    process.exit(0);
});

startServer();
