import HubSpot from 'hubspot';
import moment from 'moment';
import { HUBSPOT_API_KEY } from '../utils/consts';
import {getLogger} from '../utils/logger';
const logger = getLogger();

const FIELDS = {
    SINCE: 'cloudsync_since',
    TRIAL_EXPIRY: 'cloudsync_trial_expiry',
    DATA_BROKER_SINCE: 'cloudsync_databroker_since',
    RELATIONSHIP_SINCE: 'cloudsync_relationship_since',
    SUBSCRIPTION_SINCE: 'cloudsync_subscription_since',
    LAST_PAYMENT : 'cloudsync_last_payment_since'
};

function hasHubSpotKey() {
    return HUBSPOT_API_KEY !== null;
}

export async function createOrUpdateHubspotContact(user, hubSpotClient) {
    if (!hasHubSpotKey()) {
        logger.debug('HubSpot is not enabled');
        return;
    }
    try {
        const client = (hubSpotClient) ? hubSpotClient : new HubSpot({apiKey: HUBSPOT_API_KEY});
        const nameParts = user.name.split(' ').map(word => word.trim()).filter(word => word !== '');
        let firstname, lastname;
        switch (nameParts.length) {
            case 0:
                firstname = '';
                lastname = '';
                break;
            case 1:
                firstname = nameParts[0];
                lastname = '';
                break;
            case 2:
                firstname = nameParts[0];
                lastname = nameParts[1];
                break;
            default:
                firstname = nameParts[0];
                lastname = nameParts.slice(1).join(' ');
        }
        const properties = {
            properties: [
                {
                    property: 'email',
                    value: user.email
                },
                {
                    property: 'firstname',
                    value: firstname
                },
                {
                    property: 'lastname',
                    value: lastname
                },
                {
                    property: 'company',
                    value: user.company
                },
                {
                    property: FIELDS.SINCE,
                    value: moment(user.createdAt).utc().startOf('day').valueOf()
                },
                {
                    property: FIELDS.TRIAL_EXPIRY,
                    value: moment(user.trialExpiry).utc().startOf('day').valueOf()
                }
            ]
        };

        const data = await client.contacts.createOrUpdate(user.email, properties);
        const action = data.isNew === true ? 'Created' : 'Updated';
        logger.info(`${action} HubSpot Contact ${data.vid} for user ${user}`);
        return data;
    } catch (err) {
        logger.warn(`Failed creating or updating HubSpot Contact for user ${user}:`, err);
    }
}

async function updateContactProperty(client, contact, property, user) {
    await client.contacts.update(contact.vid, {
        properties: [
            {
                property,
                value: moment().utc().startOf('day').valueOf()
            }
        ]
    });
    logger.info(`Added HubSpot Contact property ${property} mark for user ${user}`);
}

async function setHubSpotContactPropertyIfNeeded(user, property) {
    if (!hasHubSpotKey()) {
        logger.debug('HubSpot is not enabled');
        return;
    }

    try {
        const client = new HubSpot({apiKey: HUBSPOT_API_KEY});
        const userData = await client.contacts.getByEmail(user.email);
        let contact;
        if (userData && userData.properties) {
            logger.info(`Found existing HubSpot Contact by email ${user.email}:`, userData.vid);
            contact = userData;
        } else {
            logger.info(`Existing HubSpot Contact by email ${user.email} was not found: `, userData);
            logger.info(`Creating new HubSpot Contact for user with email ${user.email}`);
            await createOrUpdateHubspotContact(user, client);
            contact = await client.contacts.getByEmail(user.email);
        }

        if (!contact.properties[property] || !contact.properties[property].value) {
            return updateContactProperty(client, contact, property, user);
        } else if (property === FIELDS.LAST_PAYMENT && contact.properties[property] === moment().utc().startOf('day').valueOf()) {
            return updateContactProperty(client, contact, property, user);
        }
    } catch (err) {
        logger.warn(`Failed setting HubSpot Contact property ${property} for user ${user}:`, err);
    }
}

export async function setHubSpotContactFirstDataBrokerSinceIfNeeded(user) {
    return setHubSpotContactPropertyIfNeeded(user, FIELDS.DATA_BROKER_SINCE);
}

export async function setHubSpotContactFirstRelationshipSinceIfNeeded(user) {
    return setHubSpotContactPropertyIfNeeded(user, FIELDS.RELATIONSHIP_SINCE);
}

export async function setHubSpotContactFirstSubscriptionSinceIfNeeded(user) {
    return setHubSpotContactPropertyIfNeeded(user, FIELDS.SUBSCRIPTION_SINCE);
}

export async function setHubSpotLastPaymentIfNeeded(user) {
    return setHubSpotContactPropertyIfNeeded(user, FIELDS.LAST_PAYMENT);
}
