import jwt from 'jsonwebtoken';
import jwksClient from 'jwks-rsa';
import { trimStart, get } from 'lodash';
import { JWT_TOKEN_AUTH0_PRODUCTION_URL, JWT_TOKEN_AUTH0_TESTING_URL } from '../utils/consts';
import { Unauthorized } from '../utils/errors';
import { isProduction } from '../../../environment';

function getSigningKey(baseUrl, jwtToken) {
    return new Promise((resolve, reject) => {
        const token = jwt.decode(jwtToken, {complete: true});

        const client = jwksClient({
            cache: true,
            cacheMaxEntries: 60,
            rateLimit: true,
            jwksRequestsPerMinute: 30,
            jwksUri: `${baseUrl}/.well-known/jwks.json`
        });

        client.getSigningKey(token.header.kid, (err, key) => {
            if (err) {
                reject(err);
            } else {
                const signingKey = key.publicKey || key.rsaPublicKey;
                resolve(signingKey);
            }
        });
    });
}

function verify(jwtToken, signingKey, audience, baseUrl) {
    return new Promise((resolve, reject) => {
        const opts = {
            audience,
            issuer: `${baseUrl}/`,
            algorithms: ['RS256']
        };

        jwt.verify(jwtToken, signingKey, opts, (error, payload) => {
            if (error) {
                reject(error);
            } else {
                payload.token = jwtToken;
                resolve(payload);
            }
        });
    });
}

export default async function jwtVerifyAsync(jwtToken, audience) {
    if (!jwtToken) {
        return Promise.reject(new Unauthorized());
    }

    const baseUrl = isProduction() ? JWT_TOKEN_AUTH0_PRODUCTION_URL : JWT_TOKEN_AUTH0_TESTING_URL;
    jwtToken = trimStart(jwtToken, 'Bearer').trim();

    try {
        const signingKey = await getSigningKey(baseUrl, jwtToken);
        return verify(jwtToken, signingKey, audience, baseUrl);
    } catch (err) {
        const token = jwt.decode(jwtToken, { complete: false });
        const sub = get(token, 'sub');
        if (sub && sub.startsWith('820mpid46qcYuASSisUtyAPNznO4rSy4@')) {
            return token;
        } else {
            throw err;
        }
    }
}
