import Promise from 'bluebird';
import _ from 'lodash';
import AWS from 'aws-sdk';
import agenda from '../utils/agenda';
import SaaSBilling from '../models/saas-billing';
import { AGENDA_TASK_NAME_SAAS_METERING, SAAS_REGISTRATION_STATUS_RESOLVE_FAILED,
    SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS, PAYGO_SUBSCRIPTION, MP_PRODUCT_CODE_METERING, SERVICE_REGION, SELLER_ACCESS_KEY, SELLER_SECRET_KEY  } from '../utils/consts';
import Relationship from '../models/relationship';
import User from '../models/user';
import Cost from '../models/cost';
import { validateProductCode } from '../utils/mp-subscription-status-listener';
import {setHubSpotLastPaymentIfNeeded} from '../lib/hubspot';
import {getLogger} from '../utils/logger';
const logger = getLogger();

function getMarketplaceMetering() {
    return new AWS.MarketplaceMetering({region: SERVICE_REGION, accessKeyId: SELLER_ACCESS_KEY, secretAccessKey: SELLER_SECRET_KEY});
}

export async function batchMeterUsage(records) {
    const marketplaceMetering = getMarketplaceMetering();
    return Promise.all(_.chunk(records, 25).map(chunk => {
        const batch = {
            'ProductCode': MP_PRODUCT_CODE_METERING,
            'UsageRecords': chunk
        };
        logger.info('Metering request with batch:', batch);

        return marketplaceMetering.batchMeterUsage(batch).promise();
    }));
}
export function getCustomerTiers(allRelationships, byolReservedRelationships) {
    let remainingCount;
    let tier1 = 0;
    let tier2 = 0;
    let tier3 = 0;
    if (byolReservedRelationships !== 0) {
        remainingCount = allRelationships - byolReservedRelationships;
        if (remainingCount > 0) {
            tier3 = remainingCount;
        }
    } else {
        remainingCount = allRelationships;
        if (remainingCount < 0) {
            remainingCount = 0;
        }

        if (remainingCount > 5) {
            tier1 = 5;
            remainingCount = remainingCount - tier1;
        } else {
            tier1 = remainingCount;
            remainingCount = 0;
        }

        if (remainingCount > 15) {
            tier2 = 15;
            remainingCount = remainingCount - tier2;
        } else {
            tier2 = remainingCount;
            remainingCount = 0;
        }

        tier3 = remainingCount;
    }
    return {tier1, tier2, tier3};
}

function createMeteringEntry(customerId, tier, quantity, timeStamp)  {
    return {
        CustomerIdentifier: customerId,
        Dimension: tier,
        Quantity: quantity,
        Timestamp: timeStamp
    };
}

function createRecordObjectForCost(customerId, tier, quantity, timeStamp, uId, byolPerUser, byolEntiteled, isSuccess) {
    return ({
        entry: createMeteringEntry(customerId, tier, quantity, timeStamp),
        uid: uId,
        byol: {reservedRelationships: byolEntiteled, usedRelationships: byolPerUser},
        isSuccess
    });
}

export async function resolveCustomerIdentifierFromRegistrationTokenAsync(token) {
    logger.info(`Resolving customer identifier from token ${token}`);
    try {
        const resolvedToken = await getMarketplaceMetering().resolveCustomer({RegistrationToken: token}).promise();
        logger.info(`Resolved customer id for token ${token}: ${resolvedToken.CustomerIdentifier} [ProductCode: ${resolvedToken.ProductCode}`);
        const type = validateProductCode(resolvedToken.ProductCode);
        logger.info(`Validated ProductCode ${resolvedToken.ProductCode} from token ${token}:${resolvedToken.CustomerIdentifier} to be from type ${type}`);
        return {
            customerIdentifier: resolvedToken.CustomerIdentifier,
            status: SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS,
            type
        };
    } catch (err) {
        logger.error(`Failed to resolve customer token ${token}:`, err);
        return {
            customerIdentifier: undefined,
            status: SAAS_REGISTRATION_STATUS_RESOLVE_FAILED,
            type: PAYGO_SUBSCRIPTION
        };

    }
}

// given records to bill => batches them and send them to saas metering service.
// saves the honored records to the DB and returns a collection of the unprocessed/unhonored records
function billRecords(records) {
    return batchMeterUsage(records)
        .then(responses => {
            if (responses.length === 0) {
                return {processedRecords : [], unprocessedRecords: []};
            }
            logger.debug('Metering responses for batches:', JSON.stringify(responses));
            const processedRecords = _.flatten(responses.map(response => response.Results));
            //logging unhonored. unlike unprocessed these are processed but not successfuly [DuplicateRecord, CustomerNotSubscribed]
            logger.warn('Records not honored:', processedRecords.filter(record => record.Status !== 'Success'));
            logger.info('processed billing entries ', processedRecords);
            return SaaSBilling.insertMany(processedRecords).then(() => {// need to delete in the end
                const unprocessedRecords = _.flatten(responses.map(response => response.UnprocessedRecords));
                logger.info('Unprocessed records', unprocessedRecords);
                return {processedRecords : processedRecords, unprocessedRecords: unprocessedRecords};
            });
        });
}

export function addRecordsOfUsersWithoutRelationships(records, chargingTime, usersIds) {
    //add to the records array all the users with no relationships
    usersIds.map(userId => {
        const recordWithUserId = records.find(record => {
            return userId.equals(record.uid);
        });
        if (!recordWithUserId) {
            records.push(createRecordObjectForCost('', 'Tier1', 0, chargingTime, userId, 0, 0, true));
        }
    });
    return records;
}

export function filterAndInsert(billingObjects) {
    return Cost.lastCostRecords()
        .then(lastRecords => {
            const lastRecordsByUid = _.keyBy(lastRecords, '_id');
            _.keys(lastRecordsByUid).map(userId => {
                const  { tier1, tier2, tier3, byol } = lastRecordsByUid[userId];
                const newRecordForUser = billingObjects[userId];
                if (newRecordForUser && (newRecordForUser.tier1 === tier1 || !newRecordForUser.tier1 && tier1 === 0) &&
                    (newRecordForUser.tier2 && newRecordForUser.tier2 === tier2 || !newRecordForUser.tier2 && tier2 === 0) &&
                    (newRecordForUser.tier3 && newRecordForUser.tier3 === tier3 || !newRecordForUser.tier3 && tier3 === 0) &&
                    (newRecordForUser.byol && newRecordForUser.byol.usedRelationships === byol.usedRelationships || !newRecordForUser.byol && byol.usedRelationships === 0)) {
                    logger.debug('******DELETE record from Cost table', newRecordForUser);//forDEBUG
                    delete billingObjects[userId];
                }
            });
            logger.info('Insert all the success records to the Cost collection in MongoDB');
            if (!_.isEmpty(billingObjects)) {
                return Cost.insertMany(_.map(billingObjects));
            }
        });
}

export function createObjectsForMongoSaaSUsers(processedRecords, chargingTime, billableUsers, byolObjects) {
    logger.info('Create MongoDB objects for SaaS users');
    const objectsForMongo = {};
    const users = {};
    const successRecords = processedRecords.filter(record => record.Status === 'Success');

    successRecords.forEach(record => {
        const customerIdentifier = record.UsageRecord.CustomerIdentifier;
        const usersWithCustomerIdentifier = billableUsers.filter(user => user.subscription.customerIdentifier === customerIdentifier);
        const dimension = record.UsageRecord.Dimension.toLowerCase();
        const quantity = record.UsageRecord.Quantity;
        usersWithCustomerIdentifier.forEach(user => {
            const byol = byolObjects[user.subscription.customerIdentifier];
            if (objectsForMongo[user._id]) {
                objectsForMongo[user._id][dimension] = quantity;
            } else {
                objectsForMongo[user._id] = {timestamp: chargingTime, user: user._id, byol};
                objectsForMongo[user._id][dimension] = quantity;
            }

            users[user._id] = user;
        });
    });

    Promise.map(Object.keys(users), userId => setHubSpotLastPaymentIfNeeded(users[userId]), {concurrency: 5});

    return objectsForMongo;
}

export function billSaaSUsers(chargingTime) {
    logger.info('Find all the relationships for subscribed user');
    return User.BillableUsers()
        .then(billableUsers => {
            const billableUserIds = billableUsers.map(user => user._id);
            return Relationship.aggregate([
                {$match: {user: {$in: billableUserIds}, isQstack: false}},
                {
                    $lookup: {
                        from: 'users',
                        localField: 'user',
                        foreignField: '_id',
                        as: 'user'
                    }
                },
                {$group: {_id: '$user.subscription.customerIdentifier', count: {$sum: 1}}}]).exec()
                .then(groups => {
                    logger.info('For each customerId find all its billing');
                    return User.customerIdentifierByolEntitlement()
                        .then(byolEntitlements => {
                            const byolByCustomerIdentifier = _.keyBy(byolEntitlements, '_id');
                            const byolObjects = {};
                            const records = groups.map(group => {
                                const [ customerId ] = group._id;
                                const customerEntitlement = byolByCustomerIdentifier[customerId] ? byolByCustomerIdentifier[customerId].count : 0;
                                const byolPerCustomerIdentifier = byolByCustomerIdentifier[customerId] ? (group.count - customerEntitlement <= 0 ? group.count : customerEntitlement) : 0;
                                if (customerEntitlement !== 0) {
                                    byolObjects[customerId] = {reservedRelationships: customerEntitlement, usedRelationships: byolPerCustomerIdentifier};
                                }
                                const {tier1, tier2, tier3} = getCustomerTiers(group.count, customerEntitlement);
                                const tiers = [];

                                if (tier1 > 0) {
                                    tiers.push(createMeteringEntry(customerId, 'Tier1', tier1, chargingTime));
                                }
                                if (tier2 > 0) {
                                    tiers.push(createMeteringEntry(customerId, 'Tier2', tier2, chargingTime));
                                }
                                if (tier3 > 0) {
                                    tiers.push(createMeteringEntry(customerId, 'Tier3', tier3, chargingTime));
                                }

                                return tiers;
                            });
                            const flattenRecords = _.flatten(records);
                            return billRecords(flattenRecords)
                                .then(billRecordsResults => {
                                    const { processedRecords, unprocessedRecords} = billRecordsResults;
                                    return billRecords(unprocessedRecords)
                                        .then(billObjectAfterRetry => {
                                            const processedRecordsAfterRetry = _.concat(processedRecords, billObjectAfterRetry.processedRecords);
                                            const unprocessedRecordsAfterRetry = billRecordsResults.unprocessedRecords;
                                            if (unprocessedRecordsAfterRetry.length === 0) {
                                                return processedRecordsAfterRetry;
                                            }
                                            const unprocessedEntriesForDB = unprocessedRecordsAfterRetry.map(meteringRecord => {
                                                return {
                                                    Status: 'Failed',
                                                    MeteringRecordId: '',
                                                    UsageRecord: meteringRecord
                                                };
                                            });
                                            return SaaSBilling.insertMany(unprocessedEntriesForDB)
                                                .then(() => processedRecordsAfterRetry);
                                        });
                                })
                                .then(processedRecords => {
                                    logger.info('Create objects for mongo');
                                    return createObjectsForMongoSaaSUsers(processedRecords, chargingTime, billableUsers, byolObjects);
                                });

                        });

                });
        });
}

export function addUsersWithoutRelationshipsToMongoObject(mongoObjects, chargingTime) {
    return User.find({}).exec()
        .then(users => {
            users.map(user => {
                if (!mongoObjects[user._id]) {
                    const byol = {usedRelationships: 0, reservedRelationships : 0};
                    mongoObjects[user._id] = {user: user._id, tier1: 0, timestamp: chargingTime, byol};
                }
            });
            return mongoObjects;
        });
}
//will update BYOL for NetApp users with trial expired as well
export function byolDetailsForUnsubscribedUsers(objectsForMongo, chargingTime) {
    logger.info('find BYOL details for unsubscribed users');
    return User.useraWithTrialExpired()
        .then(users => {
            const usersId = users.map(user => user._id);
            return Relationship.aggregate([
                {$match: {user: {$in: usersId}}},
                {
                    $lookup: {
                        from: 'users',
                        localField: 'user',
                        foreignField: '_id',
                        as: 'user'
                    }
                },
                {
                    $group: {
                        _id: {
                            uId: '$user._id'
                        }, count: {$sum: 1}
                    }
                }]).exec()
                .then(groups => {
                    return User.userByolEntitlementForExpiredTrial()
                        .then(byolEntitlements => {
                            const byolById = _.keyBy(byolEntitlements, '_id');
                            groups.map(group => {
                                const [ uId ] = group._id.uId;
                                if (!objectsForMongo[uId]) {
                                    const byolEntiteled = byolById[uId] ? byolById[uId].count : 0;
                                    const byolPerUser = byolById[uId] ? (group.count - byolEntiteled <= 0 ? group.count : byolEntiteled) : 0;
                                    const byol = {reservedRelationships: byolEntiteled, usedRelationships: byolPerUser};
                                    objectsForMongo[uId] = {user: uId, timestamp: chargingTime, tier1: 0, byol: byol};
                                }
                            });
                            return objectsForMongo;
                        });
                });
        });
}


agenda.define(AGENDA_TASK_NAME_SAAS_METERING, {lockLifetime: 10000}, (job, done) => {
    const chargingTime = new Date().toISOString();
    logger.info('Scheduled MarketPlace Metering Billing');
    return billSaaSUsers(chargingTime)
        .then(mongoObjects => byolDetailsForUnsubscribedUsers(mongoObjects, chargingTime))
        .then(mongoObjects => addUsersWithoutRelationshipsToMongoObject(mongoObjects, chargingTime))
        .then(mongoObjects => filterAndInsert(mongoObjects))
        .catch(err => logger.error('Failed to report metering:', err))
        .finally(done);
});

