import {  BadRequest } from '../utils/errors';
import {DFIO_METADATA_BUCKET_NAME} from '../utils/consts';

function getS3(region) {
    return new AWS.S3({region});
}

export async function putObject(region, bucketName, key, file, makePublic = false) {
    const s3 = getS3(region);
    const params = {
        Bucket: bucketName,
        Key: key,
        Body: file
    };

    if (makePublic) {
        params.ACL = 'public-read';
    }
    return s3.putObject(params).promise();
}

export async function listObjectsAsync(region, bucketName, prefix) {
    const s3 = getS3(region);

    const data = await s3.listObjectsV2({
        Bucket: bucketName,
        Prefix: prefix
    }).promise();
    return data.Contents.map(content => {
        return {
            key: content.Key,
            lastModified: new Date(content.LastModified).getTime(),
            size: content.Size
        };
    });
}

export async function getObjectTaggingAsync(region, bucket, key) {
    const s3 = getS3(region);

    const data = await s3.getObjectTagging({
        Bucket: bucket,
        Key: key
    }).promise();
    return data.TagSet.reduce((accumulator, currentValue) => {
        accumulator[currentValue.Key] = currentValue.Value;
        return accumulator;
    }, {});
}

export function getS3ObjectUrl(region, bucket, key, shouldReplace, isBundle) {
    const prefix = region === 'us-east-1' ? 's3' : `s3-${region}`;
    const s3Url =  `https://${prefix}.amazonaws.com/${bucket}/${encodeURIComponent(key)}`;
    if (shouldReplace) {
        let regex;
        let aliasName;
        if (isBundle) {
            regex =  new RegExp(`https:\/\/.*.\/${DFIO_METADATA_BUCKET_NAME}\/repo\/`, 'g');
            aliasName = 'repo';
        } else {
            regex =  new RegExp(`https:\/\/.*.\/${DFIO_METADATA_BUCKET_NAME}\/.*\/`, 'g');
            aliasName = 'cf';
        }
        return decodeURIComponent(s3Url).replace(regex,  `https://${aliasName}.cloudsync.netapp.com/`);
    } else {
        return s3Url;
    }
}

export function headBucket(bucket) {
    const s3 = getS3('us-east-1');
    return new Promise((resolve, reject) => {
        s3.headBucket({Bucket: bucket})
            .on('httpHeaders', (statusCode, headers) => {
                if (headers['x-amz-bucket-region']) {
                    resolve(headers['x-amz-bucket-region']);
                } else {
                    reject(new BadRequest(`Got unexpected status code ${statusCode} - for headBucket operation on bucket ${bucket}`));
                }
            })
            .on('error', error => reject(error))
            .on('httpError', error => reject(error))
            .send();
    });
}

