import {mapValues} from 'lodash';
import {SERVICE_REGION} from '../utils/consts';
import {getLogger} from '../utils/logger';

const logger = getLogger();

let sqs;

function getSQS() {
    if (!sqs) {
        sqs = new AWS.SQS({region: SERVICE_REGION});
    }
    return sqs;
}

export async function createQueue(name) {
    const sqsClient = getSQS();

    const data = await sqsClient.createQueue({
        QueueName: name,
        Attributes: {
            ReceiveMessageWaitTimeSeconds: '20',
            VisibilityTimeout: '60'
        }
    }).promise();

    return data.QueueUrl;
}

export async function sendMessage(queueUrl, message, messageAttributes = {}) {
    const sqsClient = getSQS();

    const data = await sqsClient.sendMessage({
        MessageBody: JSON.stringify(message),
        QueueUrl: queueUrl,
        MessageAttributes: mapValues(messageAttributes, value => ({StringValue: value, DataType: 'String'}))
    }).promise();

    return data.MessageId;
}

export async function receiveMessage(queueUrl) {
    const sqsClient = getSQS();

    const data = await sqsClient.receiveMessage({
        QueueUrl: queueUrl,
        MessageAttributeNames: ['All'],
        MaxNumberOfMessages: 10,
        WaitTimeSeconds: 20
    }).promise();

    if (data && data.Messages) {
        return data.Messages.map(message => ({
            receiptHandle: message.ReceiptHandle,
            body: message.Body,
            attributes: mapValues(message.MessageAttributes, value => value.StringValue)
        }));
    } else {
        return [];
    }
}

export async function deleteMessage(queueUrl, receiptHandle) {
    logger.debug('Deleting message, receipt handle:', receiptHandle);
    const sqsClient = getSQS();

    return sqsClient.deleteMessage({
        QueueUrl: queueUrl,
        ReceiptHandle: receiptHandle
    }).promise();
}

export async function deleteQueue(queueUrl) {
    const sqsClient = getSQS();

    return sqsClient.deleteQueue({QueueUrl: queueUrl}).promise();
}

export async function getQueueArn(queueUrl) {
    const sqsClient = getSQS();

    const data = await sqsClient.getQueueAttributes({
        QueueUrl: queueUrl,
        AttributeNames: ['QueueArn']
    }).promise();

    return data.Attributes.QueueArn;
}

async function setQueueAttribute(queueUrl, attributesObject) {
    const sqsClient = getSQS();

    return sqsClient.setQueueAttributes({
        QueueUrl: queueUrl,
        Attributes: attributesObject
    }).promise();
}

export async function setQueuePolicy(queueUrl, policy) {
    return setQueueAttribute(queueUrl, {
        Policy: JSON.stringify(policy)
    });
}

export async function getQueueAttributes(queueUrl, attributeNames) {
    const sqsClient = getSQS();

    try {
        const response = await sqsClient.getQueueAttributes({
            QueueUrl: queueUrl,
            AttributeNames: attributeNames
        }).promise();
        return response.Attributes;
    } catch (err) {
        if (err.code === 'AWS.SimpleQueueService.NonExistentQueue') {
            logger.warn(`Queue ${queueUrl} does not exist deleting queue`);
        } else {
            logger.error(`getQueueAttributes, got error for queue ${queueUrl}`, err);
            throw err;
        }
    }
}

export async function allowArnSendMessage(queueUrl, queueArn, arn) {
    const policy = {
        'Version': '2012-10-17',
        'Id': 'arn-send-message',
        'Statement': [
            {
                'Sid': `Sid${Date.now()}`,
                'Effect': 'Allow',
                'Principal': '*',
                'Action': 'SQS:SendMessage',
                'Resource': queueArn,
                'Condition': {
                    'ArnEquals': {
                        'aws:SourceArn': arn
                    }
                }
            }
        ]
    };

    return setQueuePolicy(queueUrl, policy);
}

