import axios from 'axios';
import User from '../models/user';
import {
    PRODUCTION_QSTACK_CLIENT_IDS, STAGING_QSTACK_CLIENT_IDS, JWT_TOKEN_AUTH0_PRODUCTION_URL,
    JWT_TOKEN_AUTH0_TESTING_URL, JWT_TOKEN_AUTH0_PAYLOAD_FIELD_USER_ID, TRIAL_EXPIRATION_MS,
    API_PATH_CONFIG, API_PATH_SAAS_REGISTRATION, JWT_TOKEN_API_AUTH0_AUDIENCE,
    JWT_TOKEN_CLOUD_MANAGER_AUTH0_AUDIENCE, API_PATH_HEALTH_CHECK
} from '../utils/consts';
import { Unauthorized, Forbidden } from '../utils/errors';
import jwtVerifyAsync from '../lib/jwt';
import { isEmptyString } from '../utils/utils';
import {createOrUpdateHubspotContact } from '../lib/hubspot';
import {isProduction } from '../../../environment';
import {getLogger} from '../utils/logger';
const logger = getLogger();


const SKIP_AUTHENTICATION_REQUESTS_PATH = [
    API_PATH_CONFIG,
    API_PATH_SAAS_REGISTRATION,
    API_PATH_HEALTH_CHECK
];


function jwtGetUserInfo(jwtToken) {
    const baseUrl = isProduction() ? JWT_TOKEN_AUTH0_PRODUCTION_URL : JWT_TOKEN_AUTH0_TESTING_URL;
    return axios.get(`${baseUrl}/userinfo`, {
        headers: {
            authorization: jwtToken,
        }
    });
}

function findClientType(info) {
    const { sub } = info;
    const [ clientId ] = sub.split('@');
    const approvedClientsIds = isProduction() ? PRODUCTION_QSTACK_CLIENT_IDS : STAGING_QSTACK_CLIENT_IDS;
    if (approvedClientsIds.includes(clientId)) {
        return 'qstack';
    } else if (info.scope && info.scope.includes('occm-access')) {
        return 'occm';
    } else {
        throw new Forbidden('Unauthorized client');
    }
}

async function loadClientToken(payload, req, authorization, onBehalf, client) {
    const portalIdFromToken = payload[JWT_TOKEN_AUTH0_PAYLOAD_FIELD_USER_ID];
    if (isEmptyString(portalIdFromToken)) {
        logger.error('Missing user ID in client JWT token payload');
        throw new Unauthorized();
    }

    const userByPortalId = await User.findOne({portalId: portalIdFromToken}).exec();
    if (!userByPortalId) {
        logger.info(`User with portalId: ${portalIdFromToken} not found`);
        // TODO-SK - ugly path for OCCM Auth0 integration
        // do not create a new user in case of OCCM and call to users/licenses
        // see routes/users.js '/users/licenses' route for more
        if ((req.path.includes('/users/info') && req.query.noCreate === 'true') || (req.path.includes('users/licenses') && onBehalf && client === 'occm')) {
            req.user = null;
        } else {
            const userInfo = await jwtGetUserInfo(onBehalf ? onBehalf : authorization);
            const portalId = userInfo.data.sub;
            const userMetadata = userInfo.data['http://cloud.netapp.com/user_metadata'];
            const name = userMetadata.name.trim();
            const email = userMetadata.email.trim().toLowerCase();
            const company = userMetadata.company.trim();
            const trialExpiry = new Date(Date.now() + TRIAL_EXPIRATION_MS);
            let newUser = new User({
                name,
                email,
                company,
                trialExpiry,
                portalId,
                lastLogin: new Date(),
                isOccm: (client === 'occm'),
                isQstack : (client === 'qstack')

            });
            newUser = await newUser.save();
            createOrUpdateHubspotContact(newUser);
            logger.info(`User ${newUser} created and loaded`);
            req.user = newUser; // set the user on the request so that access log is printed with the user information
            req.isQstack = (client === 'qstack');
            req.isCvo = (client === 'occm');
        }
    } else  {
        logger.debug(`User ${userByPortalId} content loaded`);
        req.user = userByPortalId;
        req.isQstack = (client === 'qstack');
        req.isCvo = (client === 'occm');
    }
}

async function loadRequestData(req) {
    if (!SKIP_AUTHENTICATION_REQUESTS_PATH.includes(req.path) && req.get('x-netapp-internal') !== '86ba39f2-ed8d-4044-bd3c-7beb630cb8d5') {

        // 2 cases:
        // #1 - Regular client - got single authorization header 'authorization', part of the payload is:
        //      "audience": "https://api.cloud.netapp.com"
        //      "sub": "auth0|user_id"
        // #2 - OCCM - got two authorization headers: 'authorization' and 'x-netapp-on-behalf'
        //      authorization:
        //      "audience": "https://api.cloud.netapp.com"
        //      "sub": "xxxxxxxxxxx|client"
        //
        //      on behalf:
        //      "audience": "https://ontap.cloud.netapp.com"
        //      "sub": "auth0:user_id"

        const authorization = req.get('Authorization') || req.get('authorization');
        logger.debug('Authorization header:', authorization);

        const onBehalf = req.get('x-netapp-on-behalf');
        logger.debug('On behalf header:', onBehalf);
        try {
            let payload = await jwtVerifyAsync(authorization, JWT_TOKEN_API_AUTH0_AUDIENCE);
            const client = onBehalf ? findClientType(payload) : undefined;
            if (client === 'occm') {
                payload = await jwtVerifyAsync(onBehalf, JWT_TOKEN_CLOUD_MANAGER_AUTH0_AUDIENCE);
            } else if (client === 'qstack') {
                payload = await jwtVerifyAsync(onBehalf, JWT_TOKEN_API_AUTH0_AUDIENCE);
            }
            logger.debug('Successfully verified JWT token. Payload:', payload);
            return loadClientToken(payload, req, authorization, onBehalf, client);
        } catch (error) {
            logger.debug(`Failed authenticating user with JWT token - authorization ${authorization}, x-netapp-on-behalf ${onBehalf}:`, error);
            throw new Unauthorized(error);
        }

    } else {
        logger.info('No authentication needed');
    }
}

export default async function(req, res, next) {
    logger.debug('Authenticating request to:', req.originalUrl);
    try {
        await loadRequestData(req);
        logger.debug('Authentication completed');
        next();
    } catch (err) {
        logger.debug('Authentication failed with error:', err.message);
        next(err);
    }
}
