import mongoose from 'mongoose';
import {
    DATA_BROKER_WASABI_LIVENESS_TIMEOUT_MS,
    MESSAGE_DATA_BROKER_OFFLINE,
    DATA_BROKER_CF_FOLDER,
    DFIO_METADATA_BUCKET_NAME,
    DFIO_METADATA_BUCKET_REGION,
    DATA_BROKER_PHOENIX_LIVENESS_TIMEOUT_MS
} from '../utils/consts';
import { dateDiffMs } from '../utils/utils';
import {awsRegionCodeToDisplayName, findRelationshipsForStartup, isNewDataBroker} from '../utils/common';
import { getS3ObjectUrl } from '../lib/s3';
import { sendMessage } from '../lib/sqs';

const encryptionSchema = new mongoose.Schema({
    port: { type: Number, min: 0, max: 65535, required: false },
    certificate: { type: String, required: false }
}, { _id : false });

const dataBrokerSchema = new mongoose.Schema({
    user:                   { type: mongoose.Schema.Types.ObjectId, ref: 'User', required: true, index: true },
    name:                   { type: String, required: true },
    instanceId:             { type: String, required: false },
    privateIp:              { type: String, required: false },
    publicIp:               { type: String, required: false },
    transferRate:           { type: Number, required: false },
    region:                 { type: String, required: false },
    availabilityZone:       { type: String, required: false },
    vpcId:                  { type: String, required: false },
    vpcCidrBlock:           { type: String, required: false },
    subnetId:               { type: String, required: false },
    subnetCidrBlock:        { type: String, required: false },
    version:                { type: String, required: false },
    commandsQueueUrl:       { type: String, required: false },
    statusesQueueUrl:       { type: String, required: false },
    awsAccountId:           { type: String, required: false },
    group:                  { type: String, required: false },
    proxy:                  { type: String, required: false },
    updateNewVersion:       { type: Boolean, required: false },
    groupId:                { type: mongoose.Schema.Types.ObjectId, ref: 'Group', required: false},
    lastPing:               {
        wasabi:            { type: Date, required: false }
    },
    type:                   {type: String, enum: ['AWS', 'ONPREM', 'AZURE', 'GCP'], required: true, default: 'AWS'},
    encryption:             { type: encryptionSchema, required: false },
    hostname:               { type: String, required: false },
    platform:               { type: String, required: false },
    os:                     { type: String, required: false },
    release:                { type: String, required: false },
    osTotalMem:             { type: String, required: false },
    node:                   { type: String, required: false },
    cpus:                   { type: String, required: false },
    processMaxMem:          { type: String, required: false },
    versionUpdateSent:      { type: Date }
}, {
    timestamps: true,
    toJSON: {
        transform: (doc, ret) => {
            ret.createdAt = ret.createdAt.getTime();
            delete ret.updatedAt;
            delete ret['__v'];
            delete ret.user;
            delete ret['_id'];
            delete ret.versionUpdateSent;
            ret.id = doc.id;
            if (doc.lastPing.wasabi) {
                ret.lastPing.wasabi = doc.lastPing.wasabi ? doc.lastPing.wasabi.getTime() : 0;
            }
            if (ret.type === 'AWS' && ret.instanceId) {
                ret.placement = {
                    instanceId: ret.instanceId,
                    privateIp: ret.privateIp,
                    region: awsRegionCodeToDisplayName(ret.region),
                    availabilityZone: ret.availabilityZone,
                    vpcId: ret.vpcId,
                    vpcCidrBlock: ret.vpcCidrBlock,
                    subnetId: ret.subnetId,
                    subnetCidrBlock: ret.subnetCidrBlock,
                    version: ret.version,
                    os: ret.os,
                    release: ret.release,
                    osTotalMem: ret.osTotalMem,
                    node: ret.node,
                    cpus: ret.cpus,
                    processMaxMem: ret.processMaxMem,
                    awsAccountId: ret.awsAccountId,
                    hostname: ret.hostname,
                    platform: ret.platform,
                    publicIp: ret.publicIp
                };
            } else if ((ret.type === 'ONPREM' || ret.type === 'AZURE' || ret.type === 'GCP') && ret.hostname) {
                ret.placement = {
                    hostname: ret.hostname,
                    platform: ret.platform,
                    privateIp: ret.privateIp,
                    version: ret.version,
                    os:ret.os,
                    release: ret.release,
                    osTotalMem: ret.osTotalMem,
                    node: ret.node,
                    cpus: ret.cpus,
                    processMaxMem: ret.processMaxMem,
                    awsAccountId: ret.awsAccountId, //ONPREM have an awsAccountId - the user entered awsAccess keys
                    publicIp: ret.publicIp
                };
            }
            if (ret.encryption) {
                ret.port = ret.encryption.port;
            }
            if (doc.isRegistered()) {
                ret.status = doc.status();
                if (ret.status !== 'COMPLETE') {
                    ret.message = MESSAGE_DATA_BROKER_OFFLINE;
                }
            }

            delete ret.instanceId;
            delete ret.privateIp;
            delete ret.publicIp;
            delete ret.region;
            delete ret.availabilityZone;
            delete ret.vpcId;
            delete ret.vpcCidrBlock;
            delete ret.subnetId;
            delete ret.subnetCidrBlock;
            delete ret.version;
            delete ret.commandsQueueUrl;
            delete ret.statusesQueueUrl;
            delete ret.awsAccountId;
            delete ret.os;
            delete ret.release;
            delete ret.osTotalMem;
            delete ret.node;
            delete ret.cpus;
            delete ret.processMaxMem;
            delete ret.hostname;
            delete ret.platform;
            delete ret.encryption;

            const cftKey = `${DATA_BROKER_CF_FOLDER}/${ret.id}`;
            const installerKey = `${cftKey}_installer`;
            const resource =  ret.type === 'AWS' ? cftKey : installerKey;

            ret.fileLink = getS3ObjectUrl(DFIO_METADATA_BUCKET_REGION, DFIO_METADATA_BUCKET_NAME, resource,
                ret.type !== 'AWS', false);
            ret.isNewDataBroker = isNewDataBroker(doc);
            return ret;
        }
    }
});

dataBrokerSchema.index({ user: 1,  name: 1 }, { unique: true});

dataBrokerSchema.methods.isRegistered = function() {
    return (this.version !== undefined);
};

dataBrokerSchema.methods.isAlive = function() {
    if (isNewDataBroker(this)) {
        return this.isPhoenixAlive();
    } else {
        return this.isWasabiAlive();
    }
};

dataBrokerSchema.methods.wasWasabiAlive = function() {
    return (this.lastPing.wasabi !== undefined);
};

dataBrokerSchema.methods.isPhoenixAlive = function() {
    return this.wasWasabiAlive() && (dateDiffMs(this.lastPing.wasabi, Date.now()) < DATA_BROKER_PHOENIX_LIVENESS_TIMEOUT_MS);
};

dataBrokerSchema.methods.isWasabiAlive = function() {
    return this.wasWasabiAlive() && (dateDiffMs(this.lastPing.wasabi, Date.now()) < DATA_BROKER_WASABI_LIVENESS_TIMEOUT_MS);
};

dataBrokerSchema.methods.status = function() {
    return this.isAlive() ? 'COMPLETE' : 'FAILED';
};

dataBrokerSchema.methods.message = function() {
    return this.isAlive() ? undefined : MESSAGE_DATA_BROKER_OFFLINE;
};

dataBrokerSchema.methods.sendCommand = function(requestId, command, payload) {
    const message = {
        command,
        header: {
            dataBrokerId: this.id,
            requestId,
            clientId: this.user.id
        },
        payload
    };
    if (payload) {
        message.payload = payload;
    }
    return sendMessage(this.commandsQueueUrl, message);
};

dataBrokerSchema.methods.listCifsSharesV2 = function(requestId, host, username, password, domain = undefined) {
    const payload = {
        host,
        username,
        password
    };
    if (domain) {
        payload.domain = domain;
    }
    return this.sendCommand(requestId, 'list-cifs-shares-v2', payload);
};

dataBrokerSchema.methods.listCifsShareFoldersV2 = function(requestId, host, username, password, domain = undefined, share, path = undefined) {
    const payload = {
        host,
        username,
        password,
        share
    };
    if (domain) {
        payload.domain = domain;
    }
    if (path) {
        payload.path = path;
    }
    return this.sendCommand(requestId, 'list-cifs-share-folders-v2', payload);
};

dataBrokerSchema.methods.listNfsExportsV2 = function(requestId, host) {
    const payload = {
        host
    };
    return this.sendCommand(requestId, 'list-nfs-exports-v2', payload);
};

dataBrokerSchema.methods.listNfsExportFoldersV2 = function(requestId, host, exportName, path = undefined) {
    const payload = {
        host,
        export: exportName,
        path: path || ''
    };
    return this.sendCommand(requestId, 'list-nfs-export-folders-v2', payload);
};

dataBrokerSchema.methods.listS3BucketsV2 = function(requestId) {
    return this.sendCommand(requestId, 'list-s3-buckets-v2');
};

dataBrokerSchema.methods.listS3BucketFoldersV2 = function(requestId, bucket, prefix) {
    const payload = {
        bucket
    };
    if (prefix) {
        payload.prefix = prefix;
    }
    return this.sendCommand(requestId, 'list-s3-bucket-folders-v2', payload);
};

dataBrokerSchema.methods.listSgwsBucketsV2 = function(requestId, host, port, accessKey, secretKey) {
    const payload = {
        host,
        port,
        accessKey,
        secretKey
    };
    return this.sendCommand(requestId, 'list-sgws-buckets-v2', payload);
};

dataBrokerSchema.methods.listSgwsBucketFoldersV2 = function(requestId, host, port, accessKey, secretKey, bucket, prefix = undefined) {
    const payload = {
        bucket,
        host,
        port,
        accessKey,
        secretKey
    };
    if (prefix) {
        payload.prefix = prefix;
    }
    return this.sendCommand(requestId, 'list-sgws-bucket-folders-v2', payload);
};

dataBrokerSchema.methods.startup = async function(requestId) {
    const dataBroker = this;
    const relationships = await findRelationshipsForStartup(dataBroker);
    const payload = {
        relationships: relationships
    };
    return this.sendCommand(requestId, 'startup', payload);
};

dataBrokerSchema.methods.toString = function toString() {
    const { id, name, user, version} = this;
    return `DataBroker ${name} with version ${version} (${id}) of ${user}`;
};

dataBrokerSchema.statics.findUsersByAwsAccountId = function(awsAccountId) {
    return this.aggregate([
        {
            $lookup: {
                from: 'users',
                localField: 'user',
                foreignField: '_id',
                as: 'userData'
            }
        },
        {
            $unwind: '$userData'
        },
        {
            $match: {
                'awsAccountId': `${awsAccountId}`
            }
        },
        {
            $group: {
                _id: '$user',
                subscription: { $first : '$userData.subscription'}
            }
        },
        {
            $project: {
                _id: true,
                subscription: true
            }
        }
    ]);
};

export default mongoose.model('DataBroker', dataBrokerSchema);
