import mongoose from 'mongoose';
import { defaultTransform } from '../utils/utils';


const statisticSchema = new mongoose.Schema({
    bytes:     {type: Number, required: true, default: 0},
    items:    {type: Number, required: true, default: 0}
}, {
    _id: false,
    toJSON: {
        transform: defaultTransform((doc, ret) => {
            return ret;
        })
    }
});

const counterSchema = new mongoose.Schema({
    file: {type: statisticSchema, required: true},
    directory: {type: statisticSchema, required: true},
    symbolic: {type: statisticSchema, required: true},
    other: {type: statisticSchema, required: true}
});

const TransferSchema = new mongoose.Schema({
}, {
    _id: false,
    delete: {type: counterSchema, required: true},
    failedDelete: {type: counterSchema, required: true},
    copy: {type: counterSchema, required: true},
    failedCopy: {type: counterSchema, required: true},
    abort: {type: counterSchema, required: true}
}, {
    toJSON: {
        transform: defaultTransform((doc, ret) => {
            return ret;
        })
    }
});

const ScanSchema = new mongoose.Schema({
}, {
    _id: false,
    failed: {type: counterSchema, required: true},
    delete: {type: counterSchema, required: true},
    copy: {type: counterSchema, required: true},
    grace: {type: counterSchema, required: true},
    ignore: {type: counterSchema, required: true},
    abort: {type: counterSchema, required: true}
}, {
    toJSON: {
        transform: defaultTransform((doc, ret) => {
            return ret;
        })
    }
});


const ListSchema = new mongoose.Schema({
}, {
    _id: false,
    failed: {type: counterSchema, required: true},
    list: {type: counterSchema, required: true}
}, {
    toJSON: {
        transform: defaultTransform((doc, ret) => {
            return ret;
        })
    }
});

const relationshipStatisticsSchema = mongoose.Schema({
    relationship:   {type: mongoose.Schema.Types.ObjectId, ref: 'Relationship', required: true, index: true},
    dataBroker:     {type: mongoose.Schema.Types.ObjectId, ref: 'DataBroker', required: true, index: true},
    user:           {type: mongoose.Schema.Types.ObjectId, ref: 'User', required: true, index:true},
    scan:           {type: ScanSchema, required: false},
    transfer:       {type: TransferSchema, required: false},
    list:           {type: ListSchema, required: false},
    status:         {type: String, required:false},
    reason:         {type: String}
}, {
    timestamps: true,
    toJSON: {
        transform: defaultTransform((doc, ret) => {
            return ret;
        })
    }
});

relationshipStatisticsSchema.index({ relationship: 1, dataBroker: 1, user: 1 }, {unique: true});

relationshipStatisticsSchema.statics.findRelationshipStatus = async function(relationshipId, userId) {
    return this.aggregate([
        {$match: {relationship: mongoose.Types.ObjectId(relationshipId), user: mongoose.Types.ObjectId(userId)}},
        {$group: {_id: '$relationship',
            statuses : {$addToSet: {status:'$status', reason: '$reason'}}}}]);
};

relationshipStatisticsSchema.statics.sumStatsPerRelationship = function(relationshipId) {
    return this.aggregate([
        {$match: {relationship: mongoose.Types.ObjectId(relationshipId)}},
        {$group: {_id: '$relationship',
            filesMarkedForCopy:     {$sum: '$scan.copy.file.items'},
            bytesMarkedForCopy:     {$sum: '$scan.copy.file.bytes'},
            dirsMarkedForCopy:      {$sum: '$scan.copy.directory.items'},
            symbolicMarkedForCopy:  {$sum: '$scan.copy.symbolic.items'},
            filesCopied:            {$sum : '$transfer.copy.file.items'},
            dirsCopied:             {$sum : '$transfer.copy.directory.items'},
            bytesCopied:            {$sum: '$transfer.copy.file.bytes'},
            symbolicCopied:         {$sum: '$transfer.copy.symbolic.items'},
            filesFailed:            {$sum : '$transfer.copy_failed.file.items'},
            dirsFailed :            {$sum : '$transfer.copy_failed.directory.items'},
            bytesFailed:            {$sum: '$transfer.copy_failed.file.bytes'},
            symbolicFailed:         {$sum: '$transfer.copy_failed.symbolic.items'},
            filesRemoved:           {$sum: '$transfer.delete.file.items'},
            bytesRemoved:           {$sum: '$transfer.delete.file.bytes'},
            dirsRemoved:            {$sum: '$transfer.delete.directory.items'},
            symbolicRemoved:        {$sum: '$transfer.delete.symbolic.items'},
            bytesMarkedForRemove:   {$sum: '$scan.delete.file.bytes'},
            filesMarkedForRemove:   {$sum: '$scan.delete.file.items'},
            dirsMarkedForRemove:    {$sum: '$scan.delete.directory.items'},
            symbolicMarkedForRemove: {$sum: '$scan.delete.symbolic.items'},
            bytesRemovedFailed:     {$sum: '$transfer.delete_failed.file.bytes'},
            filesRemovedFailed:     {$sum: '$transfer.delete_failed.file.items'},
            dirsRemovedFailed:      {$sum: '$transfer.delete_failed.directory.items'},
            symbolicRemovedFailed:  {$sum: '$transfer.delete_failed.symbolic.items'},
            filesMarkedForGrace:    {$sum: '$scan.grace.file.items'},
            bytesMarkedForGrace:    {$sum: '$scan.grace.file.bytes'},
            dirsMarkedForGrace:     {$sum: '$scan.grace.directory.items'},
            symbolicMarkedForGrace: {$sum: '$scan.grace.symbolic.items'},
            filesMarkedForIgnore:   {$sum: '$scan.ignore.other.items'},
            dirsScanned:             {$sum: '$list.list.directory.items'},
            filesScanned:           {$sum: '$list.list.file.items'},
            symbolicScanned:        {$sum: '$list.list.symbolic.items'},
            bytesScanned:           {$sum: '$list.list.file.bytes'},
            dirsFailedToScan:       {$sum: '$list.failed.directory.items'},
            lastMessageTime:        {$max: '$updatedAt'},
            statuses :              {$addToSet: '$statusMessages.status'}}}]);
};

export default mongoose.model('RelationshipStats', relationshipStatisticsSchema);
