import mongoose from 'mongoose';
import { get, isNumber} from 'lodash';
import { defaultTransform } from '../utils/utils';
import { sendMessage } from '../lib/sqs';
import {
    RUNNING_STATUS, COPY_RELATIONSHIP_MESSAGE, SYNC_RELATIONSHIP_MESSAGE,
    MESSAGE_STATUS_DONE, SYNC_TYPE
} from '../utils/consts';
import {calculateStatus, findMaxTimestampForAction, createObjectForAsterix,
    calculateProgress, isNewDataBroker} from '../utils/common';

import {getLogger} from '../utils/logger';
const logger = getLogger();


function calculateDuration(messages) {
    const maxTimestamp = findMaxTimestampForAction(messages, 'timestamp');
    if (maxTimestamp && messages.setup && messages.setup.status === MESSAGE_STATUS_DONE.toLowerCase()) {
        return maxTimestamp - (messages.setup.timestamp);
    }
}

function createScanningObject(status, activeMessage, isFromS3, type) {
    if (status === RUNNING_STATUS) {
        if (get(activeMessage, 'copy.status', '') !== 'started') {
            return undefined;
        } else if (isFromS3 && type === 'Sync') {
            return undefined;
        } else {
            return {
                objectsCompleted: get(activeMessage, 'index.stats.itemsIndexed', 0),
                filesCompleted: get(activeMessage, 'scan.stats.itemsScanned', 0),
                objectsIgnored: 0
            };
        }
    }
    return undefined;
}

const typesSchema = new mongoose.Schema({
    files: {type: Boolean, required: true, default: true},
    directories: {type: Boolean, required: true, default: true},
    symlinks: {type: Boolean, required: true, default: true}
}, {
    _id: false,
    toJSON: {
        transform: defaultTransform((doc, ret) => {
            return ret;
        })
    }
});

const settingsSchema = new mongoose.Schema({
    gracePeriod: {type: Number, required: true, default: 30},
    deleteOnTarget: {type: Boolean, required: true, default: false},
    retries: {type: Number, min: 0, max: 5, required: true, default: 3},
    fileTypes: {type: typesSchema, required: true, default: {}}
}, {
    _id: false,
    toJSON: {
        transform: defaultTransform((doc, ret) => {
            return ret;
        })
    }
});


const endPointSchema = new mongoose.Schema({
    protocol:   {type: String, enum: ['nfs', 'cifs', 's3', 'sgws', 'local', 'azure', 'gcp'], required: true},
    nfs:    {type: mongoose.Schema.Types.Mixed, required: false},
    cifs:   {type: mongoose.Schema.Types.Mixed, required: false},
    s3:     {type: mongoose.Schema.Types.Mixed, required: false},
    sgws:   {type: mongoose.Schema.Types.Mixed, required: false},
    local:  {type: mongoose.Schema.Types.Mixed, required: false},
    azure:  {type: mongoose.Schema.Types.Mixed, required: false},
    gcp:    {type: mongoose.Schema.Types.Mixed, required: false}
}, {
    _id: false,
    toJSON: {
        transform: defaultTransform((doc, ret) => {
            return ret;
        })
    }
});

const syncSchema = new mongoose.Schema({
    inDays:     {type: Number, required: true, default: 1},
    inHours:    {type: Number, required: true, default: 0},
    inMinutes:  {type: Number, required: true, default: 0},
    nextTime:   {type: Date, required: true},
    autoSync:   {type: String, enum: ['ON', 'OFF'], required: true, default: 'ON'}
}, {
    _id: false,
    toJSON: {
        transform: defaultTransform((doc, ret) => {
            return ret;
        })
    }
});


const relationshipSchema = mongoose.Schema({
    user:               {type: mongoose.Schema.Types.ObjectId, ref: 'User', index: true},
    dataBroker:         {type: mongoose.Schema.Types.ObjectId, ref: 'DataBroker'},
    source:             {type: endPointSchema, required: true},
    target:             {type: endPointSchema, required: true},
    messages:           {type: mongoose.Schema.Types.Mixed, required: false},
    settings:           {type: settingsSchema, required: true},
    sync:               {type : syncSchema, required: true},
    isQstack:           { type: Boolean, required: true, default: false },
    isCvo:              { type: Boolean, required: true, default: false },
    scannerQueue:       { type: String, required: false },
    transferrerQueue:   { type: String, required: false },
    phase:              {type: String, enum: ['Initial Copy', 'Sync'], required:true, default: 'Initial Copy'},
    startTime:          {type: Date, required: false},
    endTime:            {type: Date, required: false},
    group:              {type: mongoose.Schema.Types.ObjectId, ref: 'Group', index:true}
}, {
    timestamps: true,
    toJSON: {
        virtuals: true,
        transform: defaultTransform((doc, ret) => {
            // if there are still not messages - fake running copy message
            if (!isNewDataBroker(doc.dataBroker)) {
                doc.messages = doc.messages ||
                    {
                        'create-relationship-v2': {
                            setup: {
                                relationshipId: doc._id,
                                step: 'setup',
                                status: 'done',
                                timestamp: Date.now()
                            }
                        }
                    };
                const typeAndActiveMessage = doc.findTypeAndActiveMessage();
                const  { activeMessage } = typeAndActiveMessage;
                const  { type } = typeAndActiveMessage;
                const statusObject = calculateStatus(doc.messages);
                const status = get(statusObject, 'status');
                ret.activity = {
                    type,
                    status,
                    progress: calculateProgress(status, activeMessage),
                    failureMessage: get(statusObject, 'message'),
                    lastMessageTime: statusObject.timestamp,
                    filesCopied: get(activeMessage, 'copy.stats.itemsTransferred', 0),
                    filesFailed: get(activeMessage, 'copy.stats.itemsTransferFailed') || get(activeMessage, 'copy.stats.itemTransfersFailed', 0),
                    filesRemoved: get(activeMessage, 'copy.stats.itemsDeleted', 0),
                    bytesCopied: get(activeMessage, 'copy.stats.totalTransferredBytes') || get(activeMessage, 'copy.stats.bytesTransferred', 0),
                    bytesFailed: get(activeMessage, 'copy.stats.itemsTransferFailedBytes') || get(activeMessage, 'copy.stats.bytesFailedTransferring', 0),
                    bytesMarkedForCopy : get(activeMessage, 'copy.stats.itemsMarkedForTransferBytes'),
                    filesMarkedForCopy : get(activeMessage, 'scan.stats.itemsMarkedForTransfer'),
                    filesScanned: get(activeMessage, 'scan.stats.itemsScanned'),
                    dirsFailedToScan : get(activeMessage, 'scan.stats.itemsScannedFailed'),
                    filesMarkedForGrace : get(activeMessage,  'scan.stats.itemsWithinGracePeriod'),
                    filesRemovedFailed : get(activeMessage, 'copy.stats.itemsDeleteFailed'),
                    filesMarkedForIgnore: get(activeMessage, 'scan.stats.itemsTransferIgnored'),
                    filesMarkedforRemove: get(activeMessage, 'scan.stats.itemsMarkedForDelete'),
                    executionTime: calculateDuration(activeMessage),
                    dirsMarkedForCopy : 0,
                    dirsCopied: 0,
                    dirsFailed: 0,
                    bytesMarkedForRemove : 0,
                    dirsMarkedForRemove: 0,
                    bytesRemoved : 0,
                    dirsRemoved: 0,
                    bytesRemovedFailed: 0,
                    dirsMarkedForGrace: 0,
                    bytesMarkedForGrace: 0,
                    dirsScanned: 0,
                    bytesScanned:0,
                    scanning: createScanningObject(status, activeMessage, doc.isEndPointProtocol('source', 's3'), type)
                };
            }
            ret.relationshipId = doc._id;
            return ret;
        })
    }
});

relationshipSchema.methods.isEndPointProtocol = function(typeOfEndPoint, protocol) {
    return this[typeOfEndPoint].protocol  === protocol;
};

relationshipSchema.methods.findTypeAndActiveMessage = function() {
    if (this.messages) {
        const type = this.messages['sync-relationship'] ? 'Sync' : 'Initial Copy';
        const activeMessage = (type === SYNC_TYPE) ? this.messages[SYNC_RELATIONSHIP_MESSAGE] : this.messages[COPY_RELATIONSHIP_MESSAGE];
        return {type, activeMessage};
    }
};

relationshipSchema.virtual('autoSync').get(function() {
    return this.sync.autoSync;
});

relationshipSchema.index({ user: 1, _id: 1});
relationshipSchema.index({ user: 1, dataBroker: 1, source: 1, target: 1 }, {unique: true});//TODO when moving to only groups need to remove the index
// relationshipSchema.index({user: 1, group: 1, source: 1, target: 1}, {unique: true});

relationshipSchema.methods.startCopy = function(requestId, sourceCredentials, targetCredentials, additionalInfo) { //wont work now- need to change
    const source = createObjectForAsterix(this.source, sourceCredentials);
    const target = createObjectForAsterix(this.target, targetCredentials);
    const relationshipMessageJson = {userEmail : this.user.email, relationshipId: this.id, isQstack: this.isQstack, timeStamp: Date.now(), source, target, requestId};
    logger.info(`[CreateRelationship Message]${JSON.stringify(relationshipMessageJson)}`);
    const gracePeriod =  get(additionalInfo, 'gracePeriod');
    const deleteOnTarget =  get(additionalInfo, 'deleteOnTarget');
    const payload = {
        relationshipId: this.id,
        gracePeriod : (isNumber(gracePeriod)) ? `${gracePeriod} seconds` : '30 seconds',
        deleteOnTarget,
        source,
        target
    };
    if ((this.isEndPointProtocol('source', 'nfs') && this.isEndPointProtocol('target', 's3')) || (this.isEndPointProtocol('source', 's3') && this.isEndPointProtocol('target', 'nfs'))) {
        payload.restorable = true;
    }
    return sendMessage(this.dataBroker.commandsQueueUrl, {
        command: 'create-relationship-v2',
        header: {
            dataBrokerId: this.dataBroker.id,
            requestId,
            clientId: this.user.id
        },
        payload
    });
};

relationshipSchema.methods.startSync = function(requestId) {
    const relationshipMessageJson = {userEmail : this.user.email, relationshipId: this.id, isQstack: this.isQstack, timeStamp: Date.now(), source: this.source, target: this.target, requestId};
    logger.info(`[SyncRelationship Message]${JSON.stringify(relationshipMessageJson)}`);
    return sendMessage(this.dataBroker.commandsQueueUrl, {
        command: 'sync-relationship',
        header: {
            dataBrokerId: this.dataBroker.id,
            clientId: this.user.id,
            requestId
        },
        payload: {
            relationshipId : this.id
        }
    });
};

relationshipSchema.methods.startAbort = function(requestId) {
    const relationshipMessageJson = {userEmail : this.user.email, relationshipId: this.id, timeStamp: Date.now(), source: this.source, target: this.target, requestId};
    logger.info(`[AbortRelationship Message]${JSON.stringify(relationshipMessageJson)}`);
    return sendMessage(this.dataBroker.commandsQueueUrl, {
        command: 'abort-relationship',
        header: {
            dataBrokerId: this.dataBroker.id,
            clientId: this.user.id,
            requestId
        },
        payload: {
            relationshipId : this.id
        }
    });
};

relationshipSchema.statics.getRelationshipsForScheduledSync = function() {
    return this.find({$and: [
        {'sync.autoSync': 'ON'},
        {'sync.nextTime': {$lte: new Date()}},
        {$or: [{'sync.inDays' : {$gt : 0}}, {'sync.inMinutes' : {$gt : 0}}, {'sync.inHours' : {$gt : 0}}]}]}).populate('user dataBroker')
        .populate({path: 'group', populate: {path: 'dataBrokers'}}).exec();
};

relationshipSchema.methods.doDelete = function(requestId) {
    return sendMessage(this.dataBroker.commandsQueueUrl, {
        command: 'delete-relationship',
        header: {
            dataBrokerId: this.dataBroker.id,
            clientId: this.user.id,
            requestId
        },
        payload: {
            relationshipId: this.id
        }
    });
};

export default mongoose.model('Relationship', relationshipSchema);
