import mongoose from 'mongoose';
import Promise from 'bluebird';
import crypto from 'crypto';
import moment from 'moment';
import { JWT_SECRET, SAAS_REGISTRATION_STATUS_RESOLVE_FAILED,
    SAAS_REGISTRATION_STATUS_UNSUBSCRIBED_INIT, PAYGO_SUBSCRIPTION} from '../utils/consts';
import { PreconditionFailed } from '../utils/errors';
import Relationship from './relationship';
import Message from './message';
import { setHubSpotContactFirstSubscriptionSinceIfNeeded } from '../lib/hubspot';

const algorithm = 'aes-256-ctr';
const notEncryptedEncoding = 'utf8';
const encryptedEncoding = 'hex';

const nssCredentialsSchema = new mongoose.Schema({
    nssUsername: { type: String, required: false },
    nssPassword: { type: String, required: false }
}, { _id : false });

const licensesSchema = new mongoose.Schema({
    serialNumber: { type: String, required: true },
    reservedRelationships: { type: Number, required: false },
    expirationDate: { type: Date, required: false }
}, { _id : false });

const subscriptionSchema = new mongoose.Schema({
    customerIdentifier: { type: String, required: false },
    status: { type: String, required: false, default: SAAS_REGISTRATION_STATUS_UNSUBSCRIBED_INIT },
    updatedAt: { type: Date, required: false, default: new Date() },
    type: {type: String, enum: ['PAYGO', 'CONTRACT'], required: false, default: PAYGO_SUBSCRIPTION}
}, { _id : false });

const userSchema = new mongoose.Schema({
    name:           { type: String, required: true },
    email:          { type: String, required: true, lowercase: true, index: true },
    company:        { type: String, required: false },
    isOccm:         { type: Boolean, required: true, default: false },
    isQstack:         { type: Boolean, required: true, default: false },
    trialExpiry:    { type: Date, required: true },
    lastLogin:      { type: Date, required: false },
    portalId:        {type: String, required: true, index: { unique: true }},
    byol: {
        nssCredentials: { type: nssCredentialsSchema, required: false },
        licenses:    [licensesSchema]
    },
    subscription : {type: subscriptionSchema, required: true, default: {}}
}, {
    timestamps: true
});

// function can't be arrow function as this is undefined in arrow functions

userSchema.pre('save', function(next) {

    Promise.resolve()
        .then(() => {
            if (this.isModified('byol.nssCredentials')) {
                const cipher = crypto.createCipher(algorithm, JWT_SECRET);
                let crypted = cipher.update(this.byol.nssCredentials.nssPassword, notEncryptedEncoding, encryptedEncoding);
                crypted += cipher.final(encryptedEncoding);
                this.byol.nssCredentials.nssPassword = crypted;
            }

            if (this.isModified('subscription.status')) {
                if (![SAAS_REGISTRATION_STATUS_RESOLVE_FAILED, SAAS_REGISTRATION_STATUS_UNSUBSCRIBED_INIT].includes(this.subscription.status)) {
                    if (this.subscription.status === 'unsubscribe-success') {
                        this.subscription.customerIdentifier = undefined;
                        this.subscription.type = 'PAYGO';
                    }
                    this.subscription.updatedAt = new Date();
                }

                // signal to post save hook that the field was modified for additional handling
                this.subscriptionStatusWasModified = true;
            }
        })
        .then(() => next())
        .catch(next);
});

userSchema.post('save', function() {
    // flag set by the pre save hook
    if (this.subscriptionStatusWasModified) {
        delete this.subscriptionStatusWasModified;

        if (this.subscription.status === 'subscribe-success') {
            setHubSpotContactFirstSubscriptionSinceIfNeeded(this);
        }
        // we delay the message some time because the client will first need to initiate the client messages
        // poller upon login, and if we save the message immediately, the client will never receive it
        setTimeout(() => {
            Message.SendAsync(undefined, this.id, undefined, 'update-subscription-status', {
                userId: this.id,
                subscriptionStatus: this.subscription.status,
                subscriptionUpdatedAt: this.subscription.updatedAt,
                subscriptionType: this.subscription.type
            });
        }, 10000);
    }
});

userSchema.methods.isNetApp = function() {
    return /netapp/i.test(this.company) || /@netapp\.com$/i.test(this.email);
};

userSchema.methods.decryptedNssPassword = function() {
    if (!this.byol.nssCredentials) {
        throw new PreconditionFailed('No BYOL credentials for user');
    }
    const decipher = crypto.createDecipher(algorithm, JWT_SECRET);
    let dec = decipher.update(this.byol.nssCredentials.nssPassword, encryptedEncoding, notEncryptedEncoding);
    dec += decipher.final(notEncryptedEncoding);
    return dec;
};

userSchema.methods.toString = function toString() {
    const { id, name } = this;
    return `${name} (${id})`;
};

userSchema.statics.BillableUsers = function() {
    return this.find({ $and: [{ 'subscription.customerIdentifier': { $exists: true }}, {'subscription.type' : 'PAYGO'},
        {trialExpiry: { $exists: true, $lt: new Date() }}, {email: { $not : /netapp/i }}, {company: { $not : /netapp/i}} ]}).exec();
};

userSchema.statics.useraWithTrialExpired = function() {
    return this.find({ $and: [ { trialExpiry: { $exists: true } }, { trialExpiry: { $lt: new Date() } }]}).exec();
};

userSchema.statics.customerIdentifierByolEntitlement = function() {
    return this.aggregate([
        {$match: { $and :[{'subscription.customerIdentifier': { $exists: true }}, {'subscription.type' : 'PAYGO'}]}},
        {$unwind: '$byol.licenses'},
        {$match: {'byol.licenses.expirationDate': {$gt: new Date()}}},
        {$group: {_id: '$subscription.customerIdentifier', count: {$sum: '$byol.licenses.reservedRelationships'}}}]);
};

userSchema.statics.userByolEntitlementForExpiredTrial = function() {
    return this.aggregate([
        {$match: {$and: [{ trialExpiry: { $exists: true } }, { trialExpiry: { $lt: new Date() } }]}},
        {$unwind: '$byol.licenses'},
        {$match: {'byol.licenses.expirationDate': {$gt: new Date()}}},
        {$group: {_id: '$_id', count: {$sum: '$byol.licenses.reservedRelationships'}}}]);
};


userSchema.methods.relationshipsCount = function() {
    return  Relationship.find({user: this}).exec()
        .then(relationships => {
            return relationships ? relationships.length : 0;
        });
};

userSchema.methods.reservedRelationships = function() {
    let reservedRelationships = 0;
    if (this.byol.licenses.length > 0) {
        this.byol.licenses.map(license => {
            if (moment(license.expirationDate).startOf('day').isAfter()) {
                reservedRelationships = reservedRelationships + license.reservedRelationships;
            }
        });
    }
    return reservedRelationships;
};

userSchema.methods.addTrialDaysForNetappUsersAndSaveIfNeeded = function() {
    if ((this.isNetApp() || this.email.includes('unitek.com') || this.email.includes('fastlaneus.com') ||
            this.email.includes('ihcnetappu.com') || this.email.includes('trainocate.in') || this.email.includes('netapp-test.com'))
        && this.subscription.type === PAYGO_SUBSCRIPTION) {
        const today = moment();
        const trialDaysLeft = moment(this.trialExpiry).diff(today, 'days');
        if (trialDaysLeft <= 8)  {
            this.trialExpiry = today.add(14, 'day').toDate();
            return this.save();
        }
    }
    return Promise.resolve(this);
};

userSchema.methods.calculateDaysLeftForTrial = function() {
    const daysLeftForTrial = moment(this.trialExpiry).startOf('day').diff(moment().startOf('day'), 'days');
    return (daysLeftForTrial >= 0) ? daysLeftForTrial : -1;
};

export default mongoose.model('User', userSchema);
