import {isArray, remove} from 'lodash';
import {sendMessage, getQueueAttributes, setQueuePolicy} from '../lib/sqs';
import {findRelationshipsForNewDataBrokerStartup} from '../utils/common';
import {BadRequest} from '../utils/errors';
import Group from '../models/group';
import {getLogger} from '../utils/logger';
import {getUserArn} from '../lib/iam';
const logger = getLogger();


export function sendListExportsCommand(requestId, host, dataBroker) {
    const {commandsQueueUrl, id: dataBrokerId} = dataBroker;
    return sendMessage(commandsQueueUrl, {host}, {'type': 'exports', requestId, dataBrokerId});
}

export function sendListExportsFoldersCommand(requestId, host, exportName, path, version, dataBroker) {
    const {commandsQueueUrl, id: dataBrokerId} = dataBroker;
    const payload = {host, export: exportName, version};
    if (path !== '') {
        payload.path = path;
    }
    return sendMessage(commandsQueueUrl, payload, {'type': 'export-folders', requestId, dataBrokerId});
}

export function sendListBucketsCommand(requestId, dataBroker) {
    const {commandsQueueUrl, id: dataBrokerId} = dataBroker;
    return sendMessage(commandsQueueUrl, {}, {'type': 'buckets', requestId,  dataBrokerId});
}

export function sendListBucketsFolders(requestId, bucketName, prefix, dataBroker) {
    const {commandsQueueUrl, id: dataBrokerId} = dataBroker;
    const payload = {bucket: bucketName};
    if (prefix !== '') {
        payload.prefix = prefix;
    }
    return sendMessage(commandsQueueUrl, payload, {'type': 'bucket-folders', requestId, dataBrokerId});
}

export function sendListSgwsBucketsCommand(requestId, dataBroker, host, port, accessKey, secretKey) {
    const {commandsQueueUrl, id: dataBrokerId} = dataBroker;
    return sendMessage(commandsQueueUrl, {host, port, accessKey, secretKey}, {'type': 'sgws-buckets', requestId,  dataBrokerId});
}

export function sendListSgwsBucketsFoldersCommand(requestId, dataBroker, host, port, accessKey, secretKey, bucket, prefix) {
    const {commandsQueueUrl, id: dataBrokerId} = dataBroker;
    return sendMessage(commandsQueueUrl, {host, port, accessKey, secretKey, bucket, prefix}, {'type': 'sgws-buckets-folders', requestId,  dataBrokerId});
}

async function addUserToRelationshipQueues(relationshipQueue, userArn) {
    const queuePolicy = await getQueueAttributes(relationshipQueue, ['Policy']);
    const jsonPolicy = JSON.parse(queuePolicy.Policy);
    const policyPrincipals = jsonPolicy.Statement[0].Principal.AWS;
    if (isArray(policyPrincipals)) {
        if (!policyPrincipals.includes(userArn)) {
            jsonPolicy.Statement[0].Principal.AWS.push(userArn);
            await setQueuePolicy(relationshipQueue, jsonPolicy);
        }
    } else {
        if (policyPrincipals !== userArn) {
            jsonPolicy.Statement[0].Principal.AWS  = [policyPrincipals, userArn];
            await setQueuePolicy(relationshipQueue, jsonPolicy);
        }
    }
}

export async function removeUserFromRelationshipQueues(relationshipQueue, userArn) {
    const queuePolicy = await getQueueAttributes(relationshipQueue, ['Policy']);
    const jsonPolicy = JSON.parse(queuePolicy.Policy);
    const policyPrincipals = jsonPolicy.Statement[0].Principal.AWS;
    if (isArray(policyPrincipals)) {
        if (policyPrincipals.includes(userArn)) {
            remove(jsonPolicy.Statement[0].Principal.AWS, user => user === userArn);
            await setQueuePolicy(relationshipQueue, jsonPolicy);
        }
    }
}


export async function updateRelationshipQueuesWithDataBrokerUser(userName, group, removeUser) {
    const relationships = await findRelationshipsForNewDataBrokerStartup(group);

    relationships.forEach(async relationship => {
        const userArn = await getUserArn(userName);

        const { scannerQueue, transferrerQueue } = relationship;

        if (removeUser) {
            await Promise.all([removeUserFromRelationshipQueues(scannerQueue, userArn),
                removeUserFromRelationshipQueues(transferrerQueue, userArn)]);
        } else {
            await Promise.all([addUserToRelationshipQueues(scannerQueue, userArn),
                addUserToRelationshipQueues(transferrerQueue, userArn)]);
        }
    });
}

export async function sendStartupCommand(requestId, dataBroker, metadata = {}) {
    const {commandsQueueUrl, id: dataBrokerId, groupId} = dataBroker;
    const groupFromDb = await Group.findById(groupId).exec();
    const relationships = await findRelationshipsForNewDataBrokerStartup(groupFromDb);
    return sendMessage(commandsQueueUrl, { relationships, metadata }, {'type': 'startup', requestId, dataBrokerId});
}

export async function sendMetadataRequestCommand(requestId, dataBroker, targetDataBrokerId) {
    const {commandsQueueUrl, id: dataBrokerId} = dataBroker;
    logger.info(`Sending metadata command to databroker ${dataBrokerId}  to queue ${commandsQueueUrl}`);
    return sendMessage(commandsQueueUrl, { targetDataBrokerId}, {'type': 'metadata', requestId, dataBrokerId});
}


export async function sendShutdownCommand(requestId, dataBroker) {
    const {commandsQueueUrl, id: dataBrokerId} = dataBroker;
    logger.info(`Sending shutdown command to databroker ${dataBrokerId}`);
    return sendMessage(commandsQueueUrl, {}, {'type': 'shutdown', requestId, dataBrokerId});
}

export async function sendListAzureBlobStorageContainersFoldersCommand(requestId, dataBroker, container, prefix, storageAccountName,
    storageAccountConnectionString) {
    const {commandsQueueUrl, id: dataBrokerId} = dataBroker;
    let message = {container,
        prefix
    };
    if (storageAccountConnectionString) {
        message = Object.assign({}, message, {storageAccountName, storageAccountConnectionString});
    } else {
        throw new BadRequest('list-azure-blob-storage-containers request should contain storage-account-connection-string');
    }

    return sendMessage(commandsQueueUrl, message, {type: 'azure-blob-storage-container-folders', requestId, dataBrokerId});
}

export async function sendListAzureBlobStorageContainers(requestId, dataBroker, storageAccountName, storageAccountConnectionString) {
    const {id: dataBrokerId, commandsQueueUrl} = dataBroker;
    let message;
    if (storageAccountConnectionString) {
        message = {storageAccountName, storageAccountConnectionString};
    } else {
        throw new BadRequest('list-azure-blob-storage-containers request should contain storage-account-connection-string');
    }
    return sendMessage(commandsQueueUrl, message, {
        type: 'azure-blob-storage-containers',
        requestId,
        dataBrokerId
    });
}

export async function sendListGcpBucketsCommand(requestId, dataBroker, projectId, clientEmail, privateKey) {
    const {id: dataBrokerId, commandsQueueUrl} = dataBroker;
    return sendMessage(commandsQueueUrl, {projectId, clientEmail, privateKey}, {
        type: 'gcp-buckets',
        requestId,
        dataBrokerId
    });
}

export async function sendListGcpBucketFoldersCommand(requestId, dataBroker, bucket, prefix, projectId, clientEmail, privateKey) {
    const {commandsQueueUrl, id: dataBrokerId} = dataBroker;
    return sendMessage(commandsQueueUrl, {
        projectId,
        clientEmail,
        privateKey,
        bucket,
        prefix
    }, {type: 'gcp-bucket-folders', requestId, dataBrokerId});
}

export async function sendUpdateVersionCommand(requestId, dataBroker, latestVersion) {
    const {commandsQueueUrl, id: dataBrokerId, group, user} = dataBroker;
    const groupId = await Group.findOne({name: group, user}).exec();
    const  relationships = await findRelationshipsForNewDataBrokerStartup(dataBroker, groupId);
    return sendMessage(commandsQueueUrl, {
        bundleUrl: latestVersion.url,
        bundleVersion: latestVersion.version,
        relationships
    }, {'type': 'update', requestId, dataBrokerId});
}

export function sendSharesCommand(requestId, dataBroker, host, version, username, password, domain) {
    const {id: dataBrokerId, commandsQueueUrl} = dataBroker;
    return sendMessage(commandsQueueUrl, {host, username, version, password, workgroup: domain}, {
        type: 'shares',
        requestId,
        dataBrokerId
    });
}

export function sendShareFoldersCommand(requestId, dataBroker, host, share, path, version, username, password, domain) {
    const {id: dataBrokerId, commandsQueueUrl} = dataBroker;
    return sendMessage(commandsQueueUrl, {host, username, password, share, path: path === '' ? undefined : path, version, workgroup: domain}, {
        type: 'share-folders',
        requestId,
        dataBrokerId
    });
}
