import retry from 'async-retry';
import { isObject, uniq, omit, head, keys } from 'lodash';
import {sendMessage} from '../lib/sqs';
import {createEndPointForNewDataBroker, isNewDataBroker, assertIpOrHostname} from '../utils/common';
import { removeFirstSlash } from '../utils/utils';
import { BadRequest } from '../utils/errors';
import {ALLOWED_NEW_DATABROKER_SOURCE_AND_TARGET, ALLOWED_SOURCE_AND_TARGET,
    S3_TAG_KEY_REGEXP, S3_TAG_VALUE_REGEXP, SCAN_MODE_SOURCE_AND_TARGET
} from '../utils/consts';
import {getLogger} from '../utils/logger';

const logger = getLogger();


function addEndSlash(path) {
    return (path) ? (path.endsWith('/') ? path : `${path}/`) : '';
}

async function sendMessageToScanner(source, target, relationshipId, scannerQueue, dataBrokerId, requestId) {
    logger.info(`Sending message to the scanner ${scannerQueue} of relationship ${relationshipId}`);
    await sendMessage(scannerQueue, {
        relationshipId,
        source,
        target,
        type: SCAN_MODE_SOURCE_AND_TARGET
    }, {type : 'copy', requestId, dataBrokerId});
}

async function sendMessageToAllDataBrokersInAGroup(group, message, messageType, requestId) {
    const {dataBrokers} = group;
    const {relationshipId} = message;
    const commands = dataBrokers.map(async dataBroker => {
        // send to all the data brokers- if one data broker is down it wont receive done command
        // but he will also wont receive the relationship in the startup command
        const {commandsQueueUrl, id: dataBrokerId} = dataBroker;
        if (commandsQueueUrl) { //in case the data broker was not created yet
            try {
                await retry(() => sendMessage(commandsQueueUrl, message, {
                    type: messageType,
                    requestId,
                    dataBrokerId
                }), {
                    retries: 3,
                    onRetry: err => logger.warn(`RETRYING sending ${messageType} command to data broker ${dataBrokerId} for relationship ${relationshipId} with error ${JSON.stringify(err)}`)
                });
            } catch (err) {
                logger.error(`Failed to send ${messageType} to dataBroker ${dataBrokerId}`);
            }
        }
    });
    await Promise.all(commands);
}

function omitSecretKeys(object) {
    const SECRET_KEYS = ['password', 'privateKey', 'storageAccountKey', 'storageAccountConnectionString', 'secretKey', 'certificate', 'serverName'];
    return omit(object, SECRET_KEYS.map(key => `${head(keys(object))}.${key}`));
}

export async function sendCopyCommand(relationship, requestId, sourceCredentials, targetCredentials) {
    const {source: sourceFromDb, target: targetFromDb, id: relationshipId, scannerQueue, transferrerQueue,
        settings,  dataBroker : { id: dataBrokerId}, group, user} = relationship;
    const source = await createEndPointForNewDataBroker(sourceFromDb, sourceCredentials, { user });
    const target = await createEndPointForNewDataBroker(targetFromDb, targetCredentials, { user });
    const message = {
        relationshipId,
        source,
        target,
        scannerQueue,
        transferrerQueue,
        settings
    };
    await sendMessageToAllDataBrokersInAGroup(group, message, 'copy', requestId);
    return sendMessageToScanner(omitSecretKeys(source), omitSecretKeys(target), relationshipId, scannerQueue, dataBrokerId, requestId);
}

export async function sendDeleteCommand(relationship, requestId, type) {
    const {source: sourceFormDb, target: targetFromDb, group,
        id: relationshipId, scannerQueue, transferrerQueue, user} = relationship;
    const source = await createEndPointForNewDataBroker(sourceFormDb, undefined, { user });
    const target = await createEndPointForNewDataBroker(targetFromDb, undefined, { user });
    const message = {
        relationshipId,
        source,
        target,
        scannerQueue,
        transferrerQueue
    };
    logger.info(`sending ${type} command to relationship ${relationship.id} `);
    return sendMessageToAllDataBrokersInAGroup(group, message, type, requestId);
}


function validatePath(path, protocol) {
    if (path === '/.' || path.endsWith('/.') || path.includes('/./')) {
        throw new BadRequest(`${protocol} path cannot include '.' (current folder). Path inserted is ${path}`);
    }
    if (path === '/..' || path.endsWith('/..') || path.includes('/../')) {
        throw new BadRequest(`${protocol} path cannot include '..' (parent folder). Path inserted is ${path}`);
    }
    if (path.split('/').map(p => p.length).filter(length => length > 255).length) {
        throw new BadRequest(`${protocol} path's contains folders longer than 255 characters. Path inserted is ${path}`);
    }
}

function validateDifferentSourceAndTarget(protocol, relationship) {

    const adjustPath = (exportPath, path) => `${removeFirstSlash(exportPath)}/${addEndSlash(removeFirstSlash(path))}`;

    switch (protocol) {
        case 'nfs':
            {
                if (relationship.source[protocol].host === relationship.target[protocol].host) {
                    const sourceFullPath = adjustPath(relationship.source[protocol].export, relationship.source[protocol].path);
                    const targetFullPath = adjustPath(relationship.target[protocol].export, relationship.target[protocol].path);
                    if (sourceFullPath.startsWith(targetFullPath) || targetFullPath.startsWith(sourceFullPath)) {
                        throw new BadRequest('Circular relationships are not allowed');
                    }
                }
            }
            break;
        case 'cifs': {
            if (relationship.source[protocol].host === relationship.target[protocol].host) {
                const sourceFullPath = adjustPath(relationship.source[protocol].share, relationship.source[protocol].path);
                const targetFullPath = adjustPath(relationship.target[protocol].share, relationship.target[protocol].path);
                if (sourceFullPath.startsWith(targetFullPath) || targetFullPath.startsWith(sourceFullPath)) {
                    throw new BadRequest('Circular relationships are not allowed');
                }
            }
        }
            break;
        case 'local': {
            if (relationship.source.local.path === relationship.target.local.path) {
                throw new BadRequest('Circular relationships are not allowed');
            }
        }
            break;
        case 's3': {
            if (relationship.source.s3.bucket === relationship.target.s3.bucket) {
                const sourceFullPath =  adjustPath(relationship.source[protocol].bucket, relationship.source[protocol].prefix);
                const targetFullPath =  adjustPath(relationship.target[protocol].bucket, relationship.target[protocol].prefix);
                if (sourceFullPath.startsWith(targetFullPath) || targetFullPath.startsWith(sourceFullPath)) {
                    throw new BadRequest('Circular relationships are not allowed');
                }
            }
        }
            break;
        case 'azure': {
            if (relationship.source.azure.storageAccountName === relationship.target.azure.storageAccountName) {
                if (relationship.source.azure.container === relationship.target.azure.container) {
                    const sourceFullPath = adjustPath(relationship.source[protocol].container, relationship.source[protocol].prefix);
                    const targetFullPath = adjustPath(relationship.target[protocol].container, relationship.target[protocol].prefix);
                    if (sourceFullPath.startsWith(targetFullPath) || targetFullPath.startsWith(sourceFullPath)) {
                        throw new BadRequest('Circular relationships are not allowed');
                    }
                }
            }
        }
            break;
        default : {
            logger.debug(`There is no  ${protocol} to ${protocol} relationship`);
        }
    }
}


export function validateS3Tags(s3Tags) {
    const maxNumberOfTags = 5;
    if (s3Tags && s3Tags.length) {
        if (s3Tags.length > maxNumberOfTags) {
            throw new BadRequest(`Maximum of ${maxNumberOfTags} S3 tags may be provided`);
        }
        s3Tags.map(({key, value}) => {
            if (!key || !value) {
                throw new BadRequest('Each Amazon tag should contain both key and value');
            }
            if (!S3_TAG_KEY_REGEXP.test(key)) {
                throw new BadRequest('Amazon tag keys may not start with \'aws:\' and may contain letters, numbers, spaces, or the special characters + - . _ : / @');
            }
            if (!S3_TAG_VALUE_REGEXP.test(value)) {
                throw new BadRequest('Amazon tag values may not start with \'aws:\' and may contain letters, numbers, spaces, or the special characters + - . _ : / @');
            }
        });
        if (s3Tags.length !== uniq(s3Tags.map(({ key }) => key)).length) {
            throw new BadRequest('S3 tags keys should be distinct (no duplicated allowed)');
        }
    }
}

export function validateGracePeriod(gracePeriod) {
    if (gracePeriod > 48 * 60 * 60) {
        throw new BadRequest('Grace period should be less than 48 hours');
    }
}


function validateEndPoint(endPoint) {
    const {protocol} = endPoint;
    const details = endPoint[protocol];
    switch (protocol) {
        case 'nfs': {
            if (!details) {
                throw new BadRequest('source with protocol nfs should contain object : nfs: {host , export, path}');
            }
            const {host, export: exportPath, path, nfsType } = details;
            assertIpOrHostname(host, 'nfs');
            if (!exportPath && nfsType !== 'efs') {
                throw new BadRequest('source or target with protocol nfs must contain export');
            }
            if (path) {
                validatePath(path, 'nfs');
            }

            break;
        }
        case 'cifs' : {
            if (!details) {
                throw new BadRequest('source with protocol cifs should contain object : cifs: {host , share, path}');
            }
            const {host, share, path } = details;
            assertIpOrHostname(host, 'cifs');
            if (!share) {
                throw new BadRequest('Source or target with protocol nfs must contain export');
            }
            if (path) {
                validatePath(path, 'cifs');
            }
            break;
        }
        case 'sgws': {
            if (!details) {
                throw new BadRequest('source with protocol sgws should contain object : sgws: {host , bucket, prefix}');
            }
            const {host, port, bucket, prefix} = details;
            assertIpOrHostname(host, 'sgws');
            if (!bucket) {
                throw new BadRequest('source or target with protocol sgws must contain bucket');
            }
            if (port) {
                const portNum = Number.parseInt(port);
                if (portNum < 0 || portNum > 165535) {
                    throw BadRequest('SGW port should be between 0 and 165535');
                }
            }
            if (prefix && prefix.length > 1024) {
                throw new BadRequest('SGWS prefix maximum characters length can be 1024');
            }
            break;
        }
        case 's3': {
            if (!details) {
                throw new BadRequest('source with protocol s3 should contain object : s3: { bucket, prefix}');
            }
            const {prefix, bucket, encryption, tags} = details;
            if (!bucket) {
                throw new BadRequest('source or target with protocol s3 must contain bucket');
            }
            if (prefix && prefix.length > 1024) {
                throw new BadRequest('S3 prefix maximum characters length can be 1024');
            }
            if (encryption) {
                if (!['kms'].includes(encryption.protocol)) {
                    throw new BadRequest('S3 encryption protocol should be kms');
                }
                if (encryption.protocol === 'kms' && !encryption.keyId) {
                    throw new BadRequest('If S3 encryption protocol is kms you should provide keyId');
                }
            }
            if (tags) {
                validateS3Tags(tags);
            }
            break;
        }
        case 'azure': {
            const {container, storageAccountName} = details;
            //storageAccountName validation
            if (storageAccountName.length < 3 || storageAccountName.length > 24) {
                //include alphanumeric
                throw new BadRequest('Azure storageAccount name should have between 3 to 24 characters');
            }
            const storageAccountNamePattern = /[^a-z0-9]/;
            if (storageAccountName.match(storageAccountNamePattern)) {
                throw new BadRequest('Azure storageAccount name can contain only lowercase alphanumeric characters');
            }
            //container validation
            //should include dash and alphanumeric
            if (container.length < 3 || container.length > 63) {
                throw new BadRequest('Azure container name should have between 3 to 63 characters');
            }
            const containerPattern = /[^a-z\-0-9]/;
            if (container.match(containerPattern)) {
                throw new BadRequest('Azure container name can contain only lowercase alphanumeric and dash characters');
            }
            break;
        }
        case 'local': {
            if (!details) {
                throw new BadRequest('source with protocol local should contain object : local: { path }');
            }
            //TODO add validations
            break;
        }
        case 'gcp': {
            if (!details) {
                throw new BadRequest('source with protocol gcp should contain object : gcp: {storage, bucket, prefix}');
            }
            const {bucket, prefix} = details;
            if (!bucket) {
                throw new BadRequest('source or target with protocol gcp must contain bucket');
            }
            if (prefix && prefix.length > 1024) {
                throw new BadRequest('gcp prefix maximum characters length can be 1024');
            }
            break;
        }
        default:
            throw new BadRequest(`Protocol ${protocol} is not allowed`);
    }
}

export function validateSyncSchedule(sync) {
    const { inDays, inHours, inMinutes, autoSync } = sync;
    if (autoSync === 'ON' && inDays === 0 && inHours === 0 && inMinutes === 0) {
        throw new BadRequest('Please provide one of \\\'syncInHours\\\' or \\\'syncInDays\\\' or \\\'syncInMinutes \\\' for sync interval\' when pause is false');
    }
    if (inHours > 0 && inDays > 0 && inMinutes > 0) {
        throw new BadRequest('Only one of \'syncInDays\' or \'syncInHours\' or \'syncInMinutes\' can be provided for sync interval');
    }
    if (inHours > 0 && inDays > 0) {
        throw new BadRequest('Only one of \'syncInDays\' or \'syncInHours\' can be provided for sync interval');
    }
    if (inHours > 0 && inMinutes > 0) {
        throw new BadRequest('Only one of \'syncInHours\' or \'syncInMinutes\' can be provided for sync interval');
    }
    if (inDays > 0 && inMinutes > 0) {
        throw new BadRequest('Only one of  \'syncInDays\' or \'syncInMinutes\' can be provided for sync interval');
    }
    if (inDays !== 0 && (inDays < 1 || inDays > 365)) {
        throw new BadRequest('\'syncInDays\' interval must be between 1 and 365 or 0');
    }
    if (inHours !== 0 && (inHours < 1 || inHours > 24)) {
        throw new BadRequest('\'syncInHours\' interval must be between 1 and 24 or 0');
    }
    if (inMinutes !== 0 && (inMinutes < 1 || inMinutes > 60)) {
        throw new BadRequest('\'syncInMinutes\' interval must be between 1 and 60 or 0');
    }
}

export function validateFileTypes(fileTypes) {
    if (!(fileTypes.files || fileTypes.directories || fileTypes.symlinks)) {
        throw new BadRequest('Please select at least one of the following file types: files, directories, symlinks.');
    }
}

export function validateRelationship(relationship) {
    const {source, target, settings: {gracePeriod, fileTypes}, dataBroker, sync} = relationship;
    const newDAtaBroker = isNewDataBroker(dataBroker); //group must be with newDataBroker

    if (!isObject(source)) {
        throw new BadRequest('Source parameter should be an object');
    }

    if (!isObject(target)) {
        throw new BadRequest('Target parameter should be an object');
    }

    const {protocol: sourceProtocol} = source;
    const {protocol: targetProtocol} = target;

    if (newDAtaBroker) {
        if (!Object.keys(ALLOWED_NEW_DATABROKER_SOURCE_AND_TARGET).includes(sourceProtocol)) {
            throw new BadRequest(`Source type can only be one of: ${Object.keys(ALLOWED_NEW_DATABROKER_SOURCE_AND_TARGET).join(', ')}`);
        }
        if (!ALLOWED_NEW_DATABROKER_SOURCE_AND_TARGET[sourceProtocol].includes(targetProtocol)) {
            throw new BadRequest(`Target type can only be one of: ${ALLOWED_NEW_DATABROKER_SOURCE_AND_TARGET[sourceProtocol].join(', ')}`);
        }
    } else {
        if (!Object.keys(ALLOWED_SOURCE_AND_TARGET).includes(sourceProtocol)) {
            throw new BadRequest(`Source type can only be one of: ${Object.keys(ALLOWED_SOURCE_AND_TARGET).join(', ')}`);
        }
        if (!ALLOWED_SOURCE_AND_TARGET[sourceProtocol].includes(targetProtocol)) {
            throw new BadRequest(`Target type can only be one of: ${ALLOWED_SOURCE_AND_TARGET[sourceProtocol].join(', ')}`);
        }
    }
    validateEndPoint(source);
    validateEndPoint(target);
    const circularProtocols = ['nfs', 'cifs', 's3', 'azure'];
    if (sourceProtocol === targetProtocol && circularProtocols.includes(sourceProtocol)) {
        validateDifferentSourceAndTarget(sourceProtocol, relationship);
    }

    validateGracePeriod(gracePeriod);
    validateSyncSchedule(sync);
    validateFileTypes(fileTypes);
}

