import express from 'express';
import format from 'string-template';
import fs from 'fs';
import path from 'path';
import {assignIn} from 'lodash';
import DataBroker from '../models/data-broker';
import Relationship from '../models/relationship';
import Message from '../models/message';
import Group from '../models/group';
import {
    SERVER_HOST,
    DATA_BROKER_CF_FOLDER,
    DFIO_METADATA_BUCKET_REGION,
    DFIO_METADATA_BUCKET_NAME,
    CLOUD_FORMATION_ALL_REGIONS_LINK,
    MONGO_DB_ERROR_DUPLICATE_KEY,
    MESSAGE_STATUS_DONE,
    NEW_DATA_BROKER_INSTALLER_AWS_PATH,
    NEW_DATA_BROKER_AWS_CFT,
    NEW_DATA_BROKER_AZURE_CFN_PATH,
    API_PATH_DATA_BROKERS,
    NEW_DATA_BROKER_GCP_CFT
} from '../utils/consts';
import {
    getUserAndPolicyName,
    assertIpOrHostname,
    isNewDataBroker,
    findRelationshipsWithSecretsInAGroup} from '../utils/common';
import {BadRequest, NotFound, PreconditionFailed} from '../utils/errors';
import {timeout} from '../utils/utils';
import {putObject, getS3ObjectUrl} from '../lib/s3';
import {
    createQueue,
    deleteQueue,
    getQueueArn,
    setQueuePolicy
} from '../lib/sqs';
import {
    createUserAsync,
    createAccessKey,
    deleteKeys,
    putUserPolicyAsync,
    deleteUserPolicy,
    deleteUser,
    createPolicyDocument, getUserArn, waitForUserExists
} from '../lib/iam';
import {pollQueue} from '../utils/sqs-listener';
import {setHubSpotContactFirstDataBrokerSinceIfNeeded} from '../lib/hubspot';
import {
    sendListBucketsFolders, sendListBucketsCommand, sendListExportsFoldersCommand, sendListExportsCommand,
    sendListAzureBlobStorageContainersFoldersCommand, sendListAzureBlobStorageContainers, sendShutdownCommand,
    sendSharesCommand, sendShareFoldersCommand, sendListSgwsBucketsCommand, sendListSgwsBucketsFoldersCommand,
    sendListGcpBucketsCommand, sendListGcpBucketFoldersCommand, updateRelationshipQueuesWithDataBrokerUser
} from '../operations/data-broker-operations';
import {getDataBrokerLatestVersionAsync} from '../utils/update-data-broker';

import {getLogger} from '../utils/logger';

const logger = getLogger();

const router = express.Router();

const [environment] = SERVER_HOST.split('.');


function createCFN(installerLocation, type) {
    let template;
    if (type === 'AZURE') {
        template = fs.readFileSync(path.resolve(__dirname, NEW_DATA_BROKER_AZURE_CFN_PATH), 'utf-8');
    } else if (type === 'GCP') {
        template = fs.readFileSync(path.resolve(__dirname, NEW_DATA_BROKER_GCP_CFT), 'utf-8');
    } else {
        template = fs.readFileSync(path.resolve(__dirname, NEW_DATA_BROKER_AWS_CFT), 'utf-8');
    }
    return format(template, {
        installerLocation
    });
}

function createInstaller(dataBrokerId, agentBundleUrl, commandsQueue, statusesQueue, type, cloudSyncAwsAccessKey, cloudSyncAwsSecretKey, port) {
    const installer = fs.readFileSync(path.resolve(__dirname, NEW_DATA_BROKER_INSTALLER_AWS_PATH), 'utf-8');
    const installerParameters = {
        environment,
        agentBundleUrl,
        dataBrokerId,
        type,
        port,
        commandsQueue,
        statusesQueue,
        sqsAccessKeyId: cloudSyncAwsAccessKey,
        sqsSecretAccessKey: cloudSyncAwsSecretKey
    };

    return format(installer, installerParameters);
}

async function createDataBrokerQueue(dataBroker, userName, queueName, actions) {
    const {id: dataBrokerId, user: {id: userId}} = dataBroker;
    logger.info(`Create queue ${queueName} for user ${userId}`);

    try {
        const [queueUrl, userArn] = await Promise.all([createQueue(queueName), getUserArn(userName)]);
        const queueArn = await getQueueArn(queueUrl);
        await timeout(15000);
        const policy = {
            Version: '2012-10-17',
            Statement: [
                {
                    Sid: `Sid${Date.now()}`,
                    Effect: 'Allow',
                    Principal: {'AWS': [`${userArn}`]},
                    Action: actions,
                    Resource: `${queueArn}`
                }]
        };
        await setQueuePolicy(queueUrl, policy);
        logger.info('Queue successfully created:', queueUrl);

        return {queueUrl, queueArn};
    } catch (err) {
        logger.error(`Failed to create a queue ${queueName} for data broker ${dataBrokerId} and user ${userId} with error`, err);
        throw err;
    }
}

async function deleteQueues(dataBroker) {
    logger.info(`Deleting SQS queues of data broker ${dataBroker.id}`);

    const {statusesQueueUrl, commandsQueueUrl} = dataBroker;
    const promises = [commandsQueueUrl, statusesQueueUrl].map(async queueUrl => {
        try {
            logger.info(`Deleting queue ${queueUrl}`);
            await deleteQueue(queueUrl);
            logger.info('Done deleting queues');
        } catch (err) {
            logger.warn(`Failed to delete queue ${queueUrl}`, err);
        }
    });
    await Promise.all(promises);
}

async function issueDataBrokerCommandAndRespond(req, res, next, fn) {
    const {params: {id: dataBrokerId}} = req;
    const dataBroker = await DataBroker.findById(dataBrokerId).exec();
    if (!dataBroker) {
        logger.warn(`Data Broker with id ${dataBrokerId} not found`);
        next(new NotFound(`Data Broker with id ${dataBrokerId} not found`));
    } else {
        await fn(dataBroker, isNewDataBroker(dataBroker));
        res.status(202).json();
    }
}

async function attachPolicyToUserAndGenerateKeys(userAndPolicyName, commandQueueDetails, statusesQueueDetails) {
    const policyDocument = await createPolicyDocument(commandQueueDetails.queueArn, statusesQueueDetails.queueArn);
    await putUserPolicyAsync(policyDocument, userAndPolicyName, userAndPolicyName);
    return createAccessKey(userAndPolicyName);
}

async function addDataBrokerToAGroup(group, dataBroker) {
    const {groupId, name, user} = group;
    const filter = groupId ? {'_id': groupId} : {name, user};
    return Group.findOneAndUpdate(filter, {$addToSet: {dataBrokers: dataBroker}}, {new: true, upsert: true});
}

async function deleteUserAndPolicyAndQueues(dataBroker) {
    const userAndPolicyName = getUserAndPolicyName(dataBroker);
    try {
        await Promise.all([deleteKeys(userAndPolicyName), deleteUserPolicy(userAndPolicyName, userAndPolicyName)]);
        await deleteUser(userAndPolicyName);
        await deleteQueues(dataBroker);
    } catch (err) {
        logger.error(`An Error occurred while deleting data broker ${dataBroker.name} with id ${dataBroker._id} : ${err}`);
    }
}

async function removeDataBrokerFromGroupIfNeeded(dataBroker, group, databrokersMaxLength) {
    const {name, dataBrokers, user, _id: groupId} = group;
    if (group && dataBrokers.length <= databrokersMaxLength) {
        logger.info(`Removing group with 1 dataBroker ${JSON.stringify(group)}`);
        await Group.findOneAndRemove({user, name}).exec();
    } else {
        logger.info(`Removing dataBroker ${dataBroker.id} from group ${group.id}`);
        await Group.findOneAndUpdate({'_id': groupId}, {$pull: {'dataBrokers': dataBroker.id}});
        await updateRelationshipQueuesWithDataBrokerUser(getUserAndPolicyName(dataBroker), group, true);
    }
}

router.get(API_PATH_DATA_BROKERS, async(req, res, next) => {
    const {user} = req;
    try {
        const dataBrokers = await DataBroker.find({user}).sort({createdAt: 1}).exec();
        res.json(dataBrokers);
    } catch (err) {
        next(err);
    }
});

async function validateADataBrokerCanBeAddedToAGroup(group) {
    const {id: groupId} = group;
    const numberOfRelationships = await findRelationshipsWithSecretsInAGroup(groupId);
    if (numberOfRelationships > 0) {
        logger.info('Validating there is at least one data-broker in the group ');
        const {dataBrokers} = group;
        const aliveDataBroker = dataBrokers.find(dataBroker => dataBroker.isAlive());
        if (!aliveDataBroker) {
            throw new BadRequest('The existing data brokers in this relationship are not online. Start at least one of the data brokers and then try again');
        }
    }
}

async function validateGroupAndAddADataBroker(groupId, user, dataBroker) {
    let group;
    if (groupId) {
        group = await Group.findById(groupId).populate('dataBrokers').exec();
        if (!group) {
            throw new BadRequest(`There is no group with id ${groupId} `);
        }
        //need to check number of relationships that include the following as source or target : azure, cifs, sgws, gcp associated to the group
        // if it has more than one allow to add only if there is an alive data-broker in the group
        await validateADataBrokerCanBeAddedToAGroup(group);
        return addDataBrokerToAGroup({groupId: group.id}, dataBroker);
    } else {
        // no group was asked for this data-broker- we will take the data-broker name and create a new group- will be deleted once the UI have groups
        const name = dataBroker.name;
        return addDataBrokerToAGroup({name, user: user.id}, dataBroker);
    }
}

router.post(API_PATH_DATA_BROKERS, async(req, res, next) => {
    const {user, body: {name, groupId, port, type}} = req;
    const dataBroker = new DataBroker({user, name, type: type || 'AWS'}); // or on type is a temporary patch for OCCM integration

    if (port) {
        dataBroker.encryption = {port: parseInt(port)};
    }

    let group;
    let fileLink;
    try {
        if (!/^[a-zA-Z][a-zA-Z0-9-]{0,127}$/.test(name)) {
            next(new BadRequest('NetApp Data Broker name must contain only letters, numbers, dashes and start with alpha character'));
        } else {
            logger.info('Create a new data broker of type', type, 'to a group', groupId);

            req.timeline = assignIn(req.timeline, {dataBroker: name});
            req.timelineV2 = assignIn(req.timelineV2, {dataBroker: {name, dataBrokerId: dataBroker.id}});
            //first we create the group then assign the data-broker to the group.
            group = await validateGroupAndAddADataBroker(groupId, user, dataBroker);
            //Its like with aggregate and vol( first create the aggregate then creates the volume and adds the volume)
            dataBroker.groupId = group._id;

            req.timelineV2 = assignIn(req.timelineV2, {
                dataBroker: {name, dataBrokerId: dataBroker.id, groupId: group._id, type: type || 'AWS'},
                group: {name: group.name, groupId: group.id}
            });

            //creating user and queues and keys
            const userAndPolicyName = getUserAndPolicyName(dataBroker);
            const timeForUserToCreate = Date.now();
            await createUserAsync(userAndPolicyName);
            await waitForUserExists(userAndPolicyName);

            logger.info(`IAM user for data-broker ${dataBroker.id} was created after ${Date.now() - timeForUserToCreate} milliseconds`);
            const [commandsQueueDetails, statusesQueueDetails] = await Promise.all([
                createDataBrokerQueue(dataBroker, userAndPolicyName, `COMMANDS_${environment}_${dataBroker.id}`, ['SQS:DeleteMessage', 'SQS:ReceiveMessage']),
                createDataBrokerQueue(dataBroker, userAndPolicyName, `STATUSES_${environment}_${dataBroker.id}`, ['SQS:SendMessage']),
            ]);
            dataBroker.statusesQueueUrl = statusesQueueDetails.queueUrl;
            dataBroker.commandsQueueUrl = commandsQueueDetails.queueUrl;

            const keys = await attachPolicyToUserAndGenerateKeys(userAndPolicyName, commandsQueueDetails, statusesQueueDetails);
            const {AccessKeyId, SecretAccessKey} = keys;
            //creating installer
            const cftKeyPrefix = `${DATA_BROKER_CF_FOLDER}/${dataBroker.id}`;
            const installerKey = `${cftKeyPrefix}_installer`;
            const latestDataBrokerBundle = await getDataBrokerLatestVersionAsync();
            const installer = await createInstaller(dataBroker.id, latestDataBrokerBundle.url, commandsQueueDetails.queueUrl, statusesQueueDetails.queueUrl, type,
                AccessKeyId, SecretAccessKey, port);
            await putObject(DFIO_METADATA_BUCKET_REGION, DFIO_METADATA_BUCKET_NAME, installerKey, installer, true);
            const installerUrl = getS3ObjectUrl(DFIO_METADATA_BUCKET_REGION, DFIO_METADATA_BUCKET_NAME, installerKey,
                true, false);
            //creating fileLink
            if (type === 'ONPREM') {
                fileLink = installerUrl;
            } else {
                const cftKey = (type === 'GCP') ? `${cftKeyPrefix}.jinja` : cftKeyPrefix;
                await putObject(DFIO_METADATA_BUCKET_REGION, DFIO_METADATA_BUCKET_NAME, cftKey, createCFN(installerUrl, type), true);
                const templateURL = getS3ObjectUrl(DFIO_METADATA_BUCKET_REGION, DFIO_METADATA_BUCKET_NAME, cftKey,
                    type !== 'AWS', false);
                fileLink = (type === 'AZURE' || type === 'GCP') ? templateURL : `${CLOUD_FORMATION_ALL_REGIONS_LINK}#/stacks/new?stackName=${name}&templateURL=${encodeURIComponent(templateURL)}`;
            }

            await dataBroker.save();
            pollQueue(dataBroker);

            res.json({fileLink, id: dataBroker.id});

            setHubSpotContactFirstDataBrokerSinceIfNeeded(user);
            updateRelationshipQueuesWithDataBrokerUser(userAndPolicyName, group);
        }
    } catch (error) {
        let errorToThrow = error;
        if (error.code === MONGO_DB_ERROR_DUPLICATE_KEY) {
            logger.info(`DEBUG User already has a Data Broker named '${name}`);
            errorToThrow = new BadRequest(`NetApp Data Broker named '${name}' already exist`);
            if (group) {
                //Assumption- if the group has 1 data broker it is the data broker that was not created and it should be deleted
                await removeDataBrokerFromGroupIfNeeded(dataBroker, group, 1);
            }
        } else {
            const removedDataBroker = await DataBroker.findOne({
                user,
                name,
                group: (group) ? group.name : undefined
            }).exec();
            if (removedDataBroker) {
                await deleteQueues(removedDataBroker);
                await removedDataBroker.deletOne();
            }
        }
        if (group) {
            //Assumption- if the group has 1 data broker it is the data broker that was not created and it should be deleted
            await removeDataBrokerFromGroupIfNeeded(dataBroker, group, 1);
        }
        next(errorToThrow);
    }
});

async function validateDataBrokerDeletion(user, dataBrokers, group, dataBroker) {
    const filter = (group) ? {group: group.id} : {dataBroker: dataBroker.id};
    const numberOfRelationships = await Relationship.countDocuments(filter);
    if (numberOfRelationships > 0) {
        const anotherAliveDataBroker = dataBrokers.find(recievedDataBroker => recievedDataBroker.id !== dataBroker.id);
        if (!anotherAliveDataBroker) {
            //if we have no group we wouldn't find any other alive data-broker;
            throw new PreconditionFailed(`NetApp Data Broker '${dataBroker.name}' in use by ${numberOfRelationships} active relationships `);
        }
    }
}


router.delete(`${API_PATH_DATA_BROKERS}/:id`, async(req, res, next) => {
    const {user, params: {id: dataBrokerId}, id: requestId} = req;
    try {
        const dataBroker = await DataBroker.findById(dataBrokerId).populate('groupId').populate({
            path: 'groupId',
            populate: {path: 'dataBrokers'}
        }).exec();

        const {name, groupId: group} = dataBroker || {};

        //update timeline
        req.timeline = assignIn(req.timeline, {dataBroker: name});
        req.timelineV2 = assignIn(req.timelineV2, {
            dataBroker: {name, dataBrokerId, groupId: group.id},
            group: {groupId: group.id, name: group.name}
        });

        if (!dataBroker) {
            logger.warn(`Data Broker with id ${dataBrokerId} was not found for deletion`);
            next(new NotFound());
        } else {
            const {dataBrokers} = group || [dataBroker];
            await validateDataBrokerDeletion(user, dataBrokers, group, dataBroker);
            if (group) {
                const anotherDataBrokerFromTheGroup = dataBrokers.find(dataBroker => dataBroker.id !== dataBrokerId);
                await Relationship.updateOne({
                    dataBroker: dataBrokerId,
                    group: group.id
                }, {$set: {dataBroker: anotherDataBrokerFromTheGroup}}, {multi: true}).exec();
                await removeDataBrokerFromGroupIfNeeded(dataBroker, group, 1);
            }
            logger.info(`Sending shutdown command to databroker ${dataBroker._id} of user ${user._id}`);
            await sendShutdownCommand(requestId, dataBroker);
            logger.info(`removing data broker with id ${dataBrokerId} from DB `);
            await dataBroker.deleteOne();
            res.json();
            await timeout(30000);
            deleteUserAndPolicyAndQueues(dataBroker);
        }
    } catch (err) {
        next(err);
    }
});

router.get(`${API_PATH_DATA_BROKERS}/:id/list-cifs-shares`, (req, res, next) => {
    const {host, version} = req.query;
    assertIpOrHostname(host, 'cifs');
    issueDataBrokerCommandAndRespond(req, res, next, (dataBroker, isNewDataBroker) => {
        const {username, password, domain} = req.headers;
        if (isNewDataBroker) {
            return sendSharesCommand(req.id, dataBroker, host, version, username, password, domain);
        } else {
            return dataBroker.listCifsSharesV2(req.id, host, username, password, domain);
        }
    });
});

router.get(`${API_PATH_DATA_BROKERS}/:id/list-cifs-share-folders`, (req, res, next) => {
    const {host, share, path, version} = req.query;
    assertIpOrHostname(host, 'cifs');
    const {username, password, domain} = req.headers;
    issueDataBrokerCommandAndRespond(req, res, next, (dataBroker, isNewDataBroker) => {
        if (isNewDataBroker) {
            return sendShareFoldersCommand(req.id, dataBroker, host, share, path, version, username, password, domain);
        } else {
            return dataBroker.listCifsShareFoldersV2(req.id, host, username, password, domain, share, path);
        }
    });
});

router.get(`${API_PATH_DATA_BROKERS}/:id/list-nfs-exports`, (req, res, next) => {
    const {params: {id: dataBrokerId}, user: {id: userId}, query: {host, isEFS}, id: requestId} = req;
    assertIpOrHostname(host, 'nfs');
    if (isEFS) {
        return Message.SendAsync(dataBrokerId, userId, requestId, 'list-nfs-exports', {
            host,
            status: MESSAGE_STATUS_DONE,
            paths: ['/']
        })
            .then(() => res.status(202).json());
    } else {
        issueDataBrokerCommandAndRespond(req, res, next, (dataBroker, isNewDataBroker) => {
            if (isNewDataBroker) {
                return sendListExportsCommand(requestId, host, dataBroker);
            } else {
                return dataBroker.listNfsExportsV2(req.id, host);
            }
        });
    }
});

router.get(`${API_PATH_DATA_BROKERS}/:id/list-nfs-export-folders`, (req, res, next) => {
    try {
        const {exportName, host, path, version} = req.query;
        assertIpOrHostname(host, 'nfs');
        issueDataBrokerCommandAndRespond(req, res, next, (dataBroker, isNewDataBroker) => {
            const {id: requestId} = req;
            if (isNewDataBroker) {
                return sendListExportsFoldersCommand(requestId, host, exportName, path, version, dataBroker);
            } else {
                return dataBroker.listNfsExportFoldersV2(requestId, host, exportName, path);
            }
        });
    } catch (e) {
        next(e);
    }
});

router.get(`${API_PATH_DATA_BROKERS}/:id/list-s3-buckets`, (req, res, next) => {
    issueDataBrokerCommandAndRespond(req, res, next, (dataBroker, isNewDataBroker) => {
        const {id: requestId} = req;
        if (isNewDataBroker) {
            return sendListBucketsCommand(requestId, dataBroker);
        } else {
            return dataBroker.listS3BucketsV2(requestId);
        }
    });
});

router.get(`${API_PATH_DATA_BROKERS}/:id/list-s3-bucket-folders`, (req, res, next) => {
    issueDataBrokerCommandAndRespond(req, res, next, (dataBroker, isNewDataBroker) => {
        const {query: {bucket, prefix}, id: requestId} = req;
        if (isNewDataBroker) {
            return sendListBucketsFolders(requestId, bucket, prefix, dataBroker);
        } else {
            return dataBroker.listS3BucketFoldersV2(requestId, bucket, prefix);
        }
    });
});

router.get(`${API_PATH_DATA_BROKERS}/:id/list-sgws-buckets`, (req, res, next) => {
    const {host, port} = req.query;
    assertIpOrHostname(host, 'sgws');
    issueDataBrokerCommandAndRespond(req, res, next, (dataBroker, isNewDataBroker) => {
        if (isNewDataBroker) {
            return sendListSgwsBucketsCommand(req.id, dataBroker, host, port, req.headers['access-key'], req.headers['secret-key']);
        } else {
            return dataBroker.listSgwsBucketsV2(req.id, host, Number.parseInt(port), req.headers['access-key'], req.headers['secret-key']);
        }
    });
});

router.get(`${API_PATH_DATA_BROKERS}/:id/list-sgws-bucket-folders`, (req, res, next) => {
    const {host, port, bucket, prefix} = req.query;
    assertIpOrHostname(host, 'sgws');
    issueDataBrokerCommandAndRespond(req, res, next, (dataBroker, isNewDataBroker) => {
        if (isNewDataBroker) {
            return sendListSgwsBucketsFoldersCommand(req.id, dataBroker, host, port, req.headers['access-key'], req.headers['secret-key'], bucket, prefix);
        } else {
            return dataBroker.listSgwsBucketFoldersV2(req.id, host, Number.parseInt(port), req.headers['access-key'], req.headers['secret-key'], bucket, prefix);
        }
    });
});

router.get(`${API_PATH_DATA_BROKERS}/:id/list-azure-blob-storage-containers`, (req, res, next) => {
    issueDataBrokerCommandAndRespond(req, res, next, (dataBroker, isNewDataBroker) => {
        const {id: requestId} = req;
        const storageAccountName = req.query['storage-account-name'];
        const storageAccountConnectionString = req.headers['storage-account-connection-string'];
        if (isNewDataBroker) {
            return sendListAzureBlobStorageContainers(requestId, dataBroker, storageAccountName, storageAccountConnectionString);
        } else {
            throw new BadRequest('This action is not supported with this type of data broker');
        }
    });
});

router.get(`${API_PATH_DATA_BROKERS}/:id/list-azure-blob-storage-container-folders`, (req, res, next) => {
    issueDataBrokerCommandAndRespond(req, res, next, async(dataBroker, isNewDataBroker) => {
        const {id: requestId, query: {container, prefix}} = req;
        const storageAccountName = req.query['storage-account-name'];
        const storageAccountConnectionString = req.headers['storage-account-connection-string'];
        if (isNewDataBroker) {
            return sendListAzureBlobStorageContainersFoldersCommand(requestId, dataBroker, container, prefix, storageAccountName, storageAccountConnectionString);
        } else {
            next(new BadRequest('This action is not supported with this type of data broker'));
        }
    });
});

router.get(`${API_PATH_DATA_BROKERS}/:id/list-gcp-buckets`, (req, res, next) => {
    issueDataBrokerCommandAndRespond(req, res, next, (dataBroker, isNewDataBroker) => {
        const {id: requestId} = req;
        const projectId = req.query['project-id'];
        const clientEmail = req.headers['client-email'];
        const privateKey = req.headers['private-key'].replace(/\\n/g, '\n');
        if (isNewDataBroker) {
            return sendListGcpBucketsCommand(requestId, dataBroker, projectId, clientEmail, privateKey);
        } else {
            throw new BadRequest('This action is not supported with this type of data broker');
        }
    });
});

router.get(`${API_PATH_DATA_BROKERS}/:id/list-gcp-bucket-folders`, (req, res, next) => {
    issueDataBrokerCommandAndRespond(req, res, next, (dataBroker, isNewDataBroker) => {
        const {id: requestId, query: {bucket, prefix}} = req;
        const projectId = req.query['project-id'];
        const clientEmail = req.headers['client-email'];
        const privateKey = req.headers['private-key'].replace(/\\n/g, '\n');
        if (isNewDataBroker) {
            return sendListGcpBucketFoldersCommand(requestId, dataBroker, bucket, prefix, projectId, clientEmail, privateKey);
        } else {
            throw new BadRequest('This action is not supported with this type of data broker');
        }
    });
});

export default router;
