import express from 'express';
import Group from '../models/group';
import Message from '../models/message';
import { assertIpOrHostname } from '../utils/common';
import { NotFound, BadRequest } from '../utils/errors';
import {sendListBucketsCommand, sendListExportsCommand, sendListBucketsFolders, sendListExportsFoldersCommand,
    sendListAzureBlobStorageContainersFoldersCommand, sendListAzureBlobStorageContainers} from '../operations/data-broker-operations';
import {MESSAGE_STATUS_DONE, API_PATH_GROUPS} from '../utils/consts';

const router = express.Router();

import {getLogger} from '../utils/logger';
const logger = getLogger();

async function findAliveDataBrokerInAGroup(group) {
    const {dataBrokers} = group;
    const aliveDataBroker = dataBrokers.find(dataBroker => dataBroker.isAlive());
    if (!aliveDataBroker) {
        throw new BadRequest('There is no alive dataBroker inside the group');
    }
    return aliveDataBroker;
}


async function issueAliveDataBrokerCommandAndRespond(req, res, next, fn) {
    const { params : { id: groupId }, user} = req;
    try {
        const group = await Group.findOne({_id: groupId, user}).populate('dataBrokers').exec();
        if (!group) {
            logger.warn(`Group with id ${groupId} not found`);
            next(new NotFound(`Group with id ${groupId} not found`));
        } else {
            const aliveDataBroker = await findAliveDataBrokerInAGroup(group);
            await fn(aliveDataBroker);
            res.status(202).json();
        }
    } catch (e) {
        next(e);
    }
}


router.get(API_PATH_GROUPS, async(req, res, next) => {
    const { user } = req;
    try {
        const groups = await Group.find({user}).sort({createdAt: 1}).populate('dataBrokers').exec();
        res.json(groups);
    } catch (e) {
        next(e);
    }
});


router.get(`${API_PATH_GROUPS}/:id/list-nfs-exports`, async(req, res, next) => {
    const { params : { id: groupId }, user : {id: userId}, query : {host, isEFS}, id: requestId} = req;
    assertIpOrHostname(host, 'nfs');
    if (isEFS) {
        const group = await Group.findById(groupId).populate('dataBrokers').exec();
        if (!group) {
            logger.warn(`Group with id ${groupId} not found`);
            next(new NotFound(`Group with id ${groupId} not found`));
        } else {
            const aliveDataBroker = await findAliveDataBrokerInAGroup(group);
            await Message.SendAsync(aliveDataBroker.id, userId, requestId, 'list-nfs-exports', {
                host,
                status: MESSAGE_STATUS_DONE,
                paths: ['/']
            });
            res.status(202).json();
        }
    } else {
        issueAliveDataBrokerCommandAndRespond(req, res, next, dataBroker => {
            return sendListExportsCommand(requestId, host, dataBroker);
        });
    }
});

router.get(`${API_PATH_GROUPS}/:id/list-nfs-export-folders`, (req, res, next) => {
    try {
        const {exportName, host, path, version} = req.query;
        assertIpOrHostname(host, 'nfs');
        issueAliveDataBrokerCommandAndRespond(req, res, next, dataBroker => {
            const {id: requestId} = req;
            return sendListExportsFoldersCommand(requestId, host, exportName, path, version, dataBroker);
        });
    } catch (e) {
        next(e);
    }
});

router.get(`${API_PATH_GROUPS}/:id/list-s3-buckets`, (req, res, next) => {
    issueAliveDataBrokerCommandAndRespond(req, res, next, dataBroker => {
        const { id: requestId } = req;
        return sendListBucketsCommand(requestId, dataBroker);
    });
});

router.get(`${API_PATH_GROUPS}/:id/list-s3-bucket-folders`, (req, res, next) => {
    issueAliveDataBrokerCommandAndRespond(req, res, next, dataBroker => {
        const {query: {bucket, prefix}, id: requestId} = req;
        return sendListBucketsFolders(requestId, bucket, prefix, dataBroker);
    });
});

router.get(`${API_PATH_GROUPS}/:id/list-azure-blob-storage-containers`, (req, res, next) => {
    issueAliveDataBrokerCommandAndRespond(req, res, next, dataBroker => {
        const {id: requestId} = req;
        const storageAccountName = req.query['storage-account-name'];
        const storageAccountConnectionString = req.headers['storage-account-connection-string'];
        return sendListAzureBlobStorageContainers(requestId, dataBroker, storageAccountName, storageAccountConnectionString);
    });
});

router.get(`${API_PATH_GROUPS}/:id/list-azure-blob-storage-container-folders`, (req, res, next) => {
    issueAliveDataBrokerCommandAndRespond(req, res, next, dataBroker => {
        const {id: requestId, query : {container, prefix}} = req;
        const storageAccountName = req.query['storage-account-name'];
        const storageAccountConnectionString = req.headers['storage-account-connection-string'];
        return sendListAzureBlobStorageContainersFoldersCommand(requestId, dataBroker, container, prefix, storageAccountName,
            storageAccountConnectionString);
    });
});


export default router;
