import moment from 'moment';
import express from 'express';
import {get, extend, isNil} from 'lodash';
import agenda from '../utils/agenda';
import DataBroker from '../models/data-broker';
import Timeline from '../models/timeline';
import TimelineV2 from '../models/timeline-v2';
import Group from '../models/group';
import Relationship from '../models/relationship';
import RelationshipStats from '../models/relationship-statistics';
import User from '../models/user';
import {
    getEmptyRelationship, MONGO_DB_ERROR_DUPLICATE_KEY, AGENDA_TASK_NAME_SYNC_SCHEDULER, MESSAGE_STATUS_FAILED,
    FAILED_AFTER_STARTUP_MESSAGE, SERVER_HOST, RUNNING_STATUS, MESSAGE_STATUS_ABORTED,
    ABORTED_MESSAGE, API_PATH_RELATIONSHIP, API_PATH_RELATIONSHIP_V2
} from '../utils/consts';
import {BadRequest, NotFound, PreconditionFailed} from '../utils/errors';
import {
    isRelationshipRunning, createStringFromEndPointObject, isStartupRunning, calculateStatus, isRelationshipAborting,
    updateTimeLine, getUserAndPolicyName, updateTimeLineV2,
    getRelationshipResponse, updateTimeLineNewDataBroker, findTimelineRecordV2, findTimelineRecord,
    calculateRelationshipStatusV2, updateStatusForRelationship, isNewDataBroker, updateRelationshipAndTimeLine,
    handleEndOfRelationship
} from '../utils/common';
import {requestId} from '../utils/utils';
import {createQueue, getQueueArn, setQueuePolicy} from '../lib/sqs';
import {inContext} from '../middlewares/continuation-local-storage';
import {setHubSpotContactFirstRelationshipSinceIfNeeded} from '../lib/hubspot';
import updatePortalLambda from '../lib/lambda';
import {getUserArn} from '../lib/iam';
import {
    sendCopyCommand,
    validateRelationship,
    validateSyncSchedule,
    validateGracePeriod,
    validateFileTypes
} from '../operations/relationship-operations';

import {getLogger} from '../utils/logger';

const logger = getLogger();

const [environment] = SERVER_HOST.split('.');

const router = express.Router();

function calculateNextSync(inMinutes, inHours, inDays) {
    return moment().add(inMinutes, 'minutes').add(inDays, 'days').add(inHours, 'hours');
}

function createTimeLineObject(req, relationship) {
    const {source, target, dataBroker, group, id: relationshipId, isQstack, isCvo} = relationship || getEmptyRelationship();

    const timeLineSource = createStringFromEndPointObject(source);
    const timeLineTarget = createStringFromEndPointObject(target);

    req.timeline = extend(req.timeline, {
        source: timeLineSource,
        target: timeLineTarget,
        dataBroker: dataBroker.name
    });
    req.timelineV2 = extend(req.timelineV2, {
        relationship: (relationship) ? {
            relationshipId,
            source,
            target,
            isQstack,
            isCvo,
            dataBrokerId: dataBroker.id,
            group: (group) ? group.id : undefined
        } : undefined,
        dataBroker: (group) ? undefined : ((dataBroker === {}) ? {
            name: dataBroker.name,
            dataBrokerId: dataBroker.id
        } : undefined),
        group: (group) ? {name: group.name, groupId: group.id} : undefined
    });
}

function getNextSyncTime(lastSyncTime, inDays, inHours, inMinutes) {
    let nextSync = moment(lastSyncTime);
    while (nextSync.isBefore()) {
        nextSync = nextSync.add(inMinutes, 'minutes').add(inDays, 'days').add(inHours, 'hours');
    }
    return nextSync.toDate();
}

function verifyOnlyOne(relationship) {
    const {source: {protocol: sourceProtocol}, target: {protocol: targetProtocol}, user, dataBroker, group} = relationship;
    const newRelationshipSource = createStringFromEndPointObject(relationship.source);
    const newRelationshipTarget = createStringFromEndPointObject(relationship.target);

    //TODO when deleting dataBroker from relationship and using only groups - need to look for group and not dataBroker

    const query = {'source.protocol': sourceProtocol, 'target.protocol': targetProtocol, user};
    if (group) {
        query.group = group;
    } else {
        query.dataBroker = dataBroker;
    }

    return Relationship.find(query).exec()
        .then(relationships => {
            relationships.forEach(relationship => {
                const source = createStringFromEndPointObject(relationship.source);
                const target = createStringFromEndPointObject(relationship.target);
                if (source === newRelationshipSource && target === newRelationshipTarget) {
                    throw new BadRequest('Sync relationship already exists', MONGO_DB_ERROR_DUPLICATE_KEY);
                }
            });
            return relationship;
        });
}

/*
 >>>>>>>>>>>>>>>>>Phoenix>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  */

async function createRelationshipQueue(queueType, usersName, relationshipId) {
    const userArns = await Promise.all(usersName.map(userName => getUserArn(userName)));
    const queueUrl = await createQueue(`CS_${environment}_${queueType}_${relationshipId}_${Date.now()}`);
    const queueArn = await getQueueArn(queueUrl);
    const usersArns = userArns.map(user => `${user}`);
    const policy = {
        'Version': '2012-10-17',
        'Statement': [
            {
                'Sid': `Sid${Date.now()}`,
                'Effect': 'Allow',
                'Principal': {'AWS': usersArns},
                'Action': ['SQS:GetQueueAttributes', 'SQS:DeleteMessage', 'SQS:ReceiveMessage', 'SQS:SendMessage', 'SQS:ChangeMessageVisibility'],
                'Resource': `${queueArn}`
            }]
    };


    await setQueuePolicy(queueUrl, policy);
    return queueUrl;
}

async function createRelationship(relationship, requestId, isSync, sourceCredentials, targetCredentials) {
    const {id: relationshipId, group, dataBroker} = relationship;
    const usersName = (group) ? group.dataBrokers.map(dataBroker => getUserAndPolicyName(dataBroker)) : [getUserAndPolicyName(dataBroker)]; // each phoenix data broker will have a group
    const [scannerQueue, transferrerQueue] = await Promise.all([createRelationshipQueue('scanner', usersName, relationshipId),
        createRelationshipQueue('transferrer', usersName, relationshipId)]);
    relationship.scannerQueue = scannerQueue;
    relationship.transferrerQueue = transferrerQueue;
    relationship.endTime = undefined;
    relationship.startTime = Date.now();
    if (isSync) {
        relationship.phase = 'Sync';
    }

    const updatedRelationship = await relationship.save(); //in order for the relationship to be with the right phase for the simulator;
    await updateStatusForRelationship(updatedRelationship, RUNNING_STATUS, undefined, requestId); //updates timeline if needed - so there will be a timeline to update
    await updateRelationshipAndTimeLine(updatedRelationship, requestId, RUNNING_STATUS);
    await sendCopyCommand(updatedRelationship, requestId, sourceCredentials, targetCredentials);
    return updatedRelationship;
}

async function startNewDataBrokerSync(relationship, requestId, isTimeLine, throwError) {
    const {source, target, dataBroker: {id: dataBrokerId, name}, id: relationshipId, user, isQstack, isCvo, group} = relationship;
    const statusObj = await calculateRelationshipStatusV2(relationshipId, user);
    if (!statusObj || statusObj.status === RUNNING_STATUS) {
        logger.debug(`Cannot run Sync for relationship ${relationshipId} of user ${user.id}: There is a running operation right now`);
        if (throwError) {
            throw new BadRequest('A sync action is running right now, can not run another action');
        }
    } else {
        if (isTimeLine) {
            const sourceTimeLine = createStringFromEndPointObject(source);
            const targetTimeLine = createStringFromEndPointObject(target);
            await new Timeline({
                user,
                requestId: requestId,
                summary: 'Scheduled sync',
                source: sourceTimeLine,
                target: targetTimeLine,
                dataBroker: name
            }).save();
            await new TimelineV2({
                user,
                requestId,
                relationship: {
                    source,
                    target,
                    isQstack,
                    isCvo,
                    relationshipId,
                    group: (group) ? group.id : undefined,
                    dataBrokerId
                },
                dataBroker: {name, dataBrokerId},
                group: (group) ? {name: group.name, groupId: group.id} : undefined,
                summary: 'Scheduled sync'
            }).save();
        }
        await RelationshipStats.deleteMany({relationship: relationshipId}).exec();
        try {
            await createRelationship(relationship, requestId, true);
        } catch (e) {
            //relevant only to sync and scheduled sync- in case create relationship is failed the relationship will be deleted
            // in the catch part of the POST relationship route.
            logger.error(`Failed to sync relationship with id ${relationshipId} error ${e}`);
            //#1 update relationship status with FAILED
            await RelationshipStats.updateMany({'relationship': relationshipId, user:user.id}, {
                status: MESSAGE_STATUS_FAILED,
                reason: e.message || 'sync operation failed please try again'
            }, {
                multi: true,
                upsert: true,
                strict: false
            }).exec();
            //#2 send information to the UI
            await updateRelationshipAndTimeLine(relationship, requestId);
            throw e;
        }
    }
}

async function abortNewDataBroker(relationship, requestId) {
    const {id: relationshipId, user} = relationship;
    //before aborting- calculating status
    const statusObj = await calculateRelationshipStatusV2(relationshipId, user);
    if (statusObj && statusObj.status !== RUNNING_STATUS) {
        logger.debug(`Cannot run Abort for relationship ${relationshipId} of user ${user.id}: The relationship is not running`);
        throw new BadRequest('There is no running Sync right now, can not run abort action');
    }
    await handleEndOfRelationship(relationship, requestId, 'abort');
    await updateStatusForRelationship(relationship, MESSAGE_STATUS_ABORTED, ABORTED_MESSAGE, requestId);
    await updateRelationshipAndTimeLine(relationship, requestId, MESSAGE_STATUS_ABORTED);

}

/*
>>>>>>>>>>>>>>>>>>Asterix>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
 */

async function startSync(relationship, requestId) {
    relationship.messages['sync-relationship'] = {'setup': {'timestamp': new Date().getTime()}};
    relationship.markModified('messages');
    await relationship.save();
    return relationship.startSync(requestId);
}

function startAbort(relationship, requestId) {
    relationship.messages['abort-relationship'] = {'abort': {'timestamp': new Date().getTime()}};
    relationship.markModified('messages');
    return relationship.save()
        .then(() => relationship.startAbort(requestId));
}

async function validateUniqueS3BucketForBroker(dataBroker, source, target) {
    const relationship = new Relationship({source, target});
    if (relationship.isEndPointProtocol('source', 's3')) {
        const relationships = await Relationship.find({
            dataBroker: {$ne: dataBroker},
            'source.s3.bucket': source.s3.bucket
        }).exec();
        if (relationships.length) {
            /* eslint max-len: "off" */
            throw new BadRequest(`The S3 bucket '${relationship.s3Bucket}' is the source in a sync relationship that uses a different data broker. If you want to use an S3 bucket as the source in multiple sync relationships, then you must use the same data broker for each relationship`);
        }
    }
}

async function validateAndStartSync(relationship, requestId, isTimeLine, throwError) {
    const {messages, id: relationshipId, user, source, target, dataBroker: {name, id: dataBrokerId}, isQstack, isCvo, group} = relationship;
    const status = calculateStatus(messages);
    let error;
    if (isRelationshipRunning(status)) {
        logger.warn(`Cannot run Sync for relationship ${relationshipId} of user ${user._id}: There is a running operation right now`);
        error = new BadRequest('A sync action is running right now, can not run another action');
    } else if (isStartupRunning(relationship)) {
        logger.warn(`Cannot run Sync for relationship ${relationshipId}: There is a running startup operation right now`);
        error = new BadRequest('A startup action is running right now, can not run another action, please wait and try again');
    } else if (isRelationshipAborting(status)) {
        logger.warn(`Cannot run Sync for relationship ${relationshipId} of user ${user._id}: The relationship is aborting`);
        error = new BadRequest('There is an abort operation right now, can not run sync action');
    } else if (get(relationship, 'messages[\'create-relationship-v2\'].copy.status') === MESSAGE_STATUS_FAILED.toLowerCase() &&
        get(relationship, 'messages[\'create-relationship-v2\'].copy.reason') === FAILED_AFTER_STARTUP_MESSAGE) {
        logger.warn(`Cannot run sync for relationship ${relationshipId} of user ${user._id} : relationship initial copy was interrupted`);
        error = new BadRequest(`Cannot run sync action - ${FAILED_AFTER_STARTUP_MESSAGE}`);
    } else {
        if (isTimeLine) {
            const sourceTimeLine = createStringFromEndPointObject(source);
            const targetTimeLine = createStringFromEndPointObject(target);
            await new Timeline({
                user,
                requestId: requestId,
                summary: 'Scheduled sync',
                source: sourceTimeLine,
                target: targetTimeLine,
                dataBroker: name
            }).save();
            await new TimelineV2({
                user,
                requestId: requestId,
                summary: 'Scheduled sync',
                dataBroker: {name, dataBrokerId},
                relationship: {
                    relationshipId,
                    source,
                    target,
                    isQstack,
                    isCvo,
                    dataBrokerId,
                    group: (group) ? group : undefined
                }
            }).save();

        }
        return startSync(relationship, requestId);
    }
    if (error && throwError) {
        throw error;
    }
}

async function validateAndStartAbort(relationship, requestId) {
    const {messages, id: relationshipId, user: {id: userId}} = relationship;

    const status = calculateStatus(messages);

    if (!isRelationshipRunning(status)) {
        logger.debug(`Cannot run Abort for relationship ${relationshipId} of user ${userId}: The relationship is not running`);
        throw new BadRequest('There is no running Sync right now, can not run abort action');
    }

    if (isRelationshipAborting(status)) {
        logger.debug(`Cannot run Abort for relationship ${relationshipId} of user ${userId}: The relationship is not aborting`);
        throw new BadRequest('There is an abort operation right now, can not run abort action');
    }

    if (isStartupRunning(relationship)) {
        logger.debug(`Cannot run abort for relationship ${relationshipId}: There is a running startup operation right now`);
        throw new BadRequest('A startup action is running right now, can not run another action, please wait and try again');
    }

    if (relationship.isEndPointProtocol('source', 's3') || relationship.isEndPointProtocol('target', 'sgws')) {
        logger.debug(`Cannot run abort for relationship ${relationshipId}: abort only valid for upload or copy relationships`);
        throw new BadRequest('This is a download relationship, can not run abort action');
    }
    await startAbort(relationship, requestId);
}


router.get(API_PATH_RELATIONSHIP_V2, async(req, res, next) => {
    const {user} = req;
    try {
        const relationships = await Relationship.find({user}).sort({createdAt: 1}).populate('user dataBroker group').populate({
            path: 'group',
            populate: {path: 'dataBrokers'}
        }).exec();
        const newRelationships = await Promise.all(relationships.map(async r => await getRelationshipResponse(r, r.dataBroker)));
        res.json(newRelationships);

    } catch (e) {
        next(e);
    }
});


router.get(`${API_PATH_RELATIONSHIP_V2}/:id`, async(req, res, next) => {
    const {params: {id: relationshipId}} = req;
    try {
        const relationship = await Relationship.findById(relationshipId).populate('user dataBroker group').populate({
            path: 'group',
            populate: {path: 'dataBrokers'}
        }).exec();
        if (!relationship) {
            logger.warn(`Relationship with id ${relationshipId} not found `);
            return (new NotFound(`Relationship with id ${relationshipId} not found`));
        }
        const relationshipResponse = await getRelationshipResponse(relationship, relationship.dataBroker);
        res.json(relationshipResponse);
    } catch (error) {
        next(error);
    }
});

function updateRelationshipFields(endpoint) {
    if (endpoint.s3 && !endpoint.s3.storageClass) {
        endpoint.s3.storageClass = 'STANDARD';
    }
    if (endpoint.azure && !endpoint.azure.blobTier) {
        endpoint.azure.blobTier = 'HOT';
    }
    if (endpoint.nfs && !endpoint.nfs.nfsType) {
        endpoint.nfs.nfsType = 'nfs';
    }
}

function isThereAliveDataBroker(group, dataBroker) {
    const dataBrokers = (group) ? group.dataBrokers : [dataBroker];
    const aliveDataBroker = dataBrokers.find(dataBroker => dataBroker.isAlive());
    if (!aliveDataBroker) {
        throw new PreconditionFailed('Relationship must have at least one connected data-broker');
    }
}

router.post(API_PATH_RELATIONSHIP_V2, async(req, res, next) => {
    let relationship;
    let {isQstack, body: {additionalInfo, settings}} = req;
    const {
        user, id, body: {
            dataBrokerId, source, sourceCredentials, target, targetCredentials, isCvo,
            syncInDays: inDays, syncInHours: inHours, syncInMinutes, groupId
        }
    } = req;
    const inMinutes = syncInMinutes || 0;
    additionalInfo = additionalInfo || {};
    settings = settings || additionalInfo;
    const autoSync = (inDays !== 0 || inHours !== 0 || inMinutes !== 0) ? 'ON' : 'OFF';
    updateRelationshipFields(source);
    updateRelationshipFields(target);

    let dataBroker;
    let group;
    try {
        //update isQstack field in case the source or target includes cvs or anf
        const sourceNfsType = get(source, 'nfs.nfsType');
        const targetNfsType = get(target, 'nfs.nfsType');

        if (['cvs', 'anf'].includes(sourceNfsType) || ['cvs', 'anf'].includes(targetNfsType)) {
            await User.updateOne({portalId: req.user.portalId}, {$set: {isQstack: true}}).exec();
            isQstack = true;
        }

        if (!dataBrokerId && !groupId) {
            return next(new BadRequest('Use either dataBrokerId or groupId but not both'));
        }
        if (dataBrokerId && groupId) {
            return next(new BadRequest('Use either dataBrokerId or groupId but not both'));
        }
        if (dataBrokerId) {
            dataBroker = await DataBroker.findOne({_id: dataBrokerId, user}).exec();
            if (!dataBroker) {
                logger.warn(`Data broker with id ${dataBrokerId} not found`);
                return next(new NotFound(`Data broker with id ${dataBrokerId} not found`));
            } else {
                // if the dataBroker is phoenix
                if (dataBroker.groupId) {
                    group = await Group.findById(dataBroker.groupId).populate('dataBrokers');
                }
            }
        } else if (groupId) {
            group = await Group.findOne({_id: groupId, user}).populate('dataBrokers');

            if (!group) {
                logger.warn(`Group with id ${groupId} not found`);
                return next(new NotFound(`Group with id ${groupId} not found`));
            }
            const {dataBrokers} = group;
            dataBroker = await dataBrokers.find(dataBroker => dataBroker.isAlive());
        }
        // check if alive
        isThereAliveDataBroker(group, dataBroker);

        relationship = new Relationship({
            user,
            isQstack,
            isCvo,
            source,
            target,
            sync: {inDays, inHours, inMinutes, autoSync},
            settings, dataBroker, group
        });
        const newDataBroker = isNewDataBroker(dataBroker);

        createTimeLineObject(req, relationship);

        validateRelationship(relationship);
        relationship.sync.nextTime = calculateNextSync(inMinutes, inHours, inDays);
        await verifyOnlyOne(relationship);
        //start copy according to the dataBroker type
        if (newDataBroker) {
            relationship = await createRelationship(relationship, id, false, sourceCredentials, targetCredentials);
        } else {
            await validateUniqueS3BucketForBroker(dataBroker, source, target);
            await relationship.startCopy(id, sourceCredentials, targetCredentials, settings);
            relationship = await relationship.save();
        }
        const relationshipResponse = await getRelationshipResponse(relationship, dataBroker);
        res.status(202).json(relationshipResponse);
        setHubSpotContactFirstRelationshipSinceIfNeeded(user);
        updatePortalLambda(user.portalId);
    } catch (error) {
        if (error.code === MONGO_DB_ERROR_DUPLICATE_KEY || error.errorCode === MONGO_DB_ERROR_DUPLICATE_KEY) {
            logger.error(`Relationship already exists source ${JSON.stringify(source)} target ${JSON.stringify(target)}`, error);
            next(new BadRequest('Sync relationship already exists'));
        } else {
            logger.error('Failed creating relationship with error', error);
            await Relationship.findOneAndRemove({'_id': relationship._id}).exec();
            await RelationshipStats.deleteMany({'relationship': relationship._id}).exec();
            next(error);
        }
    }
});

router.delete(`${API_PATH_RELATIONSHIP}/:id`, async(req, res, next) => {
    const {params: {id: relationshipId}, id: requestId} = req;
    try {
        const relationship = await Relationship.findById(relationshipId).populate('user dataBroker').populate({
            path: 'group',
            populate: {path: 'dataBrokers'}
        }).exec();
        createTimeLineObject(req, relationship);
        if (!relationship) {
            logger.warn(`Relationship with id ${relationshipId} was not found for deletion`);
            next(new NotFound());
        } else {
            const {dataBroker} = relationship;
            if (isNewDataBroker(dataBroker)) {
                handleEndOfRelationship(relationship, requestId, 'remove');
                const relationshipResponse = await getRelationshipResponse(relationship, dataBroker);
                await RelationshipStats.deleteMany({'relationship': relationshipId}).exec();
                await Promise.all([updateTimeLineNewDataBroker(relationship, relationshipResponse, 'DELETE', findTimelineRecord, 'activity'),
                    updateTimeLineNewDataBroker(relationship, relationshipResponse, 'DELETE', findTimelineRecordV2, 'relationship.activity')]);
            } else {
                relationship.doDelete(requestId);
                await Promise.all([updateTimeLine(relationship, undefined, 'DELETE'), updateTimeLineV2(relationship, undefined, 'DELETE')]);
                await Promise.all([updateTimeLine(relationship, undefined, 'DELETE', true),
                    updateTimeLineV2(relationship, undefined, 'DELETE', true)]); //in case there was an abort
            }
            await relationship.deleteOne();
            res.json();
        }

    } catch (err) {
        logger.warn(`Failed to send delete relationship command to data broker, relationship id ${relationshipId}`, err);
        next(err);
    }
});

router.put(`${API_PATH_RELATIONSHIP_V2}/:id`, async(req, res, next) => {
    const {params: {id: relationshipId}, body: {deleteOnTarget, gracePeriod, retries, fileTypes}} = req;
    try {
        const relationship = await Relationship.findById(relationshipId).populate('user dataBroker').populate({
            path: 'group',
            populate: {path: 'dataBrokers'}
        }).exec();
        if (relationship) {
            createTimeLineObject(req, relationship);
            if (!isNil(deleteOnTarget)) {
                relationship.settings.deleteOnTarget = deleteOnTarget;
            }
            if (!isNil(gracePeriod)) {
                validateGracePeriod(gracePeriod);
                relationship.settings.gracePeriod = gracePeriod;
            }
            if (!isNil(retries)) {
                relationship.settings.retries = retries;
            }
            if (!isNil(fileTypes)) {
                validateFileTypes(Object.assign({
                    files: true, directories: true, symlinks: true
                }, fileTypes));
                if (!isNil(fileTypes.files)) {
                    relationship.settings.fileTypes.files = fileTypes.files;
                }
                if (!isNil(fileTypes.directories)) {
                    relationship.settings.fileTypes.directories = fileTypes.directories;
                }
                if (!isNil(fileTypes.symlinks)) {
                    relationship.settings.fileTypes.symlinks = fileTypes.symlinks;
                }
            }
            await relationship.save();
            return res.status(200).json({
                id: relationshipId,
                gracePeriod: relationship.settings.gracePeriod,
                deleteOnTarget: relationship.settings.deleteOnTarget,
                retries: relationship.settings.retries,
                fileTypes: relationship.settings.fileTypes
            });
        } else {
            return next(new BadRequest(`relationship with id ${relationshipId} was not found`));
        }
    } catch (err) {
        logger.error(`Error while trying to change additional info of relationship ${relationshipId}`);
        return next(err);
    }
});

router.put(`${API_PATH_RELATIONSHIP}/:id/sync`, async(req, res, next) => {
    const {params: {id: relationshipId}, id} = req;
    try {
        const relationship = await Relationship.findById(relationshipId).populate('user dataBroker').populate({
            path: 'group',
            populate: {path: 'dataBrokers'}
        }).exec();
        createTimeLineObject(req, relationship);

        if (!relationship) {
            logger.debug(`Relationship with id ${relationshipId} was not found for syncing`);
            return next(new NotFound());
        } else {
            const {dataBroker, group} = relationship;
            isThereAliveDataBroker(group, dataBroker);

            if (isNewDataBroker(dataBroker)) {
                await startNewDataBrokerSync(relationship, id, false, true);
            } else {
                await validateAndStartSync(relationship, id, false, true);
            }
            res.status(202).json();
        }
    } catch (e) {
        next(e);
    }
});

router.put(`${API_PATH_RELATIONSHIP}/:id/abort`, async(req, res, next) => {
    const {params: {id: relationshipId}, id: requestId} = req;
    logger.info(`Got abort request for relationship with id ${relationshipId}`);//added for debug
    try {
        const relationship = await Relationship.findById(relationshipId).populate('user dataBroker')
            .populate({path: 'group', populate: {path: 'dataBrokers'}}).exec();
        createTimeLineObject(req, relationship);

        if (!relationship) {
            logger.debug(`Relationship with id ${relationshipId} was not found for aborting`);
            return next(new NotFound());
        } else {
            const {dataBroker, group} = relationship;
            isThereAliveDataBroker(group, dataBroker);
            if (isNewDataBroker(dataBroker)) {
                abortNewDataBroker(relationship, requestId);
            } else {
                validateAndStartAbort(relationship, requestId);
            }
            res.status(202).json();

        }
    } catch (e) {
        next(e);
    }
});

router.put(`${API_PATH_RELATIONSHIP}/:id/sync-schedule`, async(req, res, next) => {
    const {params: {id: relationshipId}} = req;
    try {
        const relationship = await Relationship.findById(relationshipId).populate('dataBroker').populate({
            path: 'group',
            populate: {path: 'dataBrokers'}
        }).exec();
        createTimeLineObject(req, relationship);

        if (!relationship) {
            logger.warn(`Relationship with id ${relationshipId} was not found for updating sync scheduler`);
            next(new NotFound(`Relationship with id ${relationshipId} was not found for updating sync scheduler`));
        } else {
            const {syncInDays: inDays, syncInHours: inHours, syncInMinutes, pause} = req.body;
            const inMinutes = syncInMinutes || 0;
            const autoSync = (pause) ? 'OFF' : 'ON';
            const syncObj = {inDays, inHours, inMinutes, autoSync};
            validateSyncSchedule(syncObj);
            syncObj.nextTime = calculateNextSync(inMinutes, inHours, inDays);
            relationship.sync = syncObj;
            const updatedRelationship = await relationship.save();
            res.json({
                id: updatedRelationship.id,
                autoSync: updatedRelationship.autoSync,
                sync: updatedRelationship.sync
            });
        }
    } catch (e) {
        next(e);
    }
});

agenda.define(AGENDA_TASK_NAME_SYNC_SCHEDULER, {lockLifetime: 60000}, (job, done) => {
    inContext('Relationships scheduled sync', async() => {
        logger.debug('Relationships scheduled sync task started');
        try {
            const relationships = await Relationship.getRelationshipsForScheduledSync();
            const promises = relationships.map(async relationship => {
                const {user, id: relationshipId, dataBroker, group} = relationship;
                logger.debug(`Running scheduled sync for relationship ${relationshipId} of user ${user} `);
                const rId = requestId();
                try {
                    isThereAliveDataBroker(group, dataBroker);
                    if (isNewDataBroker(dataBroker)) {
                        await startNewDataBrokerSync(relationship, rId, true);
                    } else {
                        await validateAndStartSync(relationship, rId, true);
                    }
                    relationship.sync.nextTime = getNextSyncTime(relationship.sync.nextTime, relationship.sync.inDays, relationship.sync.inHours, relationship.sync.inMinutes);
                    await relationship.save();
                } catch (err) {
                    logger.error(`Failed syncing relationship ${relationship.id} of user ${user}:`, err.message);
                    relationship.sync.nextTime = getNextSyncTime(relationship.sync.nextTime, relationship.sync.inDays, relationship.sync.inHours, relationship.sync.inMinutes);
                    await relationship.save();
                }
            });
            await Promise.all(promises);
        } catch (e) {
            logger.error('Relationships scheduled sync task failed:', e);
        } finally {
            done();
        }
    });
});


export default router;
