import express from 'express';
import crypto from 'crypto';
import Promise from 'bluebird';
import moment from 'moment';
import numeral from 'numeral';
import {flatten} from 'lodash';
import {BadRequest} from '../utils/errors';
import User from '../models/user';
import Cost from '../models/cost';
import Relationship from '../models/relationship';
import DataBroker from '../models/data-broker';
import Timeline from '../models/timeline';
import TimelineV2 from '../models/timeline-v2';
import {updateSubscription} from '../utils/common';
import {resolveCustomerIdentifierFromRegistrationTokenAsync} from '../lib/marketplace-metering';
import {
    API_PATH_USERS_INFO, TIER_1_BILLING_PER_HOUR, TIER_2_BILLING_PER_HOUR,
    TIER_3_BILLING_PER_HOUR, CONTRACT_SUBSCRIPTION, PAYGO_SUBSCRIPTION,
    SAAS_REGISTRATION_STATUS_UNSUBSCRIBED_INIT, INTERCOM_SECRET
} from '../utils/consts';
import {authenticateAsync, getLicenseAsync} from '../lib/ipa-server';
import {isEmptyString} from '../utils/utils';
import {getLogger} from '../utils/logger';
const logger = getLogger();

const router = express.Router();

function calculateSumOfBillingInATier(numOfRelationships, numOfHours, tier) {
    let billingPerHour;
    if (tier === 1) {
        billingPerHour = TIER_1_BILLING_PER_HOUR;
    } else if (tier === 2) {
        billingPerHour = TIER_2_BILLING_PER_HOUR;
    } else {
        billingPerHour = TIER_3_BILLING_PER_HOUR;
    }
    return Number((numOfHours * numOfRelationships * billingPerHour).toFixed(3));
}

function validateMonthAndYear(month, year) {
    const monthsArr = moment.months();
    if (!monthsArr.find(monthInTheArr => monthInTheArr === month)) {
        throw new BadRequest('Month should be a full month name ');
    }
    if (year.length !== 4) {
        throw new BadRequest('Year should contain 4 digits');
    }
    if (year > moment().year()) {
        throw new BadRequest('Year should be current or previous year');
    }
}

function updateNonContractUsersByAWSAccountId(customerIdentifier, status, user, type) {
    return DataBroker.find({user}).exec()
        .map(dataBroker => DataBroker.findUsersByAwsAccountId(dataBroker.awsAccountId))
        .then(result => updateSubscription({customerIdentifier, status, type}, flatten(result)));
}

function resolveAndUpdateIfNeeded(mpToken, user) {
    // Assumptions:
    // 1. We can get MP token only once per AWS account
    // 2. This function is invoked only when MP token is provided
    // 3. Resolve the token for customer identifier - in case of success (customer identifier is received) : find out which type the SAAS is by the Product code
    // -put customer identifier status and type on every user with the same AWS account id
    // 4. Do not update other users if the resolving failed - for example - user can "still" AWS account id, provide wrong MP token and reset subscription status for other users
    return resolveCustomerIdentifierFromRegistrationTokenAsync(mpToken)
        .then(triple => {
            const {customerIdentifier, type, status} = triple;
            if (customerIdentifier) {
                return updateNonContractUsersByAWSAccountId(customerIdentifier, status, user, type)
                    .return(triple);
            } else {
                return triple;
            }
        });
}

function handleUserSubscriptionAndTrialExpiry(user, token) {
    //user has token and it has PAYGO subscription and its status is not subscribe-success or subscribe-fail
    if (token && user.subscription.type === PAYGO_SUBSCRIPTION) {
        logger.info('handleUserSubscription got user', user);
        //if the user has PAYGO or doesn't have any subscription at all resolve, if its contract don't do anything
        return resolveAndUpdateIfNeeded(token, user)
            .then(({customerIdentifier, status, type}) => {
                logger.info(`handleLogin customerIdentifier: ${customerIdentifier}: status: ${status}`);
                user.subscription.customerIdentifier = customerIdentifier;
                user.subscription.status = status;
                user.subscription.type = type;
                user.trialExpiry = (type === CONTRACT_SUBSCRIPTION) ? new Date() : user.trialExpiry;
                return user.save();
            });
    } else {
        return user;
    }
}

function getIntercomHash(data) {
    const hmac = crypto.createHmac('sha256', INTERCOM_SECRET);
    hmac.update(data);
    return hmac.digest('hex');
}


function getLoginResponse(user, awsAccountId) {
    return {
        uid: user.id,
        name: user.name,
        email: user.email,
        company: user.company,
        hash: getIntercomHash(user.id),
        daysLeftForTrial: user.calculateDaysLeftForTrial(),
        subscriptionStatus: user.subscription.status || SAAS_REGISTRATION_STATUS_UNSUBSCRIBED_INIT,
        subscriptionType: user.subscription.type || PAYGO_SUBSCRIPTION,
        awsAccountId //TODO delete once OCCM is updated
    };
}


router.get(API_PATH_USERS_INFO, async(req, res, next) => {
    try {
        if (req.user) {
            const user = await req.user.addTrialDaysForNetappUsersAndSaveIfNeeded();
            const userWithSubscription = await handleUserSubscriptionAndTrialExpiry(user, req.query.mpCustomerToken);
            res.json(getLoginResponse(userWithSubscription));
        } else {
            res.status(202).json();
        }
    } catch (err) {
        next(err);
    }
});

//TODO Delete the API once OCCM is updated
router.put('/users/local/update', (req, res, next) => {
    try {
        res.json(getLoginResponse(req.user, req.body.awsAccountId));
    } catch (err) {
        next(err);
    }
});

router.put('/users/nss-credentials', (req, res, next) => {
    const {nssUsername, nssPassword} = req.body;
    const {user} = req;

    if (isEmptyString(nssUsername) || isEmptyString(nssPassword)) {
        throw new BadRequest('Invalid NSS credentials');
    }

    if (user.byol.licenses.length) {
        throw new BadRequest('Cannot change NSS credentials after inserting license');
    }

    return authenticateAsync(nssUsername, nssPassword)
        .then(responseStatus => {
            logger.info(`ResponseStatus from authenticate nss credentials is ${responseStatus}`);
            if (responseStatus === 500) {
                throw new BadRequest('Invalid NSS credentials');
            }

            user.byol.nssCredentials = {
                nssUsername,
                nssPassword
            };
            return user.save();
        })
        .then(() => res.json())
        .catch(next);
});

router.put('/users/licenses', (req, res, next) => {
    const {body: {serialNumber}, user} = req;

    if (isEmptyString(serialNumber)) {
        throw new BadRequest('Invalid serialNumber');
    }
    if (!user.byol.nssCredentials) {
        throw new BadRequest('Please insert NSS credentials first');
    }

    return User.findOne({'byol.licenses.serialNumber': {$eq: serialNumber}})
        .then(userWithSerialNumber => {
            if (userWithSerialNumber) {
                if (userWithSerialNumber.id === user.id) {
                    throw new BadRequest(`License with serial number: ${serialNumber} is already defined `);
                } else {
                    throw new BadRequest(`License with serial number: ${serialNumber} in use by other user then user with id ${user.id}`);
                }
            }
        })
        .then(() => {
            return getLicenseAsync(user.byol.nssCredentials.nssUsername, user.decryptedNssPassword(), serialNumber)
                .then(licenseResponse => {
                    const response = licenseResponse.statusResp;
                    logger.info('LicenseResponse after getLicenseAsync', licenseResponse);
                    if (response.statusCode !== 'S007') {
                        logger.error(`Invalid serial number for the inserted NSS credentials got response ${JSON.stringify(response)} 
                        with statusCode ${response.statusCode} for userId ${user.id}`);
                        throw new BadRequest('Invalid serial number for the inserted NSS credentials');
                    }
                    if (!response.licenses || !response.licenses.capacity || !response.licenses.endDate) {
                        throw new BadRequest('Invalid serial number');
                    }

                    /*        let reservedRelationships = parseInt(response.licenses.capacity);
                    if (reservedRelationships < 0) {
                        logger.info(`reservedRelationships: ${reservedRelationships} is less than zero for user ${user.id}`);
                        reservedRelationships = 0;
                    }*/
                    const expirationDate = new Date(response.licenses.endDate);

                    user.byol.licenses.push({
                        serialNumber,
                        reservedRelationships : 20,
                        expirationDate
                    });
                    return user.save()
                        .then(() => {
                            return res.json({
                                serialNumber,
                                reservedRelationships: 20,
                                expirationDate: expirationDate.getTime()
                            });
                        });
                })
                .catch(e => {
                    if (e.statusCode !== 400) {
                        throw new BadRequest('Invalid serial number');
                    } else {
                        throw e;
                    }
                });
        }).catch(next);
});

router.get('/users/licenses', async(req, res, next) => {
    const {user} = req;
    if (user) {
        try {
            const userFromDb = await User.findById(user.id).exec();
            const {byol: {licenses}} = userFromDb;
            const licensesWithTimeStamp = licenses.map(license => {
                const {serialNumber, reservedRelationships, expirationDate} = license;
                return {
                    serialNumber,
                    reservedRelationships,
                    expirationDate: expirationDate.getTime()
                };
            });
            return res.json({licenses: licensesWithTimeStamp});
        } catch (e) {
            next(e);
        }
    } else {
        res.json({licenses: []});
    }
});

router.get('/users/license-settings', (req, res, next) => {
    const {user} = req;
    const totalReservedRelationships = user.reservedRelationships();

    user.relationshipsCount()
        .then(existingRelationships => {
            logger.info(`Number of relationships per user with id ${user.id} is ${existingRelationships}`);
            let freeTrial = false;
            if (user.trialExpiry && user.trialExpiry >= new Date()) {
                logger.info(`The current license is Free Trial.TrialExpiry is: ${user.trialExpiry}`);
                freeTrial = true;
            }

            let chargedRelationships = existingRelationships - totalReservedRelationships;
            let byolRelationships = totalReservedRelationships;

            if (chargedRelationships <= 0) {
                byolRelationships = existingRelationships;
                chargedRelationships = 0;
            }

            return {
                freeTrial,
                existingRelationships,
                reservedRelationships: byolRelationships,
                totalReservedRelationships,
                chargedRelationships
            };
        })
        .then(response => res.json(response))
        .catch(next);
});

router.delete('/users', (req, res, next) => {
//#1 check if the user has relationships and data brokers if do throw ERROR
//#2 Id the user doesnt have relationships and databrokers delete it from Timeline, and COST
    const deleteUserFromCollection = (userId) => {
        return Promise.map([Timeline, Cost], collection => {
            return collection.deleteOne({user: userId});
        });
    };
    const userId = req.user.id;
    if (req.user.subscription.customerIdentifier) {
        logger.info('The user is subscribed, cannot be deleted');
        throw new BadRequest('Subscribed users cannot be deleted');
    }
    return Promise.join(Relationship.findOne({'user': userId}).exec(), DataBroker.findOne({'user': userId}).exec())
        .then(searchResults => {
            if (searchResults[0] || searchResults[1]) {
                logger.info('The user has relationships or data brokers, cannot be deleted');
                throw new BadRequest('User with relationships or data brokers cannot be deleted');
            } else {
                return deleteUserFromCollection(userId)
                    .then(() => User.deleteOne({'_id': userId}).exec());
            }
        })
        .then(() => res.json())
        .catch(next);
});

router.put('/users/delete-license', async(req, res, next) => {
    const {body : {serialNumber}} = req;
    try {
        await User.findOneAndUpdate({'byol.licenses.serialNumber': serialNumber},
            {$pull: {'byol.licenses': {'serialNumber': serialNumber}}}).exec();
        res.json();
    } catch (e) {
        next(e);
    }
});

router.get('/users/relationships-stats', async(req, res, next) => {
    const {user } = req;
    try {

        //#1 find all bytes and files copied forever - all relationships that has DONE do we want ABORTED or deleted as well?
        const [aggregateResult] = await TimelineV2.aggregate([{
            $match: {
                user: user._id,
                status: {$nin: ['FAILED', 'PROGRESS']},
                'relationship.relationshipId': {$exists: true}
            }
        },
        {
            $group: {
                _id: '$user',
                allBytesCopied: {$sum: '$relationship.activity.bytesCopied'},
                allFilesCopied: {$sum: '$relationship.activity.filesCopied'}
            }
        }]);
        const totalBytes = aggregateResult ? aggregateResult.allBytesCopied : 0;
        const totalFiles = aggregateResult ? aggregateResult.allFilesCopied : 0;

        const bytesCopiedInTheLast30Days = await TimelineV2.aggregate([{
            $match: {
                user: user._id,
                status: {$nin: ['FAILED', 'PROGRESS']},
                'relationship.relationshipId': {$exists: true},
                'relationship.activity.endTime': {$gte: new Date(moment().subtract(30, 'day'))}
            }
        },
        {
            $project: {
                'date': {
                    $dateToString: {
                        format: '%Y-%m-%d',
                        date: '$relationship.activity.endTime'
                    }
                }, activity: '$relationship.activity'
            }
        },
        {$group: {_id: '$date', bytesCopied: {$sum: '$activity.bytesCopied'}}},
        {$project: {'date': '$_id', '_id': 0, 'bytesCopied': 1}}]);
        const result = {
            totalBytes,
            totalFiles,
            bytesCopiedInTheLast30Days
        };
        res.json(result);
    } catch (e) {
        next(e);
    }
});

router.get('/users/cost-times', (req, res) => {
    const creationDate = moment(req.user.createdAt).startOf('month');
    const today = moment().utc();
    const monthsAndYears = [];
    while (today.isAfter(creationDate)) {
        monthsAndYears.unshift({year: creationDate.get('year'), month: creationDate.format('MMMM')});
        creationDate.add(1, 'M');
    }
    return res.json(monthsAndYears);
});

router.get('/users/cost', (req, res, next) => {
    //Missing most important thing- if the first timestamp is from the middle of the month I need to know the last record from last month and change its from to the start of the month
    const {year, month} = req.query;
    validateMonthAndYear(month, year);
    const isCurrentMonthAndYear = (moment().format('MMMM') === month && moment().format('YYYY') === year);
    const currentTimeEpoch = Date.now();
    const endTimeStamp = moment().month(month).year(year).endOf('month').toDate();
    const startTimeStamp = moment().month(month).year(year).startOf('month').toDate();
    const startOfNextMonth = moment().month(month).year(year).add(1, 'M').startOf('month').toDate();
    const numeralObj = numeral();
    let totalCost = 0;
    let hasByol = false;
    let hasTier1 = false;
    let hasTier2 = false;
    let hasTier3 = false;

    function getCostRecord(record, from, to) {
        const numOfHours = moment(to).diff(from, 'hours');
        const tier1Sum = calculateSumOfBillingInATier(record.tier1, numOfHours, 1);
        const tier2Sum = calculateSumOfBillingInATier(record.tier2, numOfHours, 2);
        const tier3Sum = calculateSumOfBillingInATier(record.tier3, numOfHours, 3);
        const recordTotalCost = tier1Sum + tier2Sum + tier3Sum;
        totalCost += recordTotalCost;
        const totalRelationships = (record.tier1 + record.tier2 + record.tier3 + record.byol.usedRelationships);
        hasByol = hasByol || Boolean(record.byol && (record.byol.usedRelationships || record.byol.enabledRelationship));
        hasTier1 = hasTier1 || Boolean(record.tier1);
        hasTier2 = hasTier2 || Boolean(record.tier2);
        hasTier3 = hasTier3 || Boolean(record.tier3);
        return {
            from: from,
            to: to,
            hours: numOfHours,
            tier1: {
                numOfRelationships: record.tier1,
                sumOfBilling: numeralObj.set(tier1Sum).format('0.00')
            },
            tier2: {
                numOfRelationships: record.tier2,
                sumOfBilling: numeralObj.set(tier2Sum).format('0.00')
            },
            tier3: {
                numOfRelationships: record.tier3,
                sumOfBilling: numeralObj.set(tier3Sum).format('0.00')
            },
            byol: record.byol,
            totalCost: numeralObj.set(recordTotalCost).format('0.00'),
            totalRelationships
        };
    }

    const getAllRecordsInAMonth = Cost.find({
        $and: [
            {user: req.user},
            {
                timestamp: {
                    $gte: startTimeStamp,
                    $lt: endTimeStamp
                }
            }
        ]
    }).sort({timestamp: 1}).exec();
    const getLastRecordBefore = Cost.find({$and: [{'user': req.user}, {'timestamp': {$lt: startTimeStamp}}]}).sort({timestamp: -1}).limit(1).exec();
    return Promise.all([getAllRecordsInAMonth, getLastRecordBefore])
        .then(([costRecords, cost]) => {
            let costArr = costRecords.map((record, indexOfTheArray, array) => {
                const to = (indexOfTheArray === array.length - 1) ? (isCurrentMonthAndYear ? currentTimeEpoch : startOfNextMonth.getTime()) : array[indexOfTheArray + 1].timestamp.getTime();
                const from = record.timestamp.getTime();
                return getCostRecord(record, from, to);
            });
            if (cost.length && (!costArr.length || costArr[0].from !== startTimeStamp.getTime())) {
                const to = costArr.length ? costArr[0].from : (isCurrentMonthAndYear ? currentTimeEpoch : startOfNextMonth.getTime());
                costArr.unshift(getCostRecord(cost[0], startTimeStamp.getTime(), costArr.length ? costArr[0].from : to));
            }
            //need to go over all the costRecords and delete the records with 0 if there are no chargeable relationships
            if (!(hasByol || hasTier1 || hasTier2 || hasTier3)) {
                costArr = costArr.filter(cost => {
                    return cost.tier1.numOfRelationships || cost.tier2.numOfRelationships || cost.tier3.numOfRelationships || (cost.byol && cost.byol.usedRelationships);
                });
            }
            const response = {
                year,
                month,
                isTier1: hasTier1,
                isTier2: hasTier2,
                isTier3: hasTier3,
                isBYOL: hasByol,
                costArr,
                totalCost: numeralObj.set(totalCost).format('0.00')
            };
            return res.json(response);
        })
        .catch(next);
});


export default router;
