import os from 'os';
import fs from 'fs';
import path from 'path';
import uuid from 'uuid';
import moment from 'moment';
import format from 'string-template';
import { last, split, isEmpty, remove, head, drop} from 'lodash';
import DataBroker from '../models/data-broker';
import Relationship from '../models/relationship';
import { MP_PRODUCT_CODE_METERING, SCAN_MODE_SOURCE_AND_TARGET } from '../utils/consts';
import { REGION, SUBNET_CIDR_BLOCK, SUBNET_ID, VPC_CIDR_BLOCK, VPC_ID, AVAILABILITY_ZONE,
    PRIVATE_IP, INSTANCE_ID, AWS_ACCOUNT_ID_2, AWS_ACCOUNT_ID_1 } from './consts';
import { timeout } from '../utils/utils';
import {isNewDataBroker} from '../utils/common';

import {getLogger} from '../utils/logger';
const logger = getLogger();

const DELAY_TIME_MS = 3000;
const RECEIVE_TIMEOUT_MS = 10000;
const messagesQueue = [];
const FAILURE_ADDRESS_CREATE_REGEX = /fail\.com/;
const FAILURE_ADDRESS_SYNC_REGEX = /failsync\.com/;
const SUCCESS_ADDRESS_CREATE_REGEX = /success\.com/;
const relationshipsToAbort = [];
const runningRelationshipsNewDataBroker = {};

function createSQSMessage(dataBrokerId, command, payload) {
    return {
        ReceiptHandle: uuid.v4(),
        Body: JSON.stringify({
            command,
            header: {
                dataBrokerId
            },
            payload
        })
    };
}

function createNewDataBrokerSQSMessage(payload, attributes) {
    return {
        ReceiptHandle: uuid.v4(),
        Body: JSON.stringify(payload),
        MessageAttributes: attributes
    };
}

export function sendHealthMessage(dataBroker) {
    const {id: dataBrokerId} = dataBroker;
    if (isNewDataBroker(dataBroker)) {
        const payload = {
            transferrers: 4,
            scanners: 4,
            memoryUsageRss: 77336576,
            memoryUsageHeapTotal: 29220864,
            memoryUsageHeapUsed: 25610728,
            transferRate: 1223213.5435
        };
        const attributes = {type:{StringValue: 'health', DataType: 'String'}, dataBrokerId:{StringValue: dataBrokerId, DataType: 'String'}};
        messagesQueue.push(createNewDataBrokerSQSMessage(payload, attributes));
    } else {
        messagesQueue.push(createSQSMessage(dataBrokerId, 'health', {
            'usedMb' : 237,
            'freeMb' : 463,
            'maxMb' : 3568,
            'loadAvg' : 0.0
        }));
    }
}

function createFolders(path) {
    let folders;

    const lastDir = last(split(path, '/'));
    switch (lastDir) {
        case 'folder4':
        case 'folder0':
            folders = [];
            break;
        case 'folder1':
            folders = ['folder0', 'folder1', 'folder3'];
            break;
        case 'folder3':
            folders = ['folder0', 'folder1'];
            break;
        case  '' :
            folders = ['folder1', 'folder2'];
            break;
        default :
            folders = ['folder0', 'folder1', 'folder2', 'folder3', 'folder4', 'folder5'];
    }
    return folders;
}

function createBucketsFoldersList(bucket, prefix) {
    let folders;
    switch (bucket) {
        case 'small': {
            if (!prefix) {
                folders = ['/small1'];
            } else if (['/small1', 'small1'].includes(prefix)) {
                folders = ['small2', 'small3'];
            } else {
                folders = [];
            }
        }
            break;
        case 'medium': {
            if (!prefix) {
                folders = ['medium1'];
            } else if (['/medium1', 'medium1'].includes(prefix)) {
                folders = ['medium2', 'medium3'];
            } else {
                folders = [];
            }
        }
            break;
        case 'tiny': {
            if (!prefix) {
                folders = ['tiny1'];
            } else if (['/tiny1', 'tiny1'].includes(prefix)) {
                folders = ['tiny2', 'tiny3'];
            } else {
                folders = [];
            }
        }
            break;
        case 'joey':
        case 'phoebe':
        case 'chandler':
        case 'monica':
        case 'ross':
        case 'rachel':
            {
                if (!prefix) {
                    folders = ['centralPark1'];
                } else if (['/centralPark1', 'centralPark1'].includes(prefix)) {
                    folders = ['centralPark2', 'centralPark3', 'centralParkIsFullOfSpecialCharacters!@#$%^&*(){}\\[];:~.",\'?|'];
                } else {
                    folders = [];
                }
            }
            break;
        default:
            folders = [];
    }
    return folders;
}

//Asterix
function buildListNfsExportsResponse(dataBrokerId, host, isNfsExportsEmpty) {
    return createSQSMessage(dataBrokerId, 'list-nfs-exports-v2', {
        host: host,
        exports: (isNfsExportsEmpty) ? [] : ['/data1', '/data2', '/logs', '/', '/complex/junction/path'],
        status: 'success'
    });
}

function buildListNfsExportFoldersResponse(dataBrokerId, host, exportName, path, isNfsExportsFoldersEmpty) {
    const folders = (isNfsExportsFoldersEmpty) ? [] : createFolders(path);
    return createSQSMessage(dataBrokerId, 'list-nfs-export-folders-v2', {
        host,
        export: exportName,
        path,
        folders,
        status: 'success'
    });
}

function buildListS3BucketsResponse(dataBrokerId, isS3BucketsEmpty) {

    const createBucketsList = () => {
        return [
            {
                name: 'small',
                region: 'us-west-2',
                tags: {price: '100$'}
            },
            {
                name: 'medium',
                region: 'us-west-1',
                tags: {one: '1', two: '2', three: '3'}
            },
            {
                name: 'large',
                region: 'eu-central-1',
                tags: {}
            },
            {
                name: 'tiny',
                region: 'us-east-1',
                tags: {}
            },
            {
                name: 'joey',
                region: 'ap-northeast-1',
                tags: {}
            },
            {
                name: 'phoebe',
                region: 'eu-central-1',
                tags: {}
            },
            {
                name: 'chandler',
                region: 'ap-south-1',
                tags: {}
            },
            {
                name: 'monica',
                region: 'ap-south-1',
                tags: {}
            },
            {
                name: 'ross',
                region: 'ca-central-1',
                tags: {}
            },
            {
                name: 'rachel',
                region: 'eu-central-1',
                tags: {}
            },
            {
                name: 'fabric-pool-a1',
                region: 'eu-central-1',
                tags: {}
            }
        ];
    };

    return createSQSMessage(dataBrokerId, 'list-s3-buckets-v2', {
        'buckets': (isS3BucketsEmpty) ? [] : createBucketsList(),
        'status': 'success'
    });
}

function buildListS3BucketFoldersResponse(dataBrokerId, bucket, prefix, isS3BucketsFoldersEmpty) {
    return createSQSMessage(dataBrokerId, 'list-s3-bucket-folders-v2', {
        bucket,
        prefix,
        folders : (isS3BucketsFoldersEmpty) ? [] : createBucketsFoldersList(bucket, prefix),
        partial: false,
        status: 'success'
    });
}

function buildListCifsSharesResponse(command, dataBrokerId, host, isCifsSharesEmpty) {
    return createSQSMessage(dataBrokerId, command, {
        status: 'success',
        shares: (isCifsSharesEmpty) ? [] : ['/data1', '/data2', '/logs'],
        host
    });
}

function buildListCifsShareFoldersResponse(command, dataBrokerId, host, share, path = '') {
    let folders;

    switch (path.split('/').pop()) {
        case 'folder4':
        case 'folder0':
            folders = [];
            break;
        case 'folder1':
            folders = ['/folder0', '/folder1', '/folder3'];
            break;
        case 'folder3':
            folders = ['/folder0', '/folder1'];
            break;
        case  '' :
            folders = ['/folder1', '/folder2'];
            break;
        default :
            folders = ['/folder0', '/folder1', '/folder2', '/folder3', '/folder4', '/folder5'];
    }

    return createSQSMessage(dataBrokerId, command, {
        host,
        folders,
        path,
        share,
        status: 'success'
    });
}

function buildListSgwsBucketsResponse(dataBrokerId, host, port, isSgwsBucketsEmpty) {
    return createSQSMessage(dataBrokerId, 'list-sgws-buckets-v2', {
        host,
        port,
        buckets: (isSgwsBucketsEmpty) ? [] : [{
            name: 'small'
        }, {
            name: 'medium'
        }, {
            name: 'large'
        }],
        status: 'success'
    });
}

function buildListSgwsBucketFoldersResponse(dataBrokerId, host, port, bucket, prefix) {
    let folders;
    switch (bucket) {
        case 'small': {
            if (isEmpty(prefix)) {
                folders = ['/small1'];
            } else if (prefix === '/small1') {
                folders = ['/small2', '/small3'];
            } else if (prefix === '/small1/small2' || prefix === '/small1/small3') {
                folders = [];
            } else {
                folders = [];
            }
        }
            break;
        case 'medium': {
            if (isEmpty(prefix)) {
                folders = ['/medium1'];
            } else if (prefix === '/medium1') {
                folders = ['/medium2', '/medium3'];
            } else if (prefix === '/medium1/medium2' || prefix === '/medium1/medium3') {
                folders = [];
            } else {
                folders = [];
            }
        }
            break;
        default:
            folders = [];
    }

    return createSQSMessage(dataBrokerId, 'list-sgws-bucket-folders-v2', {
        bucket,
        host,
        partial: false,
        prefix,
        port,
        folders,
        status: 'success'
    });
}


//new data broker
function buildListNfsExportFoldersResponseNewDataBroker(host, exportName, path, attributes, isNfsExportsFoldersEmpty) {
    const folders = (isNfsExportsFoldersEmpty) ? [] : createFolders(path);
    return createNewDataBrokerSQSMessage({
        host,
        export: exportName,
        path,
        folders,
        status: 'SUCCEEDED'
    }, attributes);
}

function buildListNfsExportsResponseNewDataBroker(host, attributes, isNfsExportsEmpty) {
    const payload = {
        host: host,
        exports: (isNfsExportsEmpty) ? [] : ['/data1', '/data2', '/logs', '/', '/complex/junction/path'],
        status: 'SUCCEEDED'
    };
    return createNewDataBrokerSQSMessage(payload, attributes);
}

function createBucketsListNewDataBroker() {
    return [
        {
            name: 'small',
            region: 'us-west-2',
            tags: [{Key: 'price', Value: '100$'}]
        },
        {
            name: 'medium',
            region: 'us-west-1',
            tags:  [{Key: 'one', Value: '1'}, {Key: 'two', Value: '2'}, {Key: 'three', Value: '3'}]
        },
        {
            name: 'large',
            region: 'eu-central-1',
            tags: []
        },
        {
            name: 'tiny',
            region: 'us-east-1',
            tags: []
        },
        {
            name: 'joey',
            region: 'ap-northeast-1',
            tags: []
        },
        {
            name: 'phoebe',
            region: 'eu-central-1',
            tags: []
        },
        {
            name: 'chandler',
            region: 'ap-south-1',
            tags: []
        },
        {
            name: 'monica',
            region: 'ap-south-1',
            tags: []
        },
        {
            name: 'ross',
            region: 'ca-central-1',
            tags: []
        },
        {
            name: 'rachel',
            region: 'eu-central-1',
            tags: []
        },
        {
            name: 'fabric-pool-a1',
            region: 'eu-central-1',
            tags: []
        }
    ];
}

function buildListS3BucketsResponseNewDataBroker(attributes, isS3BucketsEmpty) {

    return createNewDataBrokerSQSMessage({
        'buckets': (isS3BucketsEmpty) ? [] : createBucketsListNewDataBroker(),
        'status': 'SUCCEEDED'
    }, attributes);
}

function buildListS3BucketFoldersResponseNewDataBroker(attributes, bucket, prefix, isS3BucketsFoldersEmpty) {
    return createNewDataBrokerSQSMessage({
        bucket : (isS3BucketsFoldersEmpty) ? undefined : head(createBucketsListNewDataBroker().filter(bucketObj =>  bucketObj.name === bucket)),
        prefix,
        folders : (isS3BucketsFoldersEmpty) ? [] : createBucketsFoldersList(bucket, prefix),
        status: 'SUCCEEDED'
    }, attributes);
}

function buildListAzureContainerFoldersResponse(attributes, container, prefix, storageAccountName, isEmptyDataBroker) {
    return createNewDataBrokerSQSMessage({
        container,
        prefix,
        folders : (isEmptyDataBroker) ? [] : createBucketsFoldersList(container, prefix),
        status: 'SUCCEEDED',
        storageAccountName
    }, attributes);
}

function buildListAzureContainersResponse(attributes, storageAccountName, isEmptyDataBroker) {
    return createNewDataBrokerSQSMessage({
        containers : (isEmptyDataBroker) ? [] : createBucketsListNewDataBroker(),
        status: 'SUCCEEDED',
        storageAccountName
    }, attributes);
}

function buildListGcpBucketFoldersResponse(attributes, bucket, prefix, projectId, isEmptyDataBroker) {
    return createNewDataBrokerSQSMessage({
        bucket : (isEmptyDataBroker) ? undefined : head(createBucketsListNewDataBroker().filter(bucketObj =>  bucketObj.name === bucket)),
        prefix,
        folders : (isEmptyDataBroker) ? [] : createBucketsFoldersList(bucket, prefix),
        status: 'SUCCEEDED',
        projectId
    }, attributes);
}

function buildListGcpBucketsResponse(attributes, projectId, isEmptyDataBroker) {
    return createNewDataBrokerSQSMessage({
        buckets : (isEmptyDataBroker) ? [] : createBucketsListNewDataBroker(),
        status: 'SUCCEEDED',
        projectId
    }, attributes);
}

function buildListCifsSharesResponseNewDataBroker(host, isCifsSharesEmpty, attributes) {
    return createNewDataBrokerSQSMessage({
        status: 'SUCCEEDED',
        shares: (isCifsSharesEmpty) ? [] : ['/data1', '/data2', '/logs'],
        host
    }, attributes);
}

function buildListCifsShareFoldersResponseNewDataBroker(attributes, host, share, path = '') {
    let folders;

    switch (path.split('/').pop()) {
        case 'folder4':
        case 'folder0':
            folders = [];
            break;
        case 'folder1':
            folders = ['/folder0', '/folder1', '/folder3'];
            break;
        case 'folder3':
            folders = ['/folder0', '/folder1'];
            break;
        case  '' :
            folders = ['/folder1', '/folder2'];
            break;
        default :
            folders = ['/folder0', '/folder1', '/folder2', '/folder3', '/folder4', '/folder5'];
    }

    return createNewDataBrokerSQSMessage({
        host,
        folders,
        path,
        share,
        status: 'SUCCEEDED'
    }, attributes);
}

function buildListSgwsBucketsResponseNewDataBroker(attributes, host, port, isSgwsBucketsEmpty) {
    return createNewDataBrokerSQSMessage({
        status: 'SUCCEEDED',
        buckets: (isSgwsBucketsEmpty) ? [] : [{
            name: 'small'
        }, {
            name: 'medium'
        }, {
            name: 'large'
        }],
        host,
        port
    }, attributes);
}

function buildListSgwsBucketFoldersResponseNewDataBroker(attributes, host, port, bucket, prefix) {
    let folders;
    switch (bucket) {
        case 'small': {
            if (isEmpty(prefix)) {
                folders = ['/small1'];
            } else if (prefix === '/small1') {
                folders = ['/small2', '/small3'];
            } else if (prefix === '/small1/small2' || prefix === '/small1/small3') {
                folders = [];
            } else {
                folders = [];
            }
        }
            break;
        case 'medium': {
            if (isEmpty(prefix)) {
                folders = ['/medium1'];
            } else if (prefix === '/medium1') {
                folders = ['/medium2', '/medium3'];
            } else if (prefix === '/medium1/medium2' || prefix === '/medium1/medium3') {
                folders = [];
            } else {
                folders = [];
            }
        }
            break;
        default:
            folders = [];
    }
    return createNewDataBrokerSQSMessage({
        bucket,
        host,
        prefix,
        port,
        folders,
        status: 'SUCCEEDED'
    }, attributes);
}

export function sendPlacementMessage(dataBroker) {
    const {name, type:dataBrokerType} = dataBroker;
    const awsAccountId = (dataBrokerType === 'AWS' || dataBrokerType === 'ONPREM') ? ((name.includes('diffAccount')) ? AWS_ACCOUNT_ID_2 : AWS_ACCOUNT_ID_1) : undefined;
    const isOnPremises = dataBrokerType === 'ONPREM';
    const placement = {
        instanceId: isOnPremises ? undefined : INSTANCE_ID,
        privateIp: PRIVATE_IP,
        awsAccountId: awsAccountId,
        platform: `${os.type()} ${os.release()}`,
        version: global.__VERSION__,
        availabilityZone: isOnPremises ? undefined : AVAILABILITY_ZONE,
        vpcId: isOnPremises ? undefined : VPC_ID,
        vpcCidrBlock: isOnPremises ? undefined : VPC_CIDR_BLOCK,
        subnetId: isOnPremises ? undefined : SUBNET_ID,
        subnetCidrBlock: isOnPremises ? undefined : SUBNET_CIDR_BLOCK,
        hostname: os.hostname()
    };
    if (isNewDataBroker(dataBroker)) {
        placement.cpus = 4;
        placement.node = '9.11.1';
        placement.osTotalMem = 16343470080;
        placement.processMaxMem = 55373824;
        placement.release = '4.9.76-3.78.amzn1.x86_64';
        placement.region = isOnPremises ? undefined : REGION;
        placement.publicIp = '192.168.0.30';
        messagesQueue.push(createNewDataBrokerSQSMessage(placement,
            {type: {StringValue: 'placement', DataType: 'String'},
                dataBrokerId:{StringValue: dataBroker.id, DataType: 'String'}}));
    } else {
        placement.region = isOnPremises ? 'on-premise' : REGION;
        messagesQueue.push(createSQSMessage(dataBroker.id, 'placement', placement));
    }
}

AWS.mock('SQS', 'deleteMessage', () => ({promise: () => ({})}));
AWS.mock('SQS', 'deleteQueue', () => ({promise: () => ({})}));
AWS.mock('SQS', 'setQueueAttributes', () => ({promise: () => ({})}));
AWS.mock('SQS', 'getQueueAttributes', params => ({
    promise: () => {
        const {AttributeNames: attributesNames, QueueUrl} = params;
        if (attributesNames.includes('ApproximateNumberOfMessages') || attributesNames.includes('ApproximateNumberOfMessagesNotVisible')) {
            const key = Object.keys(runningRelationshipsNewDataBroker).find(key => key.includes(QueueUrl));
            if (!key) {
                return {Attributes: {QueueArn: 'arn:aws:sqs:us-east-1:123456789012:ARN', ApproximateNumberOfMessages: '0', ApproximateNumberOfMessagesNotVisible: '0'}};
            } else {
                if (runningRelationshipsNewDataBroker[key].length > 0) {
                    return {
                        Attributes: {
                            QueueArn: 'arn:aws:sqs:us-east-1:123456789012:ARN',
                            ApproximateNumberOfMessages: '3',
                            ApproximateNumberOfMessagesNotVisible: '4'
                        }
                    };
                } else {
                    return {Attributes: {QueueArn: 'arn:aws:sqs:us-east-1:123456789013:ARN', ApproximateNumberOfMessages: '0', ApproximateNumberOfMessagesNotVisible: '0'}};
                }
            }
        } else {
            return {Attributes: {QueueArn: 'arn:aws:sqs:us-east-1:123456789012:ARN'}};
        }
    }
}));

AWS.mock('SQS', 'createQueue', params => ({
    promise: () => ({QueueUrl: `https://sqs.us-east-1.amazonaws.com/${Date.now()}/${params.QueueName}`})
}));

AWS.mock('SQS', 'receiveMessage', params => {
    return {
        promise: async() => {
            let interval;
            let timeout;

            const p1 = new Promise((resolve) => {
                timeout = setTimeout(() => {
                    clearInterval(interval);
                    resolve({Messages: []});
                }, RECEIVE_TIMEOUT_MS);
            });

            const p2 = new Promise((resolve, reject) => {
                interval = setInterval(async() => {
                    try {
                        const dataBroker = await DataBroker.findOne({statusesQueueUrl: params.QueueUrl}).exec();
                        if (dataBroker) {
                            const relevantMessages = remove(messagesQueue, message => {
                                const body = JSON.parse(message.Body);
                                if (isNewDataBroker(dataBroker)) {
                                    const attr = message.MessageAttributes;
                                    return attr && attr.dataBrokerId && attr.dataBrokerId.StringValue === dataBroker.id;
                                } else {
                                    return body.header && body.header.dataBrokerId === dataBroker.id;
                                }
                            });

                            if (relevantMessages.length > 0) {
                                clearInterval(interval);
                                clearTimeout(timeout);
                                resolve({Messages: relevantMessages});
                            }
                        }

                        if (params.QueueUrl.includes('MP_SUBSCRIPTION_STATUS_')) {
                            const mpMessages = remove(messagesQueue, message => {
                                const body = JSON.parse(message.Body);
                                if (body.Message) {
                                    message = JSON.parse(body.Message);
                                    return message['product-code'] === MP_PRODUCT_CODE_METERING;
                                }
                            });

                            if (mpMessages.length > 0) {
                                clearInterval(interval);
                                clearTimeout(timeout);
                                resolve({Messages: mpMessages});
                            }
                        }
                    } catch (err) {
                        logger.error('Received error on sqs receive messages', err);
                        clearInterval(interval);
                        clearTimeout(timeout);
                        reject(err);
                    }
                }, 500);
            });

            return Promise.race([p1, p2]);

        }};
});

export function sendSaaSSubscriptionMessage(customerIdentifier, action) {
    messagesQueue.push({
        ReceiptHandle: uuid.v4(),
        Body: JSON.stringify({
            Message: JSON.stringify({
                'customer-identifier': customerIdentifier,
                action,
                'product-code': MP_PRODUCT_CODE_METERING
            }),
            Timestamp: moment().utc().format('YYYY-MM-DDTHH:mm:ss.SSS[Z]')
        })
    });
}

async function runScenario(messages, canAbort = false) {
    if (messages.length > 0) {
        const headMessage = head(messages);
        if (canAbort && relationshipsToAbort.includes(headMessage.payload.relationshipId)) {
            remove(relationshipsToAbort, relationshipId => relationshipId === headMessage.payload.relationshipId);
            return;
        }

        const message = {
            ReceiptHandle: uuid.v4(),
            Body: JSON.stringify(headMessage)
        };

        messagesQueue.push(message);

        await timeout(DELAY_TIME_MS);
        return runScenario(drop(messages, 1), canAbort);
    }
}

function runScenarioNewDataBroker(key, attributes, canAbort = false, isUnknown = true) {
    let messages = runningRelationshipsNewDataBroker[key];
    if (messages && messages.length > 0) {
        const headMessage = head(messages);
        if (canAbort && relationshipsToAbort.includes(headMessage.relationshipId)) {
            remove(relationshipsToAbort, relationshipId => relationshipId === headMessage.relationshipId);
        } else if (!isUnknown) {
            const message = {
                ReceiptHandle: uuid.v4(),
                Body: JSON.stringify(headMessage),
                MessageAttributes: attributes
            };
            messagesQueue.push(message);
            messages = drop(messages, 1);
            runningRelationshipsNewDataBroker[key] = messages;
            setTimeout(runScenarioNewDataBroker.bind(this, key, attributes, canAbort, isUnknown), DELAY_TIME_MS);
        }
    }
}


function parseJsonAndRun(scenarioFileName, dataBrokerId, relationshipId, messageBody, canAbort = false) {
    const  { requestId } = messageBody.header;
    const scenarioFile = path.resolve(__dirname, scenarioFileName);
    if (!fs.existsSync(scenarioFile)) {
        logger.error(`Scenario file '${scenarioFileName}' not found to handle:`, messageBody);
        return;
    }
    const scenario = fs.readFileSync(scenarioFile, 'utf-8');
    const responseTime = Date.now() + DELAY_TIME_MS;
    const result = JSON.parse(format(scenario, {
        RELATIONSHIP_ID: relationshipId,
        DATA_BROKER_ID: dataBrokerId,
        REQUEST_ID: requestId,
        CURRENT_TIME_1: responseTime,
        CURRENT_TIME_2: responseTime + DELAY_TIME_MS * 2,
        CURRENT_TIME_3: responseTime + DELAY_TIME_MS * 3,
        CURRENT_TIME_4: responseTime + DELAY_TIME_MS * 4,
        CURRENT_TIME_5: responseTime + DELAY_TIME_MS * 5,
        CURRENT_TIME_6: responseTime + DELAY_TIME_MS * 6,
        CURRENT_TIME_7: responseTime + DELAY_TIME_MS * 7,
        CURRENT_TIME_8: responseTime + DELAY_TIME_MS * 8,
        CURRENT_TIME_9: responseTime + DELAY_TIME_MS * 9,
        CURRENT_TIME_10: responseTime + DELAY_TIME_MS * 10,
        INSTANCE_ID: INSTANCE_ID,
        PRIVATE_IP: PRIVATE_IP,
        REGION: REGION,
        AVAILABILITY_ZONE: AVAILABILITY_ZONE,
        VPC_ID: VPC_ID,
        VPC_CIDR_BLOCK: VPC_CIDR_BLOCK,
        SUBNET_ID: SUBNET_ID,
        SUBNET_CIDR_BLOCK: SUBNET_CIDR_BLOCK,
        VERSION: global.__VERSION__,
        HOST_NAME: os.hostname(),
        PLATFORM: `${os.type()} ${os.release()}`
    }));

    runScenario(result, canAbort);
}

AWS.mock('SQS', 'sendMessage', params  => ({
    promise: async() => {
        const messageBody = JSON.parse(params.MessageBody);
        const attributes = params.MessageAttributes;
        const dataBrokerId = messageBody.header ? messageBody.header.dataBrokerId : attributes.dataBrokerId.StringValue;
        const {payload} = messageBody;
        const commandToHandle = messageBody.command || (attributes.type ? attributes.type.StringValue : undefined);
        const dataBroker = await DataBroker.findOne({'_id': dataBrokerId}).exec();
        if (!dataBroker) {
            logger.debug(`DataBroker with id ${dataBrokerId} was not found`);
            return {MessageId: uuid.v4()};
        } else {
            const {name: dataBrokerName} = dataBroker;
            switch (commandToHandle) {
                case 'create-relationship-v2': {
                    const [sourceType] = Object.keys(payload.source);
                    const source = payload.source[sourceType];
                    const [targetType] = Object.keys(payload.target);
                    const target = payload.target[targetType];
                    const suffix = (dataBrokerName === 'asterixDataBrokerForInterruptedCopy') ? 'interrupted' :
                        ((FAILURE_ADDRESS_CREATE_REGEX.test(source.host || source.bucket || '') || FAILURE_ADDRESS_CREATE_REGEX.test(target.host || target.bucket || '')) ? 'failure' :
                            (SUCCESS_ADDRESS_CREATE_REGEX.test(source.host || source.bucket || '') || SUCCESS_ADDRESS_CREATE_REGEX.test(target.host || target.bucket || '')) ? 'success' :
                                'completed');
                    const scenarioFileName = `scenarios/create-relationship-${suffix}.json`;
                    parseJsonAndRun(scenarioFileName, dataBrokerId, payload.relationshipId, messageBody, true);
                }
                    break;
                case 'sync-relationship': {
                    const {relationshipId} = payload;
                    const relationship = await Relationship.findOne({'_id': relationshipId}).exec();
                    const {source, target} = relationship;
                    const suffix = (dataBrokerName === 'asterixDataBrokerForInterruptedSync') ? 'interrupted' :
                        ((FAILURE_ADDRESS_SYNC_REGEX.test(source) || FAILURE_ADDRESS_SYNC_REGEX.test(target)) ? 'failure' : 'success');
                    const ifFromFileSystem = source.protocol === 'nfs' || source.protocol === 'cifs';
                    const scenarioFileName = (suffix === 'failure') ? `scenarios/sync-${suffix}.json` :
                        (ifFromFileSystem) ? `scenarios/sync-fileSystem-to-storage-${suffix}.json` : `scenarios/sync-storage-to-fileSystem-${suffix}.json`;
                    parseJsonAndRun(scenarioFileName, dataBrokerId, relationshipId, messageBody, true);
                }
                    break;
                case 'startup': {
                    if (dataBrokerName === 'asterixDataBrokerForInterruptedCopy' || dataBrokerName === 'asterixDataBrokerForInterruptedSync') {
                        const scenarioFileName = 'scenarios/startup-relationship.json';
                        payload.relationships.map(relationship => {
                            parseJsonAndRun(scenarioFileName, dataBrokerId, relationship.relationshipId, messageBody);
                        });
                    }
                }
                    break;
                case 'metadata': {
                    const {targetDataBrokerId} = messageBody;
                    messagesQueue.push(createSQSMessage(dataBrokerId, 'metadata', {targetDataBrokerId, metadata : {}}));
                }
                    break;
                case 'update':
                case 'delete-relationship':
                    // do nothing
                    break;
                case 'list-cifs-shares-v2': {
                    const {payload: {host}} = messageBody;
                    const messageToSend = buildListCifsSharesResponse(commandToHandle, dataBrokerId, host, (dataBrokerName === 'asterixEmptyDataBroker'));
                    messagesQueue.push(messageToSend);
                }
                    break;
                case 'list-cifs-share-folders-v2': {
                    const {payload: {host, share, path}} = messageBody;
                    messagesQueue.push(buildListCifsShareFoldersResponse(commandToHandle, dataBrokerId, host, share, path));
                }
                    break;
                case 'list-nfs-exports-v2': {
                    const { payload : {host}} = messageBody;
                    messagesQueue.push(buildListNfsExportsResponse(dataBrokerId, host,
                        (dataBrokerName === 'asterixEmptyDataBroker')));
                }
                    break;
                case 'list-nfs-export-folders-v2': {
                    const {payload :{host, export: exportName, path}} = messageBody;
                    messagesQueue.push(buildListNfsExportFoldersResponse(dataBrokerId, host, exportName, path,
                        (dataBrokerName === 'asterixEmptyDataBroker')));
                }
                    break;
                case 'list-s3-buckets-v2': {
                    messagesQueue.push(buildListS3BucketsResponse(dataBrokerId, (dataBrokerName === 'asterixEmptyDataBroker')));
                }
                    break;
                case 'list-s3-bucket-folders-v2': {
                    const {payload: {bucket, prefix}} = messageBody;
                    messagesQueue.push(buildListS3BucketFoldersResponse(dataBrokerId, bucket, prefix,
                        (dataBrokerName === 'asterixEmptyDataBroker')));
                }
                    break;
                case 'list-sgws-buckets-v2': {
                    const {payload: {host, port}} = messageBody;
                    messagesQueue.push(buildListSgwsBucketsResponse(dataBrokerId, host, port,
                        (dataBrokerName === 'asterixEmptyDataBroker')));
                }
                    break;
                case 'list-sgws-bucket-folders-v2': {
                    const {payload : {bucket, host, port, prefix}} = messageBody;
                    messagesQueue.push(buildListSgwsBucketFoldersResponse(dataBrokerId, host, port, bucket, prefix));
                }
                    break;
                case 'abort-relationship' : {
                    const {relationshipId} = payload;
                    const relationship = await Relationship.findOne({'_id': relationshipId}).exec();
                    const {type} = relationship.findTypeAndActiveMessage();
                    relationshipsToAbort.push(relationshipId);
                    const scenarioFileName = (type === 'Sync') ? 'scenarios/abort-sync-relationship-completed.json' : 'scenarios/abort-copy-relationship-completed.json';
                    parseJsonAndRun(scenarioFileName, dataBrokerId, relationship._id, messageBody);
                }
                    break;
                case 'exports': {
                    const {host} = messageBody;
                    messagesQueue.push(buildListNfsExportsResponseNewDataBroker(host, attributes,
                        (dataBrokerName === 'EmptyDataBroker')));
                }
                    break;
                case 'export-folders': {
                    const { host, export: exportName, path} = messageBody;
                    messagesQueue.push(buildListNfsExportFoldersResponseNewDataBroker(host, exportName, path, attributes,
                        (dataBrokerName === 'EmptyDataBroker')));
                }
                    break;
                case 'shares': {
                    const {host} = messageBody;
                    const messageToSend = buildListCifsSharesResponseNewDataBroker(host, (dataBrokerName === 'EmptyDataBroker'), attributes);
                    messagesQueue.push(messageToSend);
                }
                    break;
                case 'share-folders': {
                    const {host, share, path} = messageBody;
                    messagesQueue.push(buildListCifsShareFoldersResponseNewDataBroker(attributes, host, share, path));
                }
                    break;
                case 'buckets': {
                    messagesQueue.push(buildListS3BucketsResponseNewDataBroker(attributes,
                        (dataBrokerName === 'EmptyDataBroker')));
                }
                    break;
                case 'bucket-folders' : {
                    const {bucket, prefix} = messageBody;
                    messagesQueue.push(buildListS3BucketFoldersResponseNewDataBroker(attributes, bucket, prefix,
                        (dataBrokerName === 'EmptyDataBroker')));
                }
                    break;
                case 'sgws-buckets': {
                    const {host, port} = messageBody;
                    messagesQueue.push(buildListSgwsBucketsResponseNewDataBroker(attributes, host, port, (dataBrokerName === 'EmptyDataBroker')));
                }
                    break;
                case 'sgws-buckets-folders' : {
                    const {bucket, prefix, host, port} = messageBody;
                    messagesQueue.push(buildListSgwsBucketFoldersResponseNewDataBroker(attributes, host, port, bucket, prefix,
                        (dataBrokerName === 'EmptyDataBroker')));
                }
                    break;
                case 'azure-blob-storage-containers': {
                    const {storageAccountName} = messageBody;
                    messagesQueue.push(buildListAzureContainersResponse(attributes, storageAccountName,
                        dataBrokerName === 'EmptyDataBroker'));
                }
                    break;
                case 'azure-blob-storage-container-folders' : {
                    const {container, prefix, storageAccountName} = messageBody;
                    messagesQueue.push(buildListAzureContainerFoldersResponse(attributes, container, prefix, storageAccountName,
                        dataBrokerName === 'EmptyDataBroker'));
                }
                    break;
                case 'gcp-buckets': {
                    const {projectId} = messageBody;
                    messagesQueue.push(buildListGcpBucketsResponse(attributes, projectId,
                        dataBrokerName === 'EmptyDataBroker'));
                }
                    break;
                case 'gcp-bucket-folders' : {
                    const {bucket, prefix, projectId} = messageBody;
                    messagesQueue.push(buildListGcpBucketFoldersResponse(attributes, bucket, prefix, projectId,
                        dataBrokerName === 'EmptyDataBroker'));
                }
                    break;
                case 'copy': {
                    if (messageBody && messageBody.type !== SCAN_MODE_SOURCE_AND_TARGET) {
                        const [sourceType] = Object.keys(messageBody.source);
                        const source = messageBody.source[sourceType];
                        const [targetType] = Object.keys(messageBody.target);
                        const target = messageBody.target[targetType];
                        const { scannerQueue,  transferrerQueue, relationshipId } = messageBody;
                        const relationship = await Relationship.findOne({'_id' : relationshipId, 'phase' : 'Sync'}).exec();
                        const isRelationshipInSync = Boolean(relationship);
                        const isCopyFailure = (FAILURE_ADDRESS_CREATE_REGEX.test(source.host || source.bucket || source.container || '') && !isRelationshipInSync ||
                            FAILURE_ADDRESS_CREATE_REGEX.test(target.host || target.bucket || target.container || '') && !isRelationshipInSync);
                        const isSyncFailure = (FAILURE_ADDRESS_SYNC_REGEX.test(source.host || source.bucket || source.container || '') &&
                            isRelationshipInSync || FAILURE_ADDRESS_SYNC_REGEX.test(target.host || target.bucket || target.container || '') && isRelationshipInSync);
                        const isSuccess = (SUCCESS_ADDRESS_CREATE_REGEX.test(source.host || source.bucket || source.container || '')
                            || SUCCESS_ADDRESS_CREATE_REGEX.test(target.host || target.bucket || target.container || ''));
                        const prefix = isRelationshipInSync ? 'syncPhoenix' : 'copy';
                        const suffix = (isCopyFailure || isSyncFailure) ? 'failure' : (isSuccess ? 'success' : 'completed');
                        const scenarioFileName = `scenarios/${prefix}-${suffix}.json`;
                        const scenarioFile = path.resolve(__dirname, scenarioFileName);
                        if (!fs.existsSync(scenarioFile)) {
                            logger.error(`Scenario file '${scenarioFileName}' not found to handle:`, messageBody);
                            return;
                        }
                        const scenario = fs.readFileSync(scenarioFile, 'utf-8');
                        const result =  JSON.parse(format(scenario, {
                            RELATIONSHIP_ID: relationshipId
                        }));
                        const runningRelationshipsKey = `${scannerQueue}_${transferrerQueue}`;
                        runningRelationshipsNewDataBroker[runningRelationshipsKey] = result;
                        const isUnknown = dataBroker.name === 'Unknown';
                        runScenarioNewDataBroker(runningRelationshipsKey, attributes, true, isUnknown);
                    }
                }
                    break;
                case 'abort': {
                    const {relationshipId } = messageBody;
                    const relationship = await Relationship.findById(relationshipId).exec();
                    relationshipsToAbort.push(relationshipId);
                    const {scannerQueue, transferrerQueue} = relationship;
                    const runningRelationshipsKey = `${scannerQueue}_${transferrerQueue}`;
                    delete runningRelationshipsNewDataBroker[runningRelationshipsKey];
                }
                    break;
                case 'done':
                case 'shutdown':
                case 'remove': {
                    //nothing to do
                }
                    break;
                default:
                    logger.error('Got unexpected SQS message', messageBody);
            }
            return {MessageId: uuid.v4()};
        }
    }
}));
