import fs from 'fs';
import path from 'path';
import moment from 'moment';
import User from '../models/user';
import Cost from '../models/cost';
import {connectAndMigrate} from '../db-connect-and-migrate';
import { initAgenda } from './agenda';
import { startListeningDataBrokersStatuses } from './sqs-listener';
import { startMPSubscriptionStatusListener } from './mp-subscription-status-listener';
import { isProduction } from '../../../environment';
import { TRIAL_EXPIRATION_MS, COST_RECORDS_SIMULATOR_FILE, FAKE_USERS_FILE } from '../utils/consts';
import { relationshipsStatusPoller } from '../utils/common';
import { getDataBrokerLatestVersionAsync } from './update-data-broker';

import {getLogger} from '../utils/logger';
const logger = getLogger();
const currentTime = Date.now();
const defaultTrialExpiry = new Date(currentTime + TRIAL_EXPIRATION_MS);

async function createFakeUser(fakeUser) {
    const {id, name, email, portalId} = fakeUser;
    logger.debug(`Creating '${name}' user if needed`);
    const user = await User.findOne({email}).exec();
    if (user) {
        logger.debug(`'${name}' user already exist`);
        return user;
    } else {
        logger.debug(`'${name}' user not found. Creating...`);
        let company;
        let createdAt;
        if (email.includes('occmfakeuser@') || email.includes('trialexpiryfakeuser@') || email.includes('occmfakeuser2@')) {
            company = 'OCCM';
            createdAt = new Date('1/1/2016 00:00').valueOf();
        } else {
            company = 'NetApp';
            createdAt = new Date().valueOf();
        }
        let trialExpiry;
        if (email.includes('occmfakeuser@occm.com')) {
            trialExpiry = new Date(currentTime + moment.duration(5, 'days').as('milliseconds'));
        } else if (email.includes('occmfakeuser2@')) {
            trialExpiry = new Date(currentTime + moment.duration(10, 'days').as('milliseconds'));
        } else if (email.includes('trialexpiryfakeuser@')) {
            trialExpiry = new Date('2/1/2016 00:00').valueOf();
        } else {
            trialExpiry = defaultTrialExpiry;
        }
        let awsAccountId = '282316784512';
        if (name === 'OCCM Fake User' || name === 'OCCM Fake User 2') {
            awsAccountId = '123456789011';
        }
        return new User({createdAt, _id: id, name, email, company, awsAccountId, trialExpiry, portalId}).save();
    }
}

async function createCostRecord(cost) {
    const {id, user, timestamp, tier1, tier2, tier3, byol} = cost;
    const foundCost = await Cost.findOne({_id: id}).exec();
    if (foundCost) {
        logger.debug(`'${id}' cost record is already exist`);
        return foundCost;
    } else {
        logger.debug(`'${id}' cost record was not found. Creating...`);
        return new Cost({_id: id, user, timestamp, tier1, tier2, tier3, byol}).save();
    }
}

async function createFakeUsers() {
    if (!isProduction()) {
        const fakeUsers = JSON.parse(fs.readFileSync(path.resolve(__dirname, FAKE_USERS_FILE), 'utf8'));
        await Promise.all([createFakeUser(fakeUsers.find(user => user.email === 'trialexpiryfakeuser@test.com')),
            createFakeUser(fakeUsers.find(user => user.email === 'occmfakeuser@occm.com')),
            createFakeUser(fakeUsers.find(user => user.email === 'occmfakeuser2@occm.com'))]);
    }
}

async function createCostRecords() {
    if (!isProduction()) {
        const costRecords = JSON.parse(fs.readFileSync(path.resolve(__dirname, COST_RECORDS_SIMULATOR_FILE), 'utf8'));
        return Promise.all([costRecords.map(createCostRecord)]);
    }
}

export default async function() {
    try {
        logger.info('Booting...');
        await connectAndMigrate();
        await initAgenda();
        await Promise.all([createFakeUsers(), createCostRecords(), startListeningDataBrokersStatuses(),
            startMPSubscriptionStatusListener(), relationshipsStatusPoller(), getDataBrokerLatestVersionAsync()]);
        logger.info('Booting done');
    } catch (err) {
        logger.error('Bootstrapping application failed with:', err);
        throw err;
    }
}
