import config from 'config';
import ms from 'ms';
import retry from 'async-retry';
import Message from '../models/message';
import Timeline from '../models/timeline';
import TimelineV2 from '../models/timeline-v2';
import DataBroker from '../models/data-broker';
import RelationshipStats from '../models/relationship-statistics';
import {get, isNumber, keys, last, sortBy, slice, find, isEmpty, first, set} from 'lodash';
import {
    AWS_REGIONS,
    GCP_REGIONS,
    MESSAGE_STATUS_FAILED,
    MESSAGE_STATUS_DONE,
    RUNNING_STATUS,
    MESSAGE_STATUS_INTERRUPTED,
    INTERRUPTED_MESSAGE,
    TIMELINE_DELETED_RELATIONSHIP_MESSAGE,
    FAILED_AFTER_STARTUP_MESSAGE,
    CONTRACT_SUBSCRIPTION,
    COPY_RELATIONSHIP_MESSAGE,
    SYNC_RELATIONSHIP_MESSAGE,
    ABORT_RELATIONSHIP,
    STARTUP_RELATIONSHIP,
    MESSAGE_STATUS_ABORTING,
    MESSAGE_STATUS_ABORTED,
    ABORTED_MESSAGE,
    SERVER_HOST,
    ASTERIX_EXPIRY_DATE
} from './consts';
import {BadRequest} from '../utils/errors';
import {requestId, timeout} from '../utils/utils';
import {headBucket} from '../lib/s3';
import {getQueueAttributes, deleteQueue} from '../lib/sqs';
import Relationship from '../models/relationship';
import User from '../models/user';
import {sendDeleteCommand} from '../operations/relationship-operations';

import {getLogger} from '../utils/logger';
const Joi = require('joi');
const logger = getLogger();

const [environment] = SERVER_HOST.split('.');

export function assertIpOrHostname(ipOrHostname, protocol) {
    if (!ipOrHostname) {
        throw new BadRequest(`source or target with protocol ${protocol} must contain ip or host`);
    }
    const hostNameSchema = Joi.string().hostname();
    const ipSchema = Joi.string().ip();
    const {error : ipHostError} = Joi.validate(ipOrHostname, ipSchema);
    const {error: hostNameError} = Joi.validate(ipOrHostname, hostNameSchema);
    if (ipHostError && hostNameError) {
        throw new BadRequest(`Invalid ${protocol} ip or hostname ${ipOrHostname}`);
    }

}

export function awsRegionCodeToDisplayName(region) {
    return AWS_REGIONS[region] || region;
}

export function gcpRegionCodeToDisplayName(region) {
    return GCP_REGIONS[region] || region;
}

export async function updateSubscription(subscription, users) {
    if (subscription) {
        const updateRule = {$set: {'subscription': subscription}};
        if (subscription.type === CONTRACT_SUBSCRIPTION) {
            updateRule.$set.trialExpiry = new Date();
        }
        if (users.length !== 0) {
            const usersIds = users.map(user => {
                return {'_id': user._id};
            });
            return User.updateMany({$and: [{$or: usersIds}, {'subscription.type': {$not: new RegExp(CONTRACT_SUBSCRIPTION, 'i')}}]}, updateRule).exec();
        }
    }
}

export function isNewDataBroker(dataBroker) {
    const {name, createdAt} = dataBroker;
    return  (name.startsWith('phoenix') && new Date(createdAt) < new Date(ASTERIX_EXPIRY_DATE)) || new Date(createdAt) > new Date(ASTERIX_EXPIRY_DATE);
}

export function createStringFromEndPointObject(object) {
    if (object) {
        const {protocol} = object;
        const details = object[protocol];
        let objectString;
        switch (protocol) {
            case 'sgws': {
                const {host, port, bucket, objectStorageType} = details;
                let {prefix} = details;
                const stringProtocol = objectStorageType || 'sgws';
                objectString = (port) ? `${stringProtocol}://${host}:${port}` : `${stringProtocol}://${host}`;
                prefix = (prefix) ? (prefix.startsWith('/') ? prefix.substr(1) : prefix) : prefix;
                objectString = (prefix) ? `${objectString}/${bucket}/${prefix}` : `${objectString}/${bucket}`;
                break;
            }
            case 'cifs': {
                const {host} = details;
                let {path, share} = details;
                path = (path) ? (path.startsWith('/') ? path.substr(1) : path) : path;
                share = share.startsWith('/') ? share.substr(1) : share;
                objectString = (path) ? `cifs://${host}/${share}/${path}` : `cifs://${host}/${share}`;
                break;
            }
            case 'nfs': {
                const {host, nfsType} = details;
                let {path, export: nfsExport} = details;
                const stringProtocol = nfsType || 'nfs';
                path = (path) ? (path.startsWith('/') ? path.substr(1) : path) : path;
                if (stringProtocol === 'efs') {
                    objectString = (path) ? `efs://${host}/${path}` : `efs://${host}/`;
                } else {
                    nfsExport = nfsExport.startsWith('/') ? nfsExport.substr(1) : nfsExport;
                    objectString = (path) ? `${stringProtocol}://${host}/${nfsExport}/${path}` : `${stringProtocol}://${host}/${nfsExport}`;
                }
                break;
            }
            case 's3': {
                const {bucket} = details;
                let {prefix} = details;
                prefix = (prefix) ? (prefix.startsWith('/') ? prefix.substr(1) : prefix) : prefix;
                objectString = (prefix) ? `s3://${bucket}/${prefix}` : `s3://${bucket}`;
                break;
            }
            case 'azure': {
                const {container, storageAccountName} = details;
                let {prefix} = details;
                prefix = (prefix) ? (prefix.startsWith('/') ? prefix.substr(1) : prefix) : prefix;
                objectString = (prefix) ? `azure://${storageAccountName}@${container}/${prefix}` : `azure://${storageAccountName}@${container}`;
                break;
            }
            case 'local': {
                const {path} = details;
                objectString = `local://${path}`;
                break;
            }
            case 'gcp': {
                const {bucket} = details;
                let {prefix} = details;
                prefix = (prefix) ? (prefix.startsWith('/') ? prefix.substr(1) : prefix) : prefix;
                objectString = (prefix) ? `gcp://${bucket}/${prefix}` : `gcp://${bucket}`;
                break;
            }
        }
        return objectString;
    }
}

export function findTimelineRecordV2(relationship, isAbort) {
    const {user, id: relationshipId} = relationship;
    const filter = {'relationship.relationshipId': relationshipId, status: {$nin: ['DONE', 'ABORTED', 'FAILED']}};
    if (isAbort) {
        filter.summary = 'Abort relationship data sync';
    } else {
        filter.summary = {$ne: 'Abort relationship data sync'};
    }
    return TimelineV2.findOne(filter).hint({
        'relationship.relationshipId': 1,
        'status': 1,
        'summary': 1
    }).populate('user').sort({updatedAt: -1}).limit(1).exec()
        .then(timeLineResult => {
            if (!timeLineResult) {
                logger.info(`TimeLine action with user ${user} and relationshipId ${relationshipId} was not found`);
            } else {
                return timeLineResult;
            }
        });
}

export function findTimelineRecord(relationship, isAbort) {
    const {user, source: sourceFromDb, target: targetFromDb, dataBroker: {name}} = relationship;
    const source = createStringFromEndPointObject(sourceFromDb);
    const target = createStringFromEndPointObject(targetFromDb);
    const filter = (isAbort) ? {
        user,
        source,
        target,
        dataBroker: name,
        summary: 'Abort relationship data sync',
        status: 'PROGRESS'
    } :
        {
            $or: [{
                user,
                source,
                target,
                dataBroker: name,
                summary: 'Initiate relationship data sync',
                status: 'PROGRESS',
            },
            {
                user,
                source,
                target,
                dataBroker: name,
                summary: 'Initiate relationship data sync',
                status: 'ABORTING',
            },
            {user, source, target, dataBroker: name, summary: 'Scheduled sync', status: 'PROGRESS'},
            {user, source, target, dataBroker: name, summary: 'Scheduled sync', status: 'ABORTING'},
            {
                user,
                source,
                target,
                dataBroker: name,
                summary: 'Create a new Relationship',
                status: 'PROGRESS'
            },
            {
                user,
                source,
                target,
                dataBroker: name,
                summary: 'Create a new Relationship',
                status: 'ABORTING'
            }]
        };
    return Timeline.findOne(filter).populate('user').sort({updatedAt: -1}).limit(1).exec()
        .then(timeLineResult => {
            if (!timeLineResult) {
                logger.info(`TimeLine action with user ${user} and source ${source} and target ${target} was not found and isAbort flag ${isAbort}`);
            } else {
                return timeLineResult;
            }
        });
}

function updateTimeLineV2Schema(status, failureMessage, timeLineResult, activity, isAbort) {
    const progressStatuses = [RUNNING_STATUS, 'STARTED'];
    timeLineResult.failureMessage = failureMessage;
    timeLineResult.status = progressStatuses.includes(status) ? 'PROGRESS' : status;
    if (!isAbort) {
        timeLineResult.relationship.activity = activity;
    }
    return timeLineResult.save(timeLineResult);
}

function updateTimeLineSchema(status, failureMessage, timeLineResult, activity) { //To delete when timeline-v1 is no longer relevant- there is a different function on propose
    const progressStatuses = [RUNNING_STATUS, 'STARTED'];
    timeLineResult.failureMessage = failureMessage;
    timeLineResult.status = progressStatuses.includes(status) ? 'PROGRESS' : status;
    if (!timeLineResult.summary.includes('Abort relationship data sync')) {
        timeLineResult.activity = activity;
    }
    return timeLineResult.save(timeLineResult);
}

export function getUserAndPolicyName(dataBroker) {
    return `CS_${environment}_${dataBroker.id}`;
}

//phoenix

export async function calculateRelationshipStatusV2(relationshipId, user) {
    const statusPerRelationship = await RelationshipStats.findRelationshipStatus(relationshipId, user.id);
    if (isEmpty(statusPerRelationship)) { //there is an option there will not be a status - in case a data broker joined in the middle of a relationship
        return {status: RUNNING_STATUS};
    } else {
        const statusesScores = {
            [RUNNING_STATUS]: {status: RUNNING_STATUS, score: 4},
            [MESSAGE_STATUS_FAILED]: {status: MESSAGE_STATUS_FAILED, score: 5},
            [MESSAGE_STATUS_DONE]: {status: MESSAGE_STATUS_DONE, score: 1},
            [MESSAGE_STATUS_ABORTED]: {status: MESSAGE_STATUS_ABORTED, score: 2}
        };
        const {statuses: statusArr} = first(statusPerRelationship);
        const relationshipStatusesWithScores = sortBy(statusArr.map(statusObj => ({statusObj, score: statusesScores[statusObj.status]})), 'score');
        return first(relationshipStatusesWithScores).statusObj;
    }
}

export function calculateProgressNewDataBroker(status, bytesToTransfer, bytesTransferred, dirsToTransfer, dirsTransferred) {
    let progress;
    if (isNumber(bytesToTransfer) && isNumber(dirsToTransfer)) {
        if (bytesToTransfer === 0 && dirsToTransfer === 0) {
            progress = (status === MESSAGE_STATUS_DONE) ? 100 : 0;
        } else {
            progress = Math.round((bytesTransferred + dirsTransferred) / (bytesToTransfer + dirsToTransfer) * 100);
            if (progress >= 100) {
                if (status !== MESSAGE_STATUS_DONE && status !== MESSAGE_STATUS_FAILED) {
                    progress = 99;
                } else {
                    progress = (status === MESSAGE_STATUS_DONE) ? 100 : 0;
                }
            } else {
                progress = (status === MESSAGE_STATUS_DONE) ? 100 : ((status === MESSAGE_STATUS_FAILED) ? 0 : progress);
            }
        }
    }
    return progress;
}


export async function createActivity(relationship) {
    const relationshipJson = relationship.toJSON();
    const relationshipsStats = await RelationshipStats.sumStatsPerRelationship(relationship.id, relationship.user);
    const relationshipStats = first(relationshipsStats);
    const bytesMarkedForCopy = get(relationshipStats, 'bytesMarkedForCopy', 0);
    const bytesMarkedForRemove = get(relationshipStats, 'bytesMarkedForRemove', 0);
    const bytesCopied = get(relationshipStats, 'bytesCopied', 0);
    const bytesRemoved = get(relationshipStats, 'bytesRemoved', 0);
    const bytesFailed = get(relationshipStats, 'bytesFailed', 0);
    const {status, reason} = await calculateRelationshipStatusV2(relationship.id, relationship.user);
    const filesCopied = get(relationshipStats, 'filesCopied', 0) + get(relationshipStats, 'symbolicCopied', 0);
    const filesFailed = get(relationshipStats, 'filesFailed', 0) + get(relationshipStats, 'symbolicFailed', 0);
    const filesMarkedForCopy = get(relationshipStats, 'filesMarkedForCopy', 0) + get(relationshipStats, 'symbolicMarkedForCopy', 0);
    const filesMarkedforRemove = get(relationshipStats, 'filesMarkedForRemove', 0) + get(relationshipStats, 'symbolicMarkedForRemove', 0);
    const filesRemoved = get(relationshipStats, 'filesRemoved', 0) + get(relationshipStats, 'symbolicRemoved', 0);
    const filesRemovedFailed = get(relationshipStats, 'filesRemovedFailed', 0) + get(relationshipStats, 'symbolicRemovedFailed', 0);
    const filesMarkedForGrace = get(relationshipStats, 'filesMarkedForGrace', 0) + get(relationshipStats, 'symbolicMarkedForGrace', 0);
    const dirsCopied = get(relationshipStats, 'dirsCopied', 0);
    const dirsMarkedForCopy = get(relationshipStats, 'dirsMarkedForCopy', 0);
    const dirsScanned = get(relationshipStats, 'dirsScanned', 0);
    const filesScanned = get(relationshipStats, 'filesScanned', 0) + get(relationshipStats, 'symbolicScanned', 0);
    const bytesScanned = get(relationshipStats, 'bytesScanned', 0);
    const dirsFailedToScan = get(relationshipStats, 'dirsFailedToScan', 0);

    relationshipJson.activity = {
        type: relationship.phase,
        status,
        failureMessage: (reason) ? reason : '',
        executionTime: (relationship.endTime) ? (relationship.endTime - relationship.startTime) : (Date.now() - relationship.startTime),
        startTime: relationship.startTime,
        endTime: (relationship.endTime) ? relationship.endTime : Date.now(),
        bytesMarkedForCopy,
        filesMarkedForCopy,
        dirsMarkedForCopy,
        filesCopied,
        bytesCopied,
        dirsCopied,

        filesFailed,
        bytesFailed,
        dirsFailed: get(relationshipStats, 'dirsFailed', 0),

        filesMarkedforRemove,
        bytesMarkedForRemove,
        dirsMarkedForRemove: get(relationshipStats, 'dirsMarkedForRemove', 0),
        filesRemoved,
        bytesRemoved,
        dirsRemoved: get(relationshipStats, 'dirsRemoved', 0),

        bytesRemovedFailed: get(relationshipStats, 'bytesRemovedFailed', 0),
        filesRemovedFailed,

        filesMarkedForGrace,
        bytesMarkedForGrace: get(relationshipStats, 'bytesMarkedForGrace', 0),
        dirsMarkedForGrace: get(relationshipStats, 'dirsMarkedForGrace', 0),
        filesMarkedForIgnore: get(relationshipStats, 'filesMarkedForIgnore', 0),

        dirsScanned,
        filesScanned,
        dirsFailedToScan,
        bytesScanned,

        progress: calculateProgressNewDataBroker(status,
            (bytesMarkedForCopy + bytesMarkedForRemove),
            (bytesCopied + bytesRemoved + bytesFailed), dirsMarkedForCopy, dirsCopied),
        lastMessageTime: get(relationshipStats, 'lastMessageTime', 0)

    };
    return relationshipJson;
}


export async function sendNewDataBrokerRelationshipMessage(dataBroker, relationshipResponse, requestId, user) {
    const rId = (requestId) ? requestId : requestId();
    const clientId = (user) ? user._id : dataBroker.user.id;
    return Message.QueueAsync(clientId, {
        header: {
            dataBrokerId: dataBroker.id,
            relationshipId: relationshipResponse.id,
            requestId: rId
        },
        payload: {
            'update-relationship-v2': relationshipResponse
        }
    });
}


export async function updateTimeLineNewDataBroker(relationship, relationshipResponse, action, findTimeLineRecordFunc, activityKey) {
    const progressStatuses = [RUNNING_STATUS, 'STARTED'];
    try {
        const timeLineResultRelationship = await findTimeLineRecordFunc(relationship);
        if (timeLineResultRelationship) {
            switch (action) {
                case MESSAGE_STATUS_ABORTED: {
                    timeLineResultRelationship.status = get(relationshipResponse, 'activity.status', '');
                    timeLineResultRelationship.failureMessage = get(relationshipResponse, 'activity.failureMessage');
                    set(timeLineResultRelationship, activityKey, relationshipResponse.activity);

                    //there will always be an abort record
                    const timeLineResultAbort = await findTimeLineRecordFunc(relationship, true);
                    timeLineResultAbort.status = MESSAGE_STATUS_DONE;
                    timeLineResultAbort.save();
                }
                    break;
                case 'DELETE': {
                    timeLineResultRelationship.status = MESSAGE_STATUS_FAILED;
                    timeLineResultRelationship.failureMessage = TIMELINE_DELETED_RELATIONSHIP_MESSAGE;
                    set(timeLineResultRelationship, activityKey, relationshipResponse.activity);

                    const timeLineResultAbort = await findTimeLineRecordFunc(relationship, true);
                    if (timeLineResultAbort) {
                        timeLineResultAbort.status = MESSAGE_STATUS_FAILED;
                        timeLineResultAbort.failureMessage = TIMELINE_DELETED_RELATIONSHIP_MESSAGE;
                        timeLineResultAbort.save();
                    }
                }
                    break;
                default : {
                    const statusFromRelationship = get(relationshipResponse, 'activity.status');
                    timeLineResultRelationship.status = progressStatuses.includes(statusFromRelationship) ? 'PROGRESS' : statusFromRelationship;
                    timeLineResultRelationship.failureMessage = get(relationshipResponse, 'activity.failureMessage');
                    set(timeLineResultRelationship, activityKey, relationshipResponse.activity);
                }
            }
            await timeLineResultRelationship.save();
        }

    } catch (e) {
        logger.error(`An error occurred in updateTimeLineNewDataBroker function trying to update 
        relationship ${relationship.id} with message ${JSON.stringify(relationshipResponse.activity.status)}. Error ${e}`);
        throw e;
    }
}

//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> both asterix and phoenix

export async function getRelationshipResponse(relationship, dataBroker) {
    if (relationship.group || isNewDataBroker(dataBroker)) {
        return createActivity(relationship);
    } else {
        return relationship.toJSON();
    }
}

//>>>...phoenix continue

export async function updateRelationshipAndTimeLine(relationship, requestId, status) {
    const relationshipFromDb = await Relationship.findById(relationship.id).populate('user dataBroker group').populate({
        path: 'group',
        populate: {path: 'dataBrokers'}
    }).exec();
    const relationshipResponse = await getRelationshipResponse(relationshipFromDb);
    await sendNewDataBrokerRelationshipMessage(relationship.dataBroker, relationshipResponse, requestId, relationship.user);
    await Promise.all([updateTimeLineNewDataBroker(relationship, relationshipResponse, status, findTimelineRecord, 'activity'),
        updateTimeLineNewDataBroker(relationship, relationshipResponse, status, findTimelineRecordV2, 'relationship.activity')]);
    return relationship;
}

//update all records related to relationship  and dataBrokers of relationship in relationshipStats
export async function updateStatusForRelationship(relationship, status, reason) {
    const {group, dataBroker, id: relationshipId, user} = relationship;
    if (status === RUNNING_STATUS) { //RUNNING status in order to create a document for each broker - if one is not created it is not so bad - it will be created when the statistics or failures will come
        const mastersToUpdate = (group) ? group.dataBrokers : [dataBroker];
        const promises = mastersToUpdate.map(async dataBroker => RelationshipStats.create({relationship: relationshipId, dataBroker, user, status :RUNNING_STATUS}));
        await Promise.all(promises);
    } else { //avoid  failing to update one of the documents
        await RelationshipStats.updateMany({relationship: relationshipId, user, status :{ $nin : [MESSAGE_STATUS_DONE, MESSAGE_STATUS_ABORTED,
            MESSAGE_STATUS_FAILED]}}, {$set: {status, reason}});
    }
}

async function createNfsEndPoint(details, user) {
    const {host, export: exportPath, path, version, nfsType, dataBrokerId } = details;
    const result = {host, export: exportPath, version};
    if (path !== '') {
        result.path = path;
    }
    if (nfsType === 'snfs' && dataBrokerId) {
        const listenerDatabroker = await DataBroker.findOne({_id: dataBrokerId, user }).exec();
        result.serverName = dataBrokerId;
        result.serverIP = listenerDatabroker.privateIp;
        result.port = listenerDatabroker.encryption.port;
        result.certificate = listenerDatabroker.encryption.certificate;
    } else if (nfsType === 'efs') {
        result.version = '4.1';
    }
    return result;
}

export async function createEndPointForNewDataBroker(endPointObject, credentials, additionalArguments) {
    let result;
    let changeProtocolToSnfs = false;
    const {protocol} = endPointObject;
    const details = endPointObject[protocol];
    const convertTags = (tags) => {
        return tags ? tags.map(tag => ({Key: tag.key, Value: tag.value})) : [];
    };
    switch (protocol) {
        case 'nfs': {
            const { user } = additionalArguments;
            result = await createNfsEndPoint(details, user);
            if (details.nfsType === 'snfs' && details.dataBrokerId) {
                changeProtocolToSnfs = true;
            }
            break;
        }
        case 's3': {
            const {bucket, prefix, encryption, tags: s3Tags, storageClass} = details;
            const region = await headBucket(bucket);
            result = {bucket, region, tags: convertTags(s3Tags), storageClass};
            if (prefix !== '') {
                result.path = prefix;
            }
            if (encryption && encryption.protocol === 'kms') {
                result.encryption = { 'kms': { keyId: encryption.keyId.trim() } };
            }
            break;
        }
        case 'local': {
            result = details;
            break;
        }
        case 'azure': {
            const {container, storageAccountName, prefix, tags: azureTags, blobTier} = details;
            const storageAccountConnectionString = credentials ? credentials.azure.storageAccountConnectionString : undefined;
            result = {container, storageAccountName, storageAccountConnectionString, tags: convertTags(azureTags), blobTier};

            if (prefix !== '') {
                result.path = prefix;
            }

            break;
        }
        case 'cifs': {
            const {host, share, path, version, aclServer} = details;

            result = {host, share, version, aclServer};
            if (path !== '') {
                result.path = path;
            }

            if (credentials) {
                result.username = credentials.cifs.username.trim();
                result.password = credentials.cifs.password;
                result.workgroup = credentials.cifs.domain === '' ? undefined : credentials.cifs.domain;
            }

            break;
        }
        case 'sgws': {
            const {bucket, prefix:path, host, port, tags: sgwsTags} = details;
            result = {bucket, path, host, tags: convertTags(sgwsTags)};
            if (credentials) {
                result.accessKey = credentials.sgws.accessKey.trim();
                result.secretKey = credentials.sgws.secretKey.trim();
            }
            if (port) {
                result.port = port;
            }
            break;
        }
        case 'gcp': {
            const {projectId, bucket, prefix, tags: gcpTags} = details;
            result = {projectId, bucket, tags: convertTags(gcpTags)};
            if (credentials) {
                result.clientEmail = credentials.gcp.clientEmail.trim();
                result.privateKey = credentials.gcp.privateKey.trim().replace(/\\n/g, '\n');
            }
            if (prefix !== '') {
                result.path = prefix;
            }
            break;
        }
    }
    return changeProtocolToSnfs ? {'snfs': result} : {[protocol]: result};
}

export async function findRelationshipsForNewDataBrokerStartup(group) {
    const filter = {
        scannerQueue: {$exists: true},
        transferrerQueue: {$exists: true},
        group:group._id
    };
    const relationships = await Relationship.find(filter).exec();
    const promises = relationships.map(async relationship => {
        return {
            relationshipId: relationship.id,
            source: await createEndPointForNewDataBroker(relationship.source, undefined, { user: relationship.user }),
            target: await createEndPointForNewDataBroker(relationship.target, undefined, { user: relationship.user }),
            scannerQueue: relationship.scannerQueue,
            transferrerQueue: relationship.transferrerQueue,
            settings: relationship.settings
        };
    });
    return Promise.all(promises);
}

export async function deleteRelationshipQueuesIfExist(relationship) {
    const {transferrerQueue, scannerQueue} = relationship;
    if (transferrerQueue && scannerQueue) {
        return Promise.all([
            deleteQueue(relationship.scannerQueue),
            deleteQueue(relationship.transferrerQueue),
            relationship.updateOne({$unset: {transferrerQueue: '', scannerQueue: ''}})
        ]);
    }
}

async function areEmptyQueues(relationshipId, scannerQueue, transferrerQueue) {
    const scannerAttributes = await getQueueAttributes(scannerQueue, ['ApproximateNumberOfMessages', 'ApproximateNumberOfMessagesNotVisible']);
    logger.info(`Scanner attribute of relationship ${relationshipId} are ${JSON.stringify(scannerAttributes)}`);

    if (scannerAttributes.ApproximateNumberOfMessages === '0' && scannerAttributes.ApproximateNumberOfMessagesNotVisible === '0') {
        await timeout(5000);
        const transferrerAttributes = await getQueueAttributes(transferrerQueue, ['ApproximateNumberOfMessages', 'ApproximateNumberOfMessagesNotVisible']);
        logger.info(`Transferrer attribute of relationship ${relationshipId} are ${JSON.stringify(transferrerAttributes)}`);
        return transferrerAttributes.ApproximateNumberOfMessages === '0' && transferrerAttributes.ApproximateNumberOfMessagesNotVisible === '0';
    }

    return false;
}

export async function handleEndOfRelationship(relationship, doneRequestId, commandType) {
    const {id: relationshipId} = relationship;
    //#1 send done command
    logger.info(`Sending ${commandType} command to relationship ${relationship.id}`);
    await  sendDeleteCommand(relationship, doneRequestId, commandType);

    //#3 delete relationship queues
    try {
        logger.info(`${commandType} command was sent for relationship ${relationshipId},  deleting relationship queue`);
        await retry(() => deleteRelationshipQueuesIfExist(relationship), {
            retries: 3,
            onRetry: err => logger.warn(`RETRYING deleting relationship queues of relationship ${relationshipId} with error ${JSON.stringify(err)}`)
        });
    } catch (err) {
        logger.error(`Failed deleting relationship queues for relationship ${relationshipId} with error`, err);
    }

    //#2 wait 30 seconds - in order to avoid new sync before umount was done
    await timeout(ms(config.get('end-of-relationship-timeout')));

    //update endTime of relationship
    relationship.endTime = Date.now();
    await relationship.save();
}

export async function relationshipsStatusPoller() {
    try {
        const relationships = await Relationship.find({
            scannerQueue: {$exists: true},
            transferrerQueue: {$exists: true},
            startTime: {$lte: (new Date()).getTime() - ms(config.get('relationships-poller-timeout'))}
        })
            .populate('dataBroker user')
            .populate({path: 'group', populate: {path: 'dataBrokers'}}).exec();

        await Promise.all(relationships.map(async relationship => {
            const {user, id: relationshipId, scannerQueue, transferrerQueue} = relationship;
            try {
                const doneRequestId = requestId();
                //must calculate status before checking the queue if master failed to initialize there will always be a scanner message
                const relationshipStatus = await calculateRelationshipStatusV2(relationshipId, user);
                if (relationshipStatus.status === MESSAGE_STATUS_FAILED) {
                    await handleEndOfRelationship(relationship, doneRequestId, 'done');
                } else if (scannerQueue && transferrerQueue) {
                    const emptyQueues = await areEmptyQueues(relationshipId, scannerQueue, transferrerQueue);
                    if (emptyQueues) {
                        await handleEndOfRelationship(relationship, doneRequestId, 'done');
                        await updateStatusForRelationship(relationship, MESSAGE_STATUS_DONE, undefined, doneRequestId);
                        await updateRelationshipAndTimeLine(relationship, doneRequestId, MESSAGE_STATUS_DONE);

                    }
                }
            } catch (err) {
                logger.error('failed checking relationship queues for relationship', relationshipId, err);
            }
        }));
    } catch (err) {
        logger.error('Failed to check relationships queues', err);
    } finally {
        setTimeout(() => relationshipsStatusPoller(), 15000);
    }
}

//asterix

export function calculateProgress(status, messages) {
    let progress;
    const bytesToTransfer = get(messages, 'copy.stats.itemsMarkedForTransferBytes', get(messages, 'scan,stats.bytesMarkedForTransfer'));
    const bytesTransferred = get(messages, 'copy.stats.bytesTransferred', get(messages, 'copy.stats.totalTransferredBytes', 0));
    if (isNumber(bytesToTransfer)) {
        if (bytesToTransfer === 0) {
            progress = (status === MESSAGE_STATUS_DONE) ? 100 : 0;
        } else {
            progress = Math.round((bytesTransferred / bytesToTransfer) * 100);
            if (progress >= 100) {
                if (status !== MESSAGE_STATUS_DONE && status !== MESSAGE_STATUS_FAILED) {
                    progress = 99;
                } else {
                    progress = (status === MESSAGE_STATUS_DONE) ? 100 : 0;
                }
            } else {
                progress = (status === MESSAGE_STATUS_DONE) ? 100 : ((status === MESSAGE_STATUS_FAILED) ? 0 : progress);
            }
        }
    }
    return progress;
}

function createStatusObject(status, timeStamp, reason) {
    return {
        status: status,
        message: reason,
        timestamp: timeStamp
    };
}

function createStatusObjectFromMessage(messages, step) {
    return createStatusObject(messages[step].status.toUpperCase(), messages[step].timestamp, messages[step].reason);
}

export function findMaxTimestampForAction(messages) {
    return (messages) ? last(keys(messages).filter(key => messages[key]).map(key => {
        return get(messages, `${key}.timestamp`, 0);
    }).sort()) : undefined;
}

function sortMessagesByMaxTimeStamp(messages) {
    keys(messages).filter(key => messages[key]).forEach(messageType => {
        messages[messageType].maxTimestamp = findMaxTimestampForAction(messages[messageType]);
        messages[messageType].type = messageType;
    });
    const sortedMessages = sortBy(messages, 'maxTimestamp');
    return slice(sortedMessages, sortedMessages.length - 2);
}

function findStatus(activeMessages, status) {
    return find(activeMessages, message => message && message.status === status);
}

function calculateStatusForCreateOrSyncMessage(messages) {
    const failedMessage = findStatus(messages, MESSAGE_STATUS_FAILED.toLowerCase());
    if (failedMessage && messages.maxTimestamp === failedMessage.timestamp) {
        return createStatusObjectFromMessage(messages, failedMessage.step); //FAILED Object
    } else if (messages.copy && messages.copy.status === 'done') {
        return createStatusObjectFromMessage(messages, 'copy'); //DONE object
    } else if (messages.copy && messages.copy.status === 'aborted') {
        return createStatusObject(MESSAGE_STATUS_ABORTED, messages.copy.timestamp, ABORTED_MESSAGE);
    } else {
        return createStatusObject(RUNNING_STATUS, messages.maxTimestamp);
    }
}

export function calculateStatus(messages) {
    const sortedMessages = sortMessagesByMaxTimeStamp(messages);
    const lastMeesage = !isEmpty(sortedMessages) ? last(sortedMessages) : {};
    const beforeLast = first(sortedMessages);
    switch (lastMeesage.type) {
        case COPY_RELATIONSHIP_MESSAGE:
        case SYNC_RELATIONSHIP_MESSAGE: {
            return calculateStatusForCreateOrSyncMessage(lastMeesage);
        }
        case ABORT_RELATIONSHIP: {
            const status = lastMeesage.abort.status === MESSAGE_STATUS_DONE.toLowerCase() ? MESSAGE_STATUS_ABORTED : MESSAGE_STATUS_ABORTING;
            const message = (status === MESSAGE_STATUS_ABORTED) ? ABORTED_MESSAGE : undefined;
            if ([COPY_RELATIONSHIP_MESSAGE, SYNC_RELATIONSHIP_MESSAGE].includes(beforeLast.type)) {
                const secondToLastStatus = calculateStatusForCreateOrSyncMessage(beforeLast);
                if (status === MESSAGE_STATUS_DONE) {
                    return secondToLastStatus;
                }
            }
            return createStatusObject(status, lastMeesage.timestamp, message);
        }
        case STARTUP_RELATIONSHIP: {// when there is a startup and abort there must be at least 2 messages types
            const beforeLastType = beforeLast.type;
            const failed = findStatus(lastMeesage, MESSAGE_STATUS_FAILED.toLowerCase());
            if (failed) {
                return createStatusObjectFromMessage(lastMeesage, failed.step);
            } else if (beforeLastType === ABORT_RELATIONSHIP) {
                return createStatusObject(MESSAGE_STATUS_ABORTED, beforeLast['abort'].timestamp);
            } else if ([COPY_RELATIONSHIP_MESSAGE, SYNC_RELATIONSHIP_MESSAGE].includes(beforeLastType)) {
                const secondToLastStatus = calculateStatusForCreateOrSyncMessage(beforeLast);
                if (secondToLastStatus.status !== RUNNING_STATUS) {
                    return secondToLastStatus;
                } else {
                    return createStatusObject(MESSAGE_STATUS_INTERRUPTED, lastMeesage.timestamp, INTERRUPTED_MESSAGE);
                }
            }
        }
            break;
        default : { // in case the type is different from all the other messages (old messages)
            return createStatusObject(MESSAGE_STATUS_DONE, Date.now());
        }
    }
}

export function calculateTimelineStatusForRelationship(currMessage) {
    if (currMessage) {
        if (currMessage.status === 'failed' || currMessage.step === 'copy' && currMessage.status === 'done') {
            return {status: currMessage.status.toUpperCase(), failureMessage: currMessage.reason};
        } else if (currMessage.step === 'abort') {
            if (currMessage.status === 'done') {
                return {status: MESSAGE_STATUS_ABORTED, failureMessage: ABORTED_MESSAGE};
            } else {
                return {status: MESSAGE_STATUS_ABORTING};
            }
        } else {
            return {status: RUNNING_STATUS};
        }
    }
}

export async function updateTimeLineV2(relationship, currMessage, flag, isAbort) {
    try {
        const timeLineResult = await findTimelineRecordV2(relationship, isAbort);
        if (timeLineResult) {
            let status;
            let failureMessage;
            switch (flag) {
                case 'ABORT': {
                    status = get(currMessage, 'status', '').toUpperCase();
                    failureMessage = currMessage.reason;
                }
                    break;
                case 'STARTUP': {
                    const statusObj = calculateStatus(relationship.messages);
                    status = statusObj.status;
                    failureMessage = statusObj.message;
                }
                    break;
                case 'DELETE': {
                    status = MESSAGE_STATUS_FAILED;
                    failureMessage = TIMELINE_DELETED_RELATIONSHIP_MESSAGE;
                }
                    break;
                default : {
                    const statusObj = calculateTimelineStatusForRelationship(currMessage);
                    status = statusObj.status;
                    failureMessage = statusObj.failureMessage;
                }
            }
            return updateTimeLineV2Schema(status, failureMessage, timeLineResult, relationship.toJSON().activity, isAbort);
        }
    } catch (e) {
        logger.error(`An error occurred in updateTimeLineV2 function trying to update 
        relationship ${relationship.id} with message ${currMessage}. Error ${e}`);
        throw e;
    }

}

export async function updateTimeLine(relationship, currMessage, flag, isAbort = false) {
    try {
        const timeLineResult = await findTimelineRecord(relationship, isAbort);
        if (timeLineResult) {
            let status;
            let failureMessage;
            switch (flag) {
                case 'ABORT': {
                    status = get(currMessage, 'status', '').toUpperCase();
                    failureMessage = currMessage.reason;
                }
                    break;
                case 'STARTUP': {
                    const statusObj = calculateStatus(relationship.messages);
                    status = statusObj.status;
                    failureMessage = statusObj.message;
                }
                    break;
                case 'DELETE': {
                    status = MESSAGE_STATUS_FAILED;
                    failureMessage = TIMELINE_DELETED_RELATIONSHIP_MESSAGE;
                }
                    break;
                default : {
                    const statusObj = calculateTimelineStatusForRelationship(currMessage);
                    status = statusObj.status;
                    failureMessage = statusObj.failureMessage;
                }
            }
            return updateTimeLineSchema(status, failureMessage, timeLineResult, relationship.toJSON().activity, isAbort);
        }
    } catch (e) {
        logger.error(`An error occurred in updateTimeLine function trying to update 
        relationship ${relationship.id} with message ${currMessage}. Error ${e}`);
        throw e;
    }

}

export function isRelationshipRunning(status) {
    return status.status === RUNNING_STATUS;
}

export function isRelationshipAborting(status) {
    return status.status === MESSAGE_STATUS_ABORTING;
}

export function isStartupRunning(relationship) {
    const sortedMessages = sortMessagesByMaxTimeStamp(relationship.messages);
    const lastMessage = last(sortedMessages);
    return Boolean(lastMessage.type === 'startup' && find(lastMessage, step => ['progress', 'started'].includes(step.status)));
}

export async function sendRelationshipMessages(dataBroker, relationship, requestIdFromAsterix, user) {
    const rId = (requestIdFromAsterix) ? requestIdFromAsterix : requestId();
    const clientId = (user) ? user._id : dataBroker.user.id;
    const relationshipResponse = await getRelationshipResponse(relationship, dataBroker);
    return Message.QueueAsync(clientId, {
        header: {
            dataBrokerId: dataBroker.id,
            relationshipId: relationship.id,
            requestId: rId
        },
        payload: {
            'update-relationship-v2': relationshipResponse
        }
    });
}

function updateStatusInDb(pathsToUpdate, relationship, dataBroker) {
    keys(pathsToUpdate).forEach(key => set(relationship, `${key}`, pathsToUpdate[key]));
    relationship.markModified('messages');
    return relationship.save()
        .then(relationship => sendRelationshipMessages(dataBroker, relationship))
        .then(() => updateTimeLine(relationship, undefined, 'STARTUP'))
        .then(() => updateTimeLineV2(relationship, undefined, 'STARTUP'));
}

export function findRelationshipsForStartup(dataBroker) {
    let pathsToUpdate;
    return Relationship.find({dataBroker}).populate('user dataBroker').exec()
        .map(relationship => {
            if (relationship.isEndPointProtocol('source', 's3') && relationship.messages) {
                const newMessagesStatus = get(relationship, 'messages[\'create-relationship-v2\'].copy.status');
                if (newMessagesStatus && ![MESSAGE_STATUS_DONE.toLowerCase(), MESSAGE_STATUS_FAILED.toLowerCase()].includes(newMessagesStatus)) {
                    pathsToUpdate = {
                        'messages[\'create-relationship-v2\'].copy.status': MESSAGE_STATUS_FAILED.toLowerCase(),
                        'messages[\'create-relationship-v2\'].copy.reason': FAILED_AFTER_STARTUP_MESSAGE,
                        'messages[\'create-relationship-v2\'].copy.timestamp': new Date().getTime()
                    };
                    return updateStatusInDb(pathsToUpdate, relationship, dataBroker);
                }
            }
            return relationship;
        })
        .then(relationships => {
            //filter all the relationships with the reason we entered before
            //then send object the contains relationshipId
            return relationships.filter(relationship => !isEmpty(relationship)).map(relationship => {
                return {'relationshipId': relationship.id};
            });
        });
}

export function createObjectForAsterix(endPointObject, credentials) {
    let result;
    const {protocol} = endPointObject;
    const details = endPointObject[protocol];
    switch (protocol) {
        case 'cifs': {
            if (isEmpty(credentials) || !credentials.cifs) {
                throw new BadRequest('CIFS credentials missing!');
            }
            const {username, password, domain} = credentials.cifs;
            const {host, share, path} = details;
            result = {host, share, path};
            result.username = username.trim();
            result.password = password.trim();
            if (domain) {
                result.domain = domain.trim();
            }
            break;
        }
        case 'sgws': {
            if (isEmpty(credentials) || !credentials.sgws) {
                throw new BadRequest('StorageGRID credentials missing!');
            }
            const {accessKey, secretKey} = credentials.sgws;
            const {bucket, prefix, host, port} = details;
            result = {bucket, prefix, host};
            result.accessKey = accessKey;
            result.secretKey = secretKey;
            if (port) {
                result.port = Number.parseInt(port);
            }
            break;
        }
        case 'nfs': {
            const {host, export: exportPath, path} = details;
            result = {host, export: exportPath, path};
            break;
        }
        case 's3': {
            if (isEmpty(details.encryption)) {
                result = details;
                break;
            } else {
                const {bucket, prefix} = details;
                result = {bucket, prefix};
                if (details.encryption.protocol === 'kms') {
                    result.encryption = {'kms': {keyId: details.encryption.keyId.trim()}};
                } else if (details.encryption.type === 'aes') {
                    result.encryption = {'aes': {}};
                }
            }
            break;
        }
    }
    return {[protocol]: result};
}

export async function findRelationshipsWithSecretsInAGroup(groupId) {
    const secretsRequiredProtocols = ['cifs', 'sgws', 'azure', 'gcp'];
    return Relationship.countDocuments({group: groupId, $or: [{'source.protocol': {$in : secretsRequiredProtocols}},
        {'target.protocol': {$in : secretsRequiredProtocols}}]}).exec();
}
