import Message from '../models/message';

import {getLogger} from './logger';

const logger = getLogger();
logger.addContext('module', 'messages-poller');

function getLastMessages(queue, lastMessageId) {
    return Message
        .find({
            queue,
            _id: {$gt: parseInt(lastMessageId)}
        })
        .sort({_id: 1})
        .exec();
}

function getInitLastMessage() {
    return Message
        .find()
        .sort({$natural: -1})
        .limit(1)
        .exec()
        .map(message => {
            message.payload = {
                header: {
                    requestId: 'init',
                    clientId: 'init',
                    agentId: 'init'
                },
                payload: {
                    init: {}
                }
            };
            return message;
        });
}

function extractPayloads(messages) {
    return messages.map(m => {
        const {payload} = m;
        payload.id = m.id;
        return payload;
    });
}

export async function poll(queue, lastMessageId, pollTimeout) {
    logger.debug('Polling for messages...', {queue, lastMessageId, pollTimeout});

    let messages;
    if (lastMessageId) {
        messages = await getLastMessages(queue, lastMessageId);
    } else {
        messages = await getInitLastMessage();
    }

    if (messages.length > 0) {
        logger.debug(`Found ${messages.length} for queue ${queue}`);
    } else {
        logger.debug(`Found 0 for queue ${queue}, cursoring...`);
        messages = await new Promise(resolve => {
            const cursor = Message.find({
                queue,
                _id: {$gt: parseInt(lastMessageId)}
            })
                .cursor()
                // .maxAwaitTimeMS(pollTimeout)
                .addCursorFlag('tailable', true)
                .addCursorFlag('awaitData', true);

            const timeout = setTimeout(() => {
                logger.debug('Cursor timed out');
                cursor.close();
                resolve([]);
            }, pollTimeout);

            cursor.on('data', data => {
                logger.debug('Got cursor data', data);
                cursor.close();
                clearTimeout(timeout);
                resolve([data]);
            });

            cursor.on('error', err => {
                logger.debug('Got cursor error', err);
                cursor.close();
                clearTimeout(timeout);
                resolve([]);
            });
        });
    }

    return extractPayloads(messages);
}
