import Promise from 'bluebird';
import _ from 'lodash';
import fs from 'fs';
import path from 'path';
import moment from 'moment';
import Migration from '../models/migration';
import mongoose from 'mongoose';

import {getLogger} from '../utils/logger';
const logger = getLogger();

const migrationFileTemplate =
`
/**
 * Make any changes you need to make to the database here
 * and return promise
 * 'this' is the mongoose object
 */
export async function up() {
    return this.model('Migration').countDocuments();
}

/**
 * Make any changes that UNDO the up function side effects here (if possible)
 * and return promise
 * 'this' is the mongoose object
 */
export async function down() {
    return this.model('User').countDocuments();
}
`;

const FILE_DATE_FORMAT = 'YYYYMMDD_HHmmss';

function filename(migration) {
    return `${moment(migration.createdAt).format(FILE_DATE_FORMAT)}-${migration.name}.js`;
}

let req = undefined;
if (__WEBPACK__) {
    req = require.context(
        '../../db-migrations', // context folder
        false, // include subdirectories
        /\d{8,}_\d{6,}\-.+\.js$/ // RegExp
    );
} else {
    const dbMigrationsPath = path.resolve(__dirname, '../../db-migrations');
    req = (file) => require(path.resolve(dbMigrationsPath, file));
    req.keys = () => fs.readdirSync(dbMigrationsPath);

}
logger.info('DB Migration files found:', req.keys());

class Migrator {

    async getDbAndFileMigrations() {
        const dbMigrations = await Migration.find().exec();
        return req.keys().map(filename => {
            const [created, name] = filename.replace('./', '').replace('.js', '').split('-');
            const createdAt = moment(created, FILE_DATE_FORMAT).toDate();
            const existsInDatabase = Boolean(_.find(dbMigrations, {name}));
            return {name, createdAt, filename, existsInDatabase};
        });
    }

    // create new migration
    async create(name) {
        const migration = await Migration.findOne({name}).exec();
        if (migration) {
            throw new Error(`There is already a migration with name '${name}' in the database`);
        }
        await this.sync();
        const dateString = moment().format(FILE_DATE_FORMAT);
        const createdAt = moment(dateString, FILE_DATE_FORMAT).toDate();
        const newMigrationFile = `${dateString}-${name}.js`;
        const newMigrationFilePath = path.resolve(path.join('./db-migrations', newMigrationFile));
        fs.writeFileSync(newMigrationFilePath, migrationFileTemplate);
        await Migration.create({name, createdAt});
        logger.info(`Created migration ${name} in ${newMigrationFilePath}.`);
    }

    // run migration in direction up or down up till migration name
    async run(direction = 'up', name = undefined) {
        await this.sync();
        const findQuery = name ? {name} : {};
        const untilMigration = await Migration.findOne(findQuery).exec();
        if (!untilMigration) {
            throw new Error('No migration found');
        }
        const [createdAtFilter, stateFilter, sortOrder] = (direction === 'up') ? ['$gte', 'down', 1] : ['$lte', 'up', -1];
        const query = {
            createdAt: {[createdAtFilter]: untilMigration.createdAt},
            state: stateFilter
        };
        const migrationsToRun = await Migration.find(query).sort({createdAt: sortOrder}).exec();
        if (!migrationsToRun.length) {
            return logger.info('No migrations to run');
        }
        await Promise.mapSeries(migrationsToRun, async migration => {
            migration.filename = filename(migration);
            let migrationModule;
            try {
                migrationModule = req(`./${migration.filename}`);
            } catch (e) {
                throw new Error(`Unable to load migration module ${migration.filename}: ${e.message}`);
            }
            if (!migrationModule[direction] || typeof migrationModule[direction] !== 'function') {
                throw new Error(`Migration module invalid. Missing ${direction} function`);
            }
            try {
                await migrationModule[direction].call(mongoose);
                await Migration.updateOne({name: migration.name}, {$set: {state: direction}}).exec();
                logger.info(`${direction.toUpperCase()}: ${migration.filename}`);

            } catch (e) {
                throw new Error(`Failed running migration ${migration.filename} with error: ${e}`);
            }
        });
        logger.info('All migrations finished successfully.');
    }

    // sync file system migrations with db migrations collection
    async sync() {
        const fileMigrations = await this.getDbAndFileMigrations();
        const filesNotInDb = fileMigrations.filter(migration => !migration.existsInDatabase);
        if (filesNotInDb.length) {
            logger.info('Synchronizing database with file system migrations...');
            await Promise.all(filesNotInDb.map(file => {
                logger.info(`Adding migration ${file.filename} into database from file system. State is DOWN`);
                return Migration.create({
                    name: file.name,
                    createdAt: file.createdAt
                });
            }));
        }
    }

    // list all migrations and their status
    async list() {
        await this.sync();
        const migrations = await Migration.find().sort({createdAt: 1}).exec();
        if (!migrations.length) {
            logger.info('No migrations found');
        }
        return migrations.map(migration => {
            logger.info(`${migration.state.toUpperCase()}: ${filename(migration)}`);
            return migration.toObject();
        });
    }
}

export default new Migrator();
