import Promise from 'bluebird';
import os from 'os';
import {
    createQueue,
    receiveMessage,
    deleteMessage,
    getQueueArn,
    allowArnSendMessage
} from '../lib/sqs';
import {subscribe, confirmSubscription} from '../lib/sns';
import {SERVER_HOST, MP_SUBSCRIPTION_STATUS_SNS_TOPIC, SELLER_ACCESS_KEY, SELLER_SECRET_KEY, MP_PRODUCT_CODE_METERING,
    MP_PRODUCT_CODE_CONTRACT, PAYGO_SUBSCRIPTION, CONTRACT_SUBSCRIPTION} from '../utils/consts';
import User from '../models/user';
import {BadRequest} from '../utils/errors';

import {getLogger} from '../utils/logger';
const logger = getLogger();
//returns the AWS Subscription type by the product code
export function validateProductCode(code) {
    if (code === MP_PRODUCT_CODE_METERING) {
        return PAYGO_SUBSCRIPTION;
    } else if (code === MP_PRODUCT_CODE_CONTRACT) {
        return CONTRACT_SUBSCRIPTION;
    } else {
        throw new BadRequest(`Resolved product code is incorrect. Expected: ${MP_PRODUCT_CODE_METERING} or ${MP_PRODUCT_CODE_CONTRACT}, Got: ${code}`);
    }
}

function handleSubscriptionStatusNotification(message, timestamp) {
    const customerIdentifier = message['customer-identifier'];
    const subscriptionType = validateProductCode(message['product-code']);

    // make update multi
    if (subscriptionType === PAYGO_SUBSCRIPTION) {
        return User.find({
            'subscription.customerIdentifier': customerIdentifier,
            'subscription.type': PAYGO_SUBSCRIPTION
        }).exec()
            .then(users => {
                if (users.length > 0) {
                    users.forEach(user => {
                        if (timestamp > user.subscription.updatedAt) {
                            logger.info(`Updating Subscription status for user ${user.email}, status ${message.action}, timestamp ${timestamp}`);
                            user.subscription.status = message.action;
                            user.subscription.type = 'PAYGO';
                            return user.save();
                        } else {
                            logger.warn(`Received outdated subscription notification [Action: ${message.action}. CustomerIdentifier: ${customerIdentifier}, timestamp ${timestamp}]`);
                        }
                    });
                } else {
                    logger.warn(`Subscription notification's customer identifier not found. [CustomerIdentifier ${customerIdentifier}, Action ${message.action}, timestamp ${timestamp}`);
                }
            });
    } else {
        logger.warn('Subscription type is CONTRACT and not PAYGO');
    }
}

function handleMessage(sqsMessage) {
    const isSubscriptionNotification = (message) => message.action && message['customer-identifier'] && message['product-code'];
    let subscriptionMessage, message;

    try {
        message = JSON.parse(sqsMessage);
        if (message.Message && !(message.Type && message.Type === 'SubscriptionConfirmation')) {
            subscriptionMessage = JSON.parse(message.Message);
        }
    } catch (err) {
        logger.warn('Got non JSON SQS message', err);
        return Promise.resolve();
    }

    if (message.Type && message.Type === 'SubscriptionConfirmation') {
        logger.info(`Got topic ${MP_SUBSCRIPTION_STATUS_SNS_TOPIC} subscription confirmation - confirming`);
        return confirmSubscription(MP_SUBSCRIPTION_STATUS_SNS_TOPIC, message.Token, SELLER_ACCESS_KEY, SELLER_SECRET_KEY);
    } else if (subscriptionMessage && isSubscriptionNotification(subscriptionMessage)) {
        return handleSubscriptionStatusNotification(subscriptionMessage, new Date(message.Timestamp));
    } else {
        logger.warn('Not a supported message:', message);
        return Promise.resolve();
    }
}

function pollQueue(queueUrl) {
    logger.debug(`Polling for messages from ${queueUrl}`);
    return receiveMessage(queueUrl)
        .then(messages => {
            return Promise.map(messages, message => {
                logger.debug('Got SQS message:', message);
                return handleMessage(message.body)
                    .then(() => {
                        logger.debug('Deleting message, receipt handle:', message.receiptHandle);
                        return deleteMessage(queueUrl, message.receiptHandle);
                    });
            });
        })
        .then(() => {
            setImmediate(() => {
                pollQueue(queueUrl);
            });
        })
        .catch(err => {
            logger.error(`Failed to poll messages from ${queueUrl}`, err);
            if (err.code === 'AWS.SimpleQueueService.NonExistentQueue') {
                logger.warn(`${queueUrl} does not exist - stopping polling`);
            } else {
                setTimeout(() => {
                    pollQueue(queueUrl);
                }, 5000);
            }
        });
}

export function startMPSubscriptionStatusListener() {
    // #1 - crate queue
    // #2 - translate queue url to arn
    // #3 - allow send message to the queue only from the topic
    // #4 - subscribe queue to topic (in seller account)
    // #5 - start polling from the queue
    // #6 - once SubscriptionConfirmation message is received - confirm the subscription (in seller account)
    const prefix = 'MP_SUBSCRIPTION_STATUS_';
    const [serverHost] = SERVER_HOST.split('.');
    const name = serverHost === 'local' ? `${prefix}${os.hostname().replace(/[-.]/g, '_')}` : `${prefix}${serverHost}`;

    logger.info(`Creating SQS queue ${name}`);

    return createQueue(name)
        .then(queueUrl => {
            logger.info(`Queue ${queueUrl} successfully created`);
            return getQueueArn(queueUrl)
                .then(queueArn => {
                    logger.info(`Queue ARN ${queueArn}`);
                    logger.info(`Allow send message to queue ${queueArn} from topic ${MP_SUBSCRIPTION_STATUS_SNS_TOPIC}`);
                    return allowArnSendMessage(queueUrl, queueArn, MP_SUBSCRIPTION_STATUS_SNS_TOPIC)
                        .then(() => {
                            logger.info(`Subscribing queue ${queueArn} to topic ${MP_SUBSCRIPTION_STATUS_SNS_TOPIC}`);
                            return subscribe(MP_SUBSCRIPTION_STATUS_SNS_TOPIC, queueArn, SELLER_ACCESS_KEY, SELLER_SECRET_KEY)
                                .then(subscriptionArn => {
                                    logger.info(`SQS ${queueArn} was successfully subscribed to SNS ${MP_SUBSCRIPTION_STATUS_SNS_TOPIC}, subscription ARN ${subscriptionArn}`);
                                    pollQueue(queueUrl);
                                });
                        });
                });
        });
}
