import { sortBy, get, isObject, flattenDeep, flatten } from 'lodash';
import DataBroker from '../models/data-broker';
import Group from '../models/group';
import User from '../models/user';
import Message from '../models/message';
import Relationship from '../models/relationship';
import RelationshipStats from '../models/relationship-statistics';
import Timeline from '../models/timeline';
import TimelineV2 from '../models/timeline-v2';
import {receiveMessage, deleteMessage} from '../lib/sqs';
import {
    awsRegionCodeToDisplayName, sendRelationshipMessages, updateTimeLine, updateRelationshipAndTimeLine,
    updateTimeLineV2, updateSubscription, isNewDataBroker, gcpRegionCodeToDisplayName, findRelationshipsWithSecretsInAGroup
} from './common';
import { requestId } from '../utils/utils';
import {
    STATUS_SUCCESS, MESSAGE_STATUS_DONE, MESSAGE_STATUS_FAILED,
    SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS, PAYGO_SUBSCRIPTION, RUNNING_STATUS
} from '../utils/consts';
import { sendStartupCommand, sendMetadataRequestCommand } from '../operations/data-broker-operations';
import {updateDataBrokerWithVersion} from '../utils/update-data-broker';


import {getLogger} from '../utils/logger';
const logger = getLogger();

async function updateSubscriptionByAccountId(dataBroker) {
    const user = await User.findOne({'_id': dataBroker.user}).exec();
    const users = await DataBroker.findUsersByAwsAccountId(dataBroker.awsAccountId);
    const flattenUsers = flatten(users);
    if (user.subscription.status === SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS) {
        return updateSubscription(user.subscription, flattenUsers);
    } else {
        const subscription = flattenUsers.find(user => user.subscription.status === SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS && user.subscription.type === PAYGO_SUBSCRIPTION);
        return updateSubscription(subscription, flattenUsers);
    }
}

async function registerDataBrokerPlacement(dataBroker, placement) {
    const {instanceId, privateIp, region, availabilityZone, vpcId, vpcCidrBlock, subnetId, updateNewVersion,
        subnetCidrBlock, platform, hostname, awsAccountId, os, release, osTotalMem, node, proxy,
        cpus, processMaxMem, version, certificate, publicIp } = placement;
    dataBroker.instanceId = instanceId;
    dataBroker.privateIp = privateIp;
    dataBroker.region = awsRegionCodeToDisplayName(region);
    dataBroker.availabilityZone = availabilityZone;
    dataBroker.vpcId = vpcId;
    dataBroker.vpcCidrBlock = vpcCidrBlock;
    dataBroker.subnetId = subnetId;
    dataBroker.subnetCidrBlock = subnetCidrBlock;
    dataBroker.version = version;
    dataBroker.platform = platform;
    dataBroker.hostname = hostname;
    dataBroker.awsAccountId = awsAccountId;
    dataBroker.versionUpdateSent = undefined;
    dataBroker.os = os;
    dataBroker.release = release;
    dataBroker.osTotalMem = osTotalMem;
    dataBroker.node = node;
    dataBroker.cpus = cpus;
    dataBroker.processMaxMem = processMaxMem;
    dataBroker.publicIp = publicIp;
    dataBroker.proxy = proxy;
    dataBroker.updateNewVersion = updateNewVersion;
    if (certificate) {
        dataBroker.encryption.certificate = certificate;
    }

    const updatedDataBroker = await dataBroker.save();
    return updateSubscriptionByAccountId(updatedDataBroker);
}

function sendDataBrokerPlacement(dataBroker, userId, requestId, placement) {
    placement.region = awsRegionCodeToDisplayName(placement.region);
    logger.info(`Sending data broker ${dataBroker.id} placement information`);

    Message.SendAsync(dataBroker.id, userId, requestId, 'data-mover-status', {
        id: dataBroker.id,
        status: 'COMPLETE',
        message: undefined,
        placement
    });
}

async function generatePlacementTimelineIfNeeded(dataBroker, placement) {
    let summary;
    const {platform, version:dataBrokerVersion, user, name, id:dataBrokerId} = dataBroker;
    const {version: placementVersion} = placement;
    if (!platform) {
        logger.info(`${dataBroker} created and registered`);
        summary = 'New Data Broker created and registered';
    } else if (dataBrokerVersion !== placement.version) {
        logger.info(`${dataBroker.toString()} version updated ${dataBrokerVersion} => ${placementVersion}`);
        summary = `Data-Broker version updated ${dataBrokerVersion} => ${placementVersion}`;
    }
    if (summary) {
        await new Timeline({
            user,
            requestId: requestId(),
            status: MESSAGE_STATUS_DONE,
            summary,
            dataBroker: name
        }).save();
        await new TimelineV2({
            user,
            requestId: requestId(),
            status: MESSAGE_STATUS_DONE,
            summary,
            dataBroker: {dataBrokerId, name }
        }).save();
    }
}

function sendSmallCommandsResponse(message, fieldsName, paths, isStatusSuccess, dataBroker, msg, msgTitle, requestId, isPhoenix) {
    const {id: dataBrokerId} = dataBroker;
    if (isStatusSuccess) {
        message[fieldsName] = paths.sort();
    } else {
        message.message = (isPhoenix) ? msg.reason : msg.payload.error;
    }
    return Message.SendAsync(dataBrokerId, dataBroker.user.id, requestId, msgTitle, message);
}

async function handleRelationshipMessage(payload, isStartup, requestId, dataBroker, command) {
    const {relationshipId} = payload;
    const {step} = payload;
    const filter = {
        _id: relationshipId,
        user: dataBroker.user,
    };
    if (isStartup) {
        filter.$or = [{[ `messages.${command}.${step}`]: {$exists: false}}, {[`messages.${command}.${step}.timestamp`]: {$lte: payload.timestamp}}];
    } else {
        filter.$and = [{$or: [{[ `messages.${command}.${step}`]: {$exists: false}}, {[`messages.${command}.${step}.timestamp`]: {$lte: payload.timestamp}}]},
            {[`messages.${command}.${step}.status`]: {$ne: 'done'}}];
    }
    const update = {
        $set: {
            [`messages.${command}.${step}`]: payload
        }
    };
    const relationship = await Relationship.findOneAndUpdate(filter, update, {
        upsert: false,
        new: true
    }).populate('user dataBroker').exec();
    if (!relationship) {
        logger.warn(`Got  relationship message for relationship id '${relationshipId}' from ${dataBroker} but relationship ${relationship} not found (might have been deleted`);
        return;
    }
    await sendRelationshipMessages(dataBroker, relationship, requestId);
    if (isStartup) {
        await updateTimeLine(relationship, undefined, 'STARTUP');
        await updateTimeLineV2(relationship, undefined, 'STARTUP');
    } else {
        await updateTimeLine(relationship, payload);
        await updateTimeLineV2(relationship, payload);
    }
    return relationship;
}

async function handleMessage(dataBroker, message) {
    const {statusesQueueUrl} = dataBroker;
    const {body, attributes} = message;

    const newDataBroker = isNewDataBroker(dataBroker);

    logger.debug(`Received message from queue '${statusesQueueUrl}' of ${dataBroker}:`, JSON.stringify(message));
    // parse message body
    const msg = JSON.parse(body);
    // handle message
    const {command, payload} = msg;
    const commandToHandle = command || attributes.type;
    const requestIdReceived = (msg.header) ? msg.header.requestId : attributes.requestId;
    const isStatusSuccess = payload && payload.status === STATUS_SUCCESS || msg && msg.status === 'SUCCEEDED';
    const status = isStatusSuccess ? MESSAGE_STATUS_DONE : MESSAGE_STATUS_FAILED;

    switch (commandToHandle) {
        case 'create-relationship-v2' :
        case 'sync-relationship' : {
            await handleRelationshipMessage(payload, false, requestIdReceived, dataBroker, command);
        }
            break;
        case 'abort-relationship': {
            const relationship = await handleRelationshipMessage(payload, false, requestIdReceived, dataBroker, command);
            await updateTimeLine(relationship, payload, 'ABORT', true);
            await updateTimeLineV2(relationship, payload, 'ABORT', true);
        }
            break;
        case 'startup' : {
            if (newDataBroker) {
                const {relationships} = msg;
                await Promise.all(relationships.map(async relationshipObject => {
                    const {relationshipId} = relationshipObject;
                    const relationship = await Relationship.findById(relationshipId).populate('user dataBroker group').populate({
                        path: 'group',
                        populate: {path: 'dataBrokers'}
                    }).exec();
                    if (!relationship) {
                        logger.warn(`Got  relationship message for relationship id '${relationshipId}' from ${dataBroker._id} but relationship ${relationship} not found (might have been deleted`);
                    } else {
                        const filter = {
                            relationship: relationshipId,
                            dataBroker: dataBroker._id,
                            user: relationship.user
                        };
                        const {status, reason} = msg;
                        await RelationshipStats.findOneAndUpdate(filter, {$set: {status, reason}}, {
                            upsert: true,
                            new: true,
                            strict: false
                        }).exec();
                        await updateRelationshipAndTimeLine(relationship, requestIdReceived);
                    }
                }));
            } else {
                await handleRelationshipMessage(payload, true, requestIdReceived, dataBroker, command);
            }
        }
            break;
        case 'health': {
            const { transferRate } = msg;
            dataBroker.transferRate = transferRate || 0;
            await dataBroker.save();
            const runningRelationships = await RelationshipStats.find({
                dataBroker: dataBroker.id,
                $or: [
                    {status : { $exists : false } }, //in case the broker was added to a relationship it will not have a RUNNING status
                    { status : RUNNING_STATUS}]
            }).exec();
            if (runningRelationships.length) {
                await Message.SendAsync(dataBroker.id, dataBroker.user.id, requestId(),
                    'databroker-transfer-rate', {transferRate, dataBrokerId: dataBroker.id});
            }
        }
            break;
        case 'placement': {
            const placementInfo = (newDataBroker) ? msg : payload;
            await generatePlacementTimelineIfNeeded(dataBroker, placementInfo);
            await registerDataBrokerPlacement(dataBroker, placementInfo);
            if (newDataBroker) {
                const {groupId, _id: dataBrokerId} = dataBroker;
                const group = await Group.findById(groupId).populate('dataBrokers').exec();
                const {dataBrokers} = group;
                logger.info('There are more than 1 data brokers inside the group', groupId.toString());
                const numberOfRelationshipsWithPasswordInAGroup = await findRelationshipsWithSecretsInAGroup(groupId);
                if (dataBrokers.length > 1 && numberOfRelationshipsWithPasswordInAGroup > 0) {
                    const aliveDataBrokerInAGroup = dataBrokers.find(dataBroker => dataBroker.isAlive() && dataBroker.id.toString() !== dataBrokerId.toString());
                    if (aliveDataBrokerInAGroup) {
                        logger.info(`There is at least one alive data broker in ${group._id.toString()}`);
                        await sendMetadataRequestCommand(requestId(), aliveDataBrokerInAGroup, dataBrokerId);
                    }
                } else {
                    await sendStartupCommand(requestId(), dataBroker);
                    await updateDataBrokerWithVersion(dataBroker.id);
                }
            } else {
                await dataBroker.startup(requestId());
            }
            await sendDataBrokerPlacement(dataBroker, dataBroker.user.id, requestId(), placementInfo);
        }
            break;
        case 'metadata': {
            const {targetDataBrokerId, metadata} = msg;
            //receive metadata response -
            const targetDataBroker = await DataBroker.findById(targetDataBrokerId).exec();
            if (!targetDataBroker) {
                logger.error(`DataBroker with id ${targetDataBrokerId} was not found cant send metadata-response message `);
            } else if (isStatusSuccess) {
                logger.info(`Received metadata message ${msg} from ${dataBroker.id}, sending startup command`);
                //sending the message to the target metadata
                await sendStartupCommand(requestId(), targetDataBroker, metadata);
            } else {
                logger.error('Failed to send metadata to databroker', targetDataBrokerId);
            }
        }
            break;
        case 'update': {
            //in case of error there is a response from the data-broker
            const {bundleVersion, bundleUrl, reason} = msg;
            if (!isStatusSuccess) {
                logger.error(`Error updating data broker ${dataBroker.id} with updateUrl '${bundleUrl}' and version '${bundleVersion}':`, reason);
            }
        }
            break;
        case 'azure-blob-storage-containers':
            await Message.SendAsync(dataBroker.id, dataBroker.user.id, requestIdReceived,
                'list-azure-blob-storage-containers', msg);
            break;
        case 'azure-blob-storage-container-folders':
            await Message.SendAsync(dataBroker.id, dataBroker.user.id, requestIdReceived,
                'list-azure-blob-storage-container-folders', msg);
            break;
        case 'gcp-buckets': {
            const {buckets, reason, projectId} = msg;
            let message = {projectId};
            if (isStatusSuccess) {
                message.buckets = sortBy(buckets.map(bucket => {
                    bucket.region = {
                        name: bucket.region,
                        displayName: gcpRegionCodeToDisplayName(bucket.region)
                    };
                    return bucket;
                }), 'name');
                message.status = status;
            } else {
                message = {
                    status,
                    message: reason
                };
            }
            await Message.SendAsync(dataBroker.id, dataBroker.user.id, requestIdReceived,
                'list-gcp-buckets', message);
        }
            break;
        case 'gcp-bucket-folders': {
            const {projectId, bucket, prefix, folders} = msg;
            const message = {
                projectId,
                bucket: bucket.name,
                status,
                prefix
            };
            await sendSmallCommandsResponse(message, 'list', folders, isStatusSuccess, dataBroker, msg,
                'list-gcp-bucket-folders', requestIdReceived, true);
        }
            break;
        case 'shares':
        case 'list-cifs-shares-v2': {
            const {host, shares} = newDataBroker ? msg : payload;
            await sendSmallCommandsResponse({
                host,
                status
            }, 'paths', shares, isStatusSuccess, dataBroker, msg, 'list-cifs-shares', requestIdReceived, newDataBroker);
        }
            break;
        case 'share-folders':
        case 'list-cifs-share-folders-v2': {
            const {host, folders, path, share} = newDataBroker ? msg : payload;
            const message = {
                status,
                host,
                share,
                path
            };
            await sendSmallCommandsResponse(message, 'folders', folders, isStatusSuccess, dataBroker, msg,
                'list-cifs-share-folders', requestIdReceived, newDataBroker);
        }
            break;
        case 'exports': //phoenix
        case 'list-nfs-exports-v2': {
            const exportsInfo = (newDataBroker) ? msg : payload;
            const {exports, host} = exportsInfo;
            await sendSmallCommandsResponse({host, status}, 'paths', exports, isStatusSuccess, dataBroker, msg,
                'list-nfs-exports', requestIdReceived, newDataBroker);
        }
            break;
        case 'export-folders': //phoenix
        case 'list-nfs-export-folders-v2': {
            const exportFoldersInfo = (newDataBroker) ? msg : payload;

            const {host: nfsServer, export: exportName, path: rawPath, folders} = exportFoldersInfo;
            const path = rawPath ? `${exportName}${rawPath}` : exportName;
            const message = {
                status,
                nfsServer,
                export: exportName,
                path
            };
            await sendSmallCommandsResponse(message, 'folders', folders, isStatusSuccess, dataBroker, msg,
                'list-nfs-export-folders', requestIdReceived, newDataBroker);
        }
            break;
        case 'list-s3-buckets-v2': {
            const {buckets, error} = payload;
            let message;
            if (isStatusSuccess) {
                message = sortBy(buckets.map(bucket => {
                    bucket.region = {
                        name: bucket.region,
                        displayName: awsRegionCodeToDisplayName(bucket.region)
                    };
                    bucket.tags = Object.keys(bucket.tags).map(key => {
                        return {Key: key, Value: bucket.tags[key]};
                    });
                    return bucket;
                }), 'name').filter(bucket => !bucket.name.startsWith('fabric-pool-'));
            } else {
                message = {
                    status: MESSAGE_STATUS_FAILED,
                    message: error
                };
            }
            await Message.SendAsync(dataBroker.id, dataBroker.user.id, requestIdReceived,
                'list-s3-buckets', message);
        }
            break;
        case 'list-s3-bucket-folders-v2': {
            const {bucket, prefix, partial, folders} = payload;
            const message = {
                status,
                bucket,
                prefix,
                partial
            };
            await sendSmallCommandsResponse(message, 'list', folders, isStatusSuccess, dataBroker, msg,
                'list-s3-bucket-folders', requestIdReceived);
        }
            break;
        case 'sgws-buckets':
        case 'list-sgws-buckets-v2': {
            const sgwsBucketsInfo = (newDataBroker) ? msg : payload;

            const {buckets, host, port} = sgwsBucketsInfo;
            const message = {
                status,
                host,
                port
            };
            if (isStatusSuccess) {
                message.buckets = sortBy(buckets, 'name');
            } else {
                message.message = (newDataBroker) ? sgwsBucketsInfo.reason : sgwsBucketsInfo.error;
            }
            await Message.SendAsync(dataBroker.id, dataBroker.user.id,
                requestIdReceived, 'list-sgws-buckets', message);
        }
            break;
        case 'sgws-buckets-folders':
        case 'list-sgws-bucket-folders-v2': {
            const sgwsBucketsFoldersInfo = (newDataBroker) ? msg : payload;
            const {host, port, bucket, prefix, folders} = sgwsBucketsFoldersInfo ;
            const message = {
                status,
                host,
                port,
                bucket,
                prefix
            };
            await sendSmallCommandsResponse(message, 'list',  (folders) ? folders.filter(folder => folder !== '.') : [],
                isStatusSuccess, dataBroker, msg, 'list-sgws-bucket-folders', requestIdReceived);
        }
            break;
        case 'buckets': {
            let message;
            const {buckets, reason} = msg;
            if (isStatusSuccess) {
                message = buckets.map(bucket => {
                    bucket.region = {
                        name: bucket.region,
                        displayName: awsRegionCodeToDisplayName(bucket.region)
                    };
                    return bucket;
                });
                message.status = status;
            } else {
                message = {
                    status,
                    message: reason
                };
            }
            await Message.SendAsync(dataBroker.id, dataBroker.user.id, requestIdReceived,
                'list-s3-buckets', message);
        }
            break;
        case 'bucket-folders': {
            const {bucket, prefix, folders} = msg;
            const message = {
                bucket: (isStatusSuccess) ? bucket.name : bucket,
                status,
                prefix
            };
            await sendSmallCommandsResponse(message, 'list', folders, isStatusSuccess, dataBroker, msg,
                'list-s3-bucket-folders', requestIdReceived, true);
        }
            break;
        case 'abort':
        case 'copy': {
            const {relationshipId} = msg;
            const isStatusMessage = Boolean(msg.status);
            const relationship = await Relationship.findById(relationshipId).populate('user dataBroker group').populate({
                path: 'group',
                populate: {path: 'dataBrokers'}
            }).exec();
            if (!relationship) {
                logger.warn(`Got  relationship message for relationship id '${relationshipId}' from ${dataBroker._id} but relationship ${relationship} not found (might have been deleted`);
            } else {
                const filter = {
                    relationship: relationshipId,
                    dataBroker: dataBroker._id,
                    user: relationship.user
                };

                let update;
                if (isStatusMessage) {
                    const {status, reason} = msg;
                    update = {$set: {status, reason: (reason) ? reason : undefined}};
                } else {
                    const updateObj = {};

                    const findAllKeys = (accumString, obj) => {
                        if (isObject(obj)) {
                            return Object.keys(obj).map(key => {
                                const nextString = (accumString === '') ? `${key}` : `${accumString}.${key}`;
                                return findAllKeys(nextString, obj[key]);
                            });
                        } else {
                            return accumString;
                        }
                    };
                    const allKeys = flattenDeep(findAllKeys('', msg));
                    allKeys.forEach(key => {
                        if (key !== 'relationshipId') {
                            updateObj[key.toLowerCase()] = get(msg, key);
                        }
                    });

                    update = {$inc: updateObj};

                }
                await RelationshipStats.findOneAndUpdate(filter, update, {
                    upsert: true,
                    strict: false
                });
                await updateRelationshipAndTimeLine(relationship, requestIdReceived);
            }
        }
            break;
        default:
            logger.debug(`Unknown command '${commandToHandle}' while processing message from SQS queue '${statusesQueueUrl}' of ${dataBroker}`, msg);
    }
}

export async function pollQueue(dataBroker) {
    const {id, statusesQueueUrl} = dataBroker;

    try {
        logger.debug(`Polling messages from SQS queue '${statusesQueueUrl}' of ${dataBroker}`);
        const messages = await receiveMessage(statusesQueueUrl);
        if (messages.length > 0) {
            // retrieve data broker and update data broker last ping in one shot
            const updatedDataBroker = await DataBroker.findByIdAndUpdate(id, {$set: {'lastPing.wasabi': new Date()}}, {new: true}).populate('user').exec();

            for (const message of messages) {
                try {
                    await handleMessage(updatedDataBroker, message);
                } catch (err) {
                    logger.error(`Failed handling SQS message from queue ${statusesQueueUrl} of ${dataBroker}`, {message, err});
                } finally {
                    await deleteMessage(statusesQueueUrl, message.receiptHandle);
                }
            }
        }

        pollQueue(dataBroker);
    } catch (err) {
        if (err.code === 'AWS.SimpleQueueService.NonExistentQueue') {
            logger.warn(`'${statusesQueueUrl}' SQS queue of ${dataBroker} does not exist anymore. Will not continue polling`);
        } else {
            logger.warn(`Failed to poll message - delaying polling for SQS queue '${statusesQueueUrl}' of ${dataBroker}`, err);
            setTimeout(() => pollQueue(dataBroker, statusesQueueUrl), 5000);
        }
    }
}

export function startListeningDataBrokersStatuses() {
    logger.info('Looking for status queues...');
    DataBroker.find({}).populate('user').exec().map(pollQueue);
}
