import Promise from 'bluebird';
import agenda from './agenda';
import { filter, orderBy, last} from 'lodash';
import DataBroker from '../models/data-broker';
import {
    AGENDA_TASK_NAME_UPDATE_DATABROKER,
    DFIO_METADATA_BUCKET_REGION,
    DFIO_METADATA_BUCKET_NAME,
    DATA_BROKER_REPO_FOLDER,
    DATA_BROKER_BUNDLE_LABEL
} from './consts';
import { listObjectsAsync, getObjectTaggingAsync, getS3ObjectUrl } from '../lib/s3';
import { getRuntimeEnvironment} from '../../../environment';
import { NotFound } from '../utils/errors';
import {sendUpdateVersionCommand} from '../operations/data-broker-operations';
import {requestId} from './utils';
import {isNewDataBroker} from './common';

import {getLogger} from '../utils/logger';
const logger = getLogger();

export async function getDataBrokerLatestVersionAsync() {
    const prefix = `${DATA_BROKER_REPO_FOLDER}/${getRuntimeEnvironment()}/phoenix`;
    logger.info(`Looking for data broker bundle in bucket ${DFIO_METADATA_BUCKET_NAME}, region ${DFIO_METADATA_BUCKET_REGION}, prefix ${prefix}, label ${DATA_BROKER_BUNDLE_LABEL}`);

    const objects = await listObjectsAsync(DFIO_METADATA_BUCKET_REGION, DFIO_METADATA_BUCKET_NAME, prefix);
    const promises = objects.map(async object => {
        const tags = await getObjectTaggingAsync(DFIO_METADATA_BUCKET_REGION, DFIO_METADATA_BUCKET_NAME, object.key);
        object['environment'] = tags['environment'];
        object['buildTime'] = tags['buildTime'];
        object['version'] = tags['version'];
        object['label'] = tags['label'];

        return (object);
    });
    const objectsWithTags = await Promise.all(promises);
    logger.info(`${objectsWithTags.length} agent bundles found in ${DFIO_METADATA_BUCKET_NAME}/${prefix}`);
    const filtered = filter(objectsWithTags, o => {
        // use labeled bundles only if bundle name is defined in server configuration
        // use not labeled bundles only if bundle name is not defined
        return (DATA_BROKER_BUNDLE_LABEL ? o.label && o.label === DATA_BROKER_BUNDLE_LABEL : !o.label) && o.key.includes('phoenix');
    });

    const orderedObjects = orderBy(filtered, ['lastModified']);
    const lastObject = last(orderedObjects);
    if (lastObject) {
        // the key is publicly available since create-data-broker-bundle script makes bundle 'public-read'
        const dataBrokerBundleUrl = getS3ObjectUrl(DFIO_METADATA_BUCKET_REGION, DFIO_METADATA_BUCKET_NAME, lastObject.key,
            true, true);
        logger.info('Data broker bundle URL:', dataBrokerBundleUrl);
        lastObject.url = dataBrokerBundleUrl;
        return lastObject;
    } else {
        throw new NotFound(`Data Broker bundle not found. Bucket ${DFIO_METADATA_BUCKET_NAME}, prefix ${prefix}, label ${DATA_BROKER_BUNDLE_LABEL}`);
    }
}

export async function updateDataBrokerWithVersion(dataBrokerId) {
    //#1 get latest data broker version
    const latestVersion = await getDataBrokerLatestVersionAsync();
    logger.info(`Latest online data broker version for phoenix '${latestVersion.version}'`);
    const dataBroker = await DataBroker.findOne({version: {$ne : latestVersion.version}, '_id': dataBrokerId}).exec();
    if (dataBroker) {
        logger.info(`Sending dataBroker with ${dataBrokerId} from version ${dataBroker.version} to version ${latestVersion.version}`);
        await DataBroker.findOneAndUpdate({'_id': dataBroker.id}, {$set: {versionUpdateSent: new Date()}}).exec();
        await sendUpdateVersionCommand(requestId(), dataBroker, latestVersion);
    } else {
        logger.info(`dataBroker with id ${dataBrokerId} doesnt need an update`);
    }
}

export async function updateDataBrokers() {
    //#1 get latest data broker version
    const latestVersion = await getDataBrokerLatestVersionAsync();
    logger.info(`Latest online data broker version for phoenix '${latestVersion.version}'`);
    //#2 get all phoenix data brokers
    const allDataBrokers = await DataBroker.find({version: {$ne : latestVersion.version}}).exec();
    const alivePhoenixDataBrokers = allDataBrokers.filter(dataBroker => isNewDataBroker(dataBroker) && dataBroker.isAlive());
    //#3 check if the version is equal
    //#4 send update command
    const promises = alivePhoenixDataBrokers.map(async dataBroker => {
        logger.info(`Sending ${dataBroker} update version command to version '${latestVersion.version}' with url '${latestVersion.url}'`);
        await DataBroker.findOneAndUpdate({'_id': dataBroker.id}, {$set: {versionUpdateSent: new Date()}}).exec();
        await sendUpdateVersionCommand(requestId(), dataBroker, latestVersion);
        logger.info(`Sent ${dataBroker} update version command to version '${latestVersion.version}'`);
    });
    await Promise.all(promises);
}

agenda.define(AGENDA_TASK_NAME_UPDATE_DATABROKER, {lockLifetime: 10000}, async(job, done) => {
    logger.info('update data broker task started');
    try {
        await updateDataBrokers();
        logger.info('update data broker task finished');
    } catch (e) {
        logger.error('update-data-broker task failed:', e);
    } finally {
        done();
    }
});
