import Promise from 'bluebird';
import fs from 'fs';
import path from 'path';
import mongoose from 'mongoose';
import * as utils from '../../src/lib/marketplace-metering';
import DataBroker from '../../src/models/data-broker';
import Relationship from '../../src/models/relationship';
import User from '../../src/models/user';
import Cost from '../../src/models/cost';
import { expect } from '../globals';

const AWS = require('mock-aws');

describe('marketplace-metering library', () => {
    const timestamp = new Date().toISOString();
    const record1 = {
        entry: {
            CustomerIdentifier: 'customerId',
            Dimension: 'Tier1',
            Quantity: 5,
            Timestamp: timestamp
        },
        uid: '58933ddf96d0230a142f840f',
        byol: { reservedRelationships: 20, usedRelationships: 3 },
        isSuccess: false
    };
    const record2 = {
        entry: {
            CustomerIdentifier: 'customerId',
            Dimension: 'Tier2',
            Quantity: 12,
            Timestamp: timestamp
        },
        uid: '58933ddf96d0230a142f840f',
        byol: { reservedRelationships: 20, usedRelationships: 3 },
        isSuccess: false
    };
    describe('getCustomerTiers', () => {
        it('Tiers should be partitioned according to the number of relationships (only tier1 relationships)', () => {
            const result = utils.getCustomerTiers(4, 0);
            result.tier1.should.equal(4);
            result.tier2.should.equal(0);
            result.tier3.should.equal(0);
        });

        it('Tiers should be partitioned according to the number of relationships (only tier1 and tier2 relationships)', () => {
            const result = utils.getCustomerTiers(18, 0);
            result.tier1.should.equal(5);
            result.tier2.should.equal(13);
            result.tier3.should.equal(0);
        });

        it('Tiers should be partitioned according to the number of relationships (all tiers relationships)', () => {
            const result = utils.getCustomerTiers(23, 0);
            result.tier1.should.equal(5);
            result.tier2.should.equal(15);
            result.tier3.should.equal(3);
        });

        it('Tiers should be partitioned according to the number of relationships (tiers and BYOL relationships)', () => {
            const result = utils.getCustomerTiers(23, 20);
            result.tier1.should.equal(0);
            result.tier2.should.equal(0);
            result.tier3.should.equal(3);
        });

        it('Tiers should be partitioned according to the number of relationships  BYOL relationships)', () => {
            const result = utils.getCustomerTiers(23, 40);
            result.tier1.should.equal(0);
            result.tier2.should.equal(0);
            result.tier3.should.equal(0);
        });
    });

    describe('marketplace function with mongoDb actions', () => {
        const generateMeteringId = () => Math.floor((1 + Math.random()) * 0x10000).toString(8);

        function createProcessedAwsResponse(customerIdentifier, dimension, quantyity, awsTimestamp, status) {
            return {
                MeteringRecordId: generateMeteringId(),
                Status: status,
                UsageRecord: {
                    CustomerIdentifier: customerIdentifier,
                    Dimension: dimension,
                    Quantity: quantyity,
                    Timestamp: awsTimestamp
                }
            };
        }

        function createUnprocessedAwsResponse(customerIdentifier, dimension, quantyity, awsTimestamp) {
            return {
                CustomerIdentifier: customerIdentifier,
                Dimension: dimension,
                Quantity: quantyity,
                Timestamp: awsTimestamp
            };
        }

        AWS.mock('MarketplaceMetering', 'batchMeterUsage', (params) => ({
            promise: () => {
                const testTimestamp = new Date().toISOString();
                const customerIdObj = {};
                // building results from the requested records themselves and adding some mixed data
                const usageRecords = params.UsageRecords;
                let results;
                let unproccessedRecords;
                if (usageRecords.length > 1) {
                    results = usageRecords.slice(0, usageRecords.length - 1);
                    unproccessedRecords = [createUnprocessedAwsResponse(
                        usageRecords[usageRecords.length - 1].CustomerIdentifier,
                        usageRecords[usageRecords.length - 1].Dimension, usageRecords[usageRecords.length - 1].Quantity, testTimestamp
                    )];
                } else {
                    results = usageRecords;
                    unproccessedRecords = [];
                }
                results = params.UsageRecords.map(record => {
                    if (customerIdObj[record.CustomerIdentifier]) {
                        if (customerIdObj[record.CustomerIdentifier][record.Dimension]) {
                            return createProcessedAwsResponse(record.CustomerIdentifier, record.Dimension, record.Quantity, testTimestamp, 'DuplicateRecord');
                        }
                        customerIdObj[record.CustomerIdentifier][record.Dimension] = true;
                        return createProcessedAwsResponse(record.CustomerIdentifier, record.Dimension, record.Quantity, testTimestamp, 'Success');
                    }
                    customerIdObj[record.CustomerIdentifier] = { [record.Dimension]: true };
                    return createProcessedAwsResponse(record.CustomerIdentifier, record.Dimension, record.Quantity, testTimestamp, 'Success');
                });
                results.push({
                    MeteringRecordId: generateMeteringId(),
                    Status: 'DuplicateRecord',
                    UsageRecord: {
                        CustomerIdentifier: params.UsageRecords[0].CustomerIdentifier,
                        Dimension: params.UsageRecords[0].Dimension,
                        Quantity: params.UsageRecords[0].Quantity,
                        Timestamp: testTimestamp
                    }
                });

                results.push({
                    MeteringRecordId: generateMeteringId(),
                    Status: 'CustomerNotSubscribed',
                    UsageRecord: {
                        CustomerIdentifier: 'unsubscrpibedUserId'
                    }
                });
                unproccessedRecords.push(createUnprocessedAwsResponse('someBandoned', 'Tier1', 5, testTimestamp));
                return {
                    Results: results,
                    UnprocessedRecords: unproccessedRecords
                };
            }
        }));

        const users = JSON.parse(fs.readFileSync(path.resolve('./test/resources', 'users.json'), 'utf8'));
        const relationships = JSON.parse(fs.readFileSync(path.resolve('./test/resources', 'relationships.json'), 'utf8'));
        const dataBrokers = JSON.parse(fs.readFileSync(path.resolve('./test/resources', 'dataBrokers.json'), 'utf8'));


        before(async() => {
            await Promise.all([User.deleteMany(), Relationship.deleteMany(), DataBroker.deleteMany()]);
            const userWithUnExpiredTrial = users.find(user => user._id === '58a40e987200292bd0f942f8');
            const date = new Date();
            userWithUnExpiredTrial.trialExpiry = date.setMonth(date.getMonth() + 12);
            return User.insertMany(users)
                .then(() => DataBroker.insertMany(dataBrokers))
                .then(() => Relationship.insertMany(relationships));
        });

        it('billSaaSUsers- should return one user with tiers and no BYOL', () => {
            const chargingTime = new Date().toISOString();
            return utils.billSaaSUsers(chargingTime)
                .then(obj => {
                    expect(obj).to.be.an('object');
                    expect(obj['58a40e987200292bd0f942f8']).to.be.an('undefined');
                    expect(obj['58a40eba7200292bd0f942fa']).to.be.an('undefined');
                    expect(obj['58a40edb7200292bd0f942fc']).to.be.an('undefined');
                    obj['58a40efb7200292bd0f942fd'].tier1.should.equal(5);
                    obj['58a40efb7200292bd0f942fd'].tier2.should.equal(15);
                    obj['58a40efb7200292bd0f942fd'].tier3.should.equal(4);
                });
        });

        it('byolDetailsForUnsubscribedUsers- should return 3 users with BYOL object- 1 with 0 relationships and 2 with 1 relationship', () => {
            const chargingTime = new Date().toISOString();
            return utils.byolDetailsForUnsubscribedUsers({}, chargingTime)
                .then(obj => {
                    expect(obj).to.be.an('object');
                    expect(obj['58a40e987200292bd0f942f8']).to.be.an('undefined');
                    obj['58a40efb7200292bd0f942fd'].byol.reservedRelationships.should.equal(0);
                    obj['58a40efb7200292bd0f942fd'].byol.usedRelationships.should.equal(0);
                    obj['58a40edb7200292bd0f942fc'].byol.reservedRelationships.should.equal(20);
                    obj['58a40edb7200292bd0f942fc'].byol.usedRelationships.should.equal(1);
                    obj['58a40eba7200292bd0f942fa'].byol.reservedRelationships.should.equal(20);
                    obj['58a40eba7200292bd0f942fa'].byol.usedRelationships.should.equal(1);
                });
        });

        it('Agenda flow- Cost collection should have records for each user, should have 1 user with 0 relationships and 5 others with relationships', () => {
            const chargingTime = new Date().toISOString();
            return utils.billSaaSUsers(chargingTime)
                .then(mongoObjects => utils.byolDetailsForUnsubscribedUsers(mongoObjects, chargingTime))
                .then(mongoObjects => utils.addUsersWithoutRelationshipsToMongoObject(mongoObjects, chargingTime))
                .then(mongoObjects => utils.filterAndInsert(mongoObjects))
                .then(() => Cost.find({}).exec())
                .then(costRecords => {
                    const costForExpiredUserWithByol = costRecords.find(cost => cost.user.toString() === '58a40edb7200292bd0f942fc');
                    costForExpiredUserWithByol.tier1.should.equal(0);
                    costForExpiredUserWithByol.tier2.should.equal(0);
                    costForExpiredUserWithByol.tier3.should.equal(0);
                    costForExpiredUserWithByol.byol.usedRelationships.should.equal(1);
                    costForExpiredUserWithByol.byol.reservedRelationships.should.equal(20);
                    const costForSubscribedNotNetappUser = costRecords.find(cost => cost.user.toString() === '58a40efb7200292bd0f942fd');
                    costForSubscribedNotNetappUser.tier1.should.equal(5);
                    costForSubscribedNotNetappUser.tier2.should.equal(15);
                    costForSubscribedNotNetappUser.tier3.should.equal(4);
                    costForSubscribedNotNetappUser.byol.usedRelationships.should.equal(0);
                    costForSubscribedNotNetappUser.byol.reservedRelationships.should.equal(0);
                    const costForSubscribedNetappUser = costRecords.find(cost => cost.user.toString() === '58a40eba7200292bd0f942fa');
                    costForSubscribedNetappUser.tier1.should.equal(0);
                    costForSubscribedNetappUser.tier2.should.equal(0);
                    costForSubscribedNetappUser.tier3.should.equal(0);
                    costForSubscribedNetappUser.byol.usedRelationships.should.equal(1);
                    costForSubscribedNetappUser.byol.reservedRelationships.should.equal(20);
                    const costForUserWithFreeTrial = costRecords.find(cost => cost.user.toString() === '58a40e987200292bd0f942f8');
                    costForUserWithFreeTrial.tier1.should.equal(0);
                    costForUserWithFreeTrial.tier2.should.equal(0);
                    costForUserWithFreeTrial.tier3.should.equal(0);
                    costForUserWithFreeTrial.byol.usedRelationships.should.equal(0);
                    costForUserWithFreeTrial.byol.reservedRelationships.should.equal(0);
                    const costForUserWithSubscriptionAndBYOL = costRecords.find(cost => cost.user.toString() === '58a4610c740ffd7cc870e5fb');
                    costForUserWithSubscriptionAndBYOL.tier1.should.equal(0);
                    costForUserWithSubscriptionAndBYOL.tier2.should.equal(0);
                    costForUserWithSubscriptionAndBYOL.tier3.should.equal(4);
                    costForUserWithSubscriptionAndBYOL.byol.usedRelationships.should.equal(20);
                    costForUserWithSubscriptionAndBYOL.byol.reservedRelationships.should.equal(20);
                    const costForSubscribeUserWithExistingCustomerIdentifierAndNoRelationships = costRecords.find(cost => cost.user.toString() === '58a47410bee5ae664080cc26');
                    costForSubscribeUserWithExistingCustomerIdentifierAndNoRelationships.tier1.should.equal(0);
                    costForSubscribeUserWithExistingCustomerIdentifierAndNoRelationships.tier2.should.equal(0);
                    costForSubscribeUserWithExistingCustomerIdentifierAndNoRelationships.tier3.should.equal(4);
                    costForSubscribeUserWithExistingCustomerIdentifierAndNoRelationships.byol.usedRelationships.should.equal(20);
                    costForSubscribeUserWithExistingCustomerIdentifierAndNoRelationships.byol.reservedRelationships.should.equal(20);
                });
        });
    });

    describe('addRecordsOfUsersWithoutRelationships', () => {
        const userIds = [mongoose.Types.ObjectId('58933deb96d0230a142f8410'), mongoose.Types.ObjectId('58933ddf96d0230a142f840f')];

        it('should add another record of user that doesnt exists in the records array', () => {
            const records = [record1, record2];
            utils.addRecordsOfUsersWithoutRelationships(records, timestamp, userIds);
            const addedRecord = records.find(record => record.entry.CustomerIdentifier === '');
            records.length.should.equal(3);
            addedRecord.uid.toString().should.equal('58933deb96d0230a142f8410');
        });

        it('should not add another record to the records array', () => {
            const records = [record1, record2];
            records.push({
                entry: {
                    CustomerIdentifier: 'customerId',
                    Dimension: 'Tier1',
                    Quantity: 1,
                    Timestamp: timestamp
                },
                uid: '58933deb96d0230a142f8410',
                byol: { reservedRelationships: 0, usedRelationships: 0 },
                isSuccess: false
            });
            utils.addRecordsOfUsersWithoutRelationships(records, timestamp, userIds);
            records.length.should.equal(3);
        });
    });
});
