import Promise from 'bluebird';
import User from '../../src/models/user';
import DataBroker from '../../src/models/data-broker';
import * as consts from '../../src/utils/consts';
import { should } from '../globals';

describe('DataBroker Model', () => {
    let user;
    const name = 'DM1';
    const instanceId = 'i-122345656';
    const privateIp = '10.0.0.1';
    const region = 'us-west-1';
    const availabilityZone = 'az';
    const vpcId = 'vpc-1234567';
    const vpcCidrBlock = 'vpcCidrBlock';
    const subnetId = 'sub-1234567';
    const subnetCidrBlock = 'subnetCidrBlock';
    const lastPing = {
        wasabi: new Date()
    };

    beforeEach(() => Promise.all([User.deleteMany(), DataBroker.deleteMany()])
        .then(() => User.create({
            name: 'User Name',
            email: 'user@email.com',
            password: 'Netapp234',
            company: 'company',
            trialExpiry: new Date(),
            awsAccountId: '282316784512',
            portalId: '1'
        }))
        .then(u => user = u));

    describe('CRUD', () => {
        it('create', () => DataBroker.create({ user, name })
            .then(dataBroker => {
                should.exist(dataBroker);
                dataBroker.user.id.should.equal(user.id);
                dataBroker.name.should.equal(name);
                should.not.exist(dataBroker.instanceId);
                should.not.exist(dataBroker.privateIp);
                should.not.exist(dataBroker.region);
                should.not.exist(dataBroker.availabilityZone);
                should.not.exist(dataBroker.vpcId);
                should.not.exist(dataBroker.vpcCidrBlock);
                should.not.exist(dataBroker.subnetId);
                should.not.exist(dataBroker.subnetCidrBlock);
                should.exist(dataBroker.createdAt);
                should.exist(dataBroker.updatedAt);
            }));

        it('create with optional fields', () => DataBroker.create({
            user,
            name,
            instanceId,
            privateIp,
            region,
            availabilityZone,
            vpcId,
            vpcCidrBlock,
            subnetId,
            subnetCidrBlock,
            lastPing
        })
            .then(dataBroker => {
                should.exist(dataBroker);
                dataBroker.user.id.should.equal(user.id);
                dataBroker.name.should.equal(name);
                dataBroker.instanceId.should.equal(instanceId);
                dataBroker.privateIp.should.equal(privateIp);
                dataBroker.region.should.equal(region);
                dataBroker.availabilityZone.should.equal(availabilityZone);
                dataBroker.vpcId.should.equal(vpcId);
                dataBroker.vpcCidrBlock.should.equal(vpcCidrBlock);
                dataBroker.subnetId.should.equal(subnetId);
                dataBroker.subnetCidrBlock.should.equal(subnetCidrBlock);
                dataBroker.lastPing.wasabi.should.equalTime(lastPing.wasabi);
                should.exist(dataBroker.createdAt);
                should.exist(dataBroker.updatedAt);
            }));

        it('get', () => DataBroker.create({
            user,
            name,
            instanceId,
            privateIp,
            region,
            availabilityZone,
            vpcId,
            vpcCidrBlock,
            subnetId,
            subnetCidrBlock,
            lastPing
        })
            .then(dataBroker => DataBroker.findOne({ _id: dataBroker.id }).populate('user'))
            .then(dataBroker => {
                should.exist(dataBroker);
                dataBroker.user.id.should.equal(user.id);
                dataBroker.name.should.equal(name);
                dataBroker.instanceId.should.equal(instanceId);
                dataBroker.privateIp.should.equal(privateIp);
                dataBroker.region.should.equal(region);
                dataBroker.availabilityZone.should.equal(availabilityZone);
                dataBroker.vpcId.should.equal(vpcId);
                dataBroker.vpcCidrBlock.should.equal(vpcCidrBlock);
                dataBroker.subnetId.should.equal(subnetId);
                dataBroker.subnetCidrBlock.should.equal(subnetCidrBlock);
                dataBroker.lastPing.wasabi.should.equalTime(lastPing.wasabi);
                should.exist(dataBroker.createdAt);
                should.exist(dataBroker.updatedAt);
            }));

        it('update', () => {
            const updatedRegion = 'us-west-2';
            return DataBroker.create({
                user,
                name,
                instanceId,
                privateIp,
                region,
                availabilityZone,
                vpcId,
                vpcCidrBlock,
                subnetId,
                subnetCidrBlock,
                lastPing
            })
                .then(dataBroker => {
                    dataBroker.region = updatedRegion;
                    return dataBroker.save();
                })
                .then(dataBroker => DataBroker.findOne({ _id: dataBroker.id }).populate('user'))
                .then(dataBroker => {
                    should.exist(dataBroker);
                    dataBroker.user.id.should.equal(user.id);
                    dataBroker.name.should.equal(name);
                    dataBroker.instanceId.should.equal(instanceId);
                    dataBroker.privateIp.should.equal(privateIp);
                    dataBroker.region.should.equal(updatedRegion);
                    dataBroker.availabilityZone.should.equal(availabilityZone);
                    dataBroker.vpcId.should.equal(vpcId);
                    dataBroker.vpcCidrBlock.should.equal(vpcCidrBlock);
                    dataBroker.subnetId.should.equal(subnetId);
                    dataBroker.subnetCidrBlock.should.equal(subnetCidrBlock);
                    dataBroker.lastPing.wasabi.should.equalTime(lastPing.wasabi);
                    should.exist(dataBroker.createdAt);
                    should.exist(dataBroker.updatedAt);
                });
        });

        it('add optional fields', () => DataBroker.create({
            user,
            name,
            instanceId,
            privateIp,
            region,
            availabilityZone,
            vpcId,
            vpcCidrBlock,
            subnetId,
            subnetCidrBlock
        })
            .then(dataBroker => {
                dataBroker.lastPing = lastPing;
                return dataBroker.save();
            })
            .then(dataBroker => DataBroker.findOne({ _id: dataBroker.id }).populate('user'))
            .then(dataBroker => {
                should.exist(dataBroker);
                dataBroker.user.id.should.equal(user.id);
                dataBroker.name.should.equal(name);
                dataBroker.instanceId.should.equal(instanceId);
                dataBroker.privateIp.should.equal(privateIp);
                dataBroker.region.should.equal(region);
                dataBroker.availabilityZone.should.equal(availabilityZone);
                dataBroker.vpcId.should.equal(vpcId);
                dataBroker.vpcCidrBlock.should.equal(vpcCidrBlock);
                dataBroker.subnetId.should.equal(subnetId);
                dataBroker.subnetCidrBlock.should.equal(subnetCidrBlock);
                dataBroker.lastPing.wasabi.should.equalTime(lastPing.wasabi);
                should.exist(dataBroker.createdAt);
                should.exist(dataBroker.updatedAt);
            }));

        it('delete', () => DataBroker.create({
            user,
            name,
            instanceId,
            privateIp,
            region,
            availabilityZone,
            vpcId,
            vpcCidrBlock,
            subnetId,
            subnetCidrBlock,
            lastPing
        })
            .then(dataBroker => DataBroker.deleteOne({ _id: dataBroker.id }))
            .then(() => DataBroker.findOne({ user }))
            .should.eventually.not.exist);
    });

    describe('Custom Methods', () => {
        consts.DATA_BROKER_WASABI_LIVENESS_TIMEOUT_MS = 10000;

        it('isWasabiAlive', () => DataBroker.create({
            user,
            name,
            instanceId,
            privateIp,
            region,
            availabilityZone,
            vpcId,
            vpcCidrBlock,
            subnetId,
            subnetCidrBlock
        })
            .then(dataBroker => {
                dataBroker.isWasabiAlive().should.equal(false);
                dataBroker.lastPing.wasabi = new Date(Date.now() - (consts.DATA_BROKER_WASABI_LIVENESS_TIMEOUT_MS * 2));
                dataBroker.isWasabiAlive().should.equal(false);
                dataBroker.lastPing.wasabi = new Date(Date.now() - (consts.DATA_BROKER_WASABI_LIVENESS_TIMEOUT_MS / 2));
                dataBroker.isWasabiAlive().should.equal(true);
            }));
    });
});

