import Promise from 'bluebird';
import moment from 'moment';
import { should } from '../globals';
import { COPY_RELATIONSHIP_MESSAGE, SYNC_RELATIONSHIP_MESSAGE } from '../../src/utils/consts';
import User from '../../src/models/user';
import Relationship from '../../src/models/relationship';
import DataBroker from '../../src/models/data-broker';
import Group from '../../src/models/group';

describe('Relationship Model', () => {
    let user,
        dataBroker;
    const nfsServer = 'source';
    const source = {
        protocol: 'nfs',
        nfs: {
            host: nfsServer,
            export: '/'
        }
    };
    const target = {
        protocol: 's3',
        s3: {
            bucket: 'bucket',
            prefix: 'folder'
        }
    };
    const autoSync = 'OFF';

    const dataBrokerName = 'dataBrokerName';
    const sync = {
        inDays: 0, inHours: 1, inMinutes: 0, nextTime: new Date(), autoSync
    };
    const settings = { gracePeriod: 30 };

    beforeEach(() => Promise.all([User.deleteMany(), Relationship.deleteMany(), DataBroker.deleteMany()])
        .then(() => User.create({
            name: 'User Name',
            email: 'user@email.com',
            password: 'Netapp234',
            company: 'company',
            trialExpiry: new Date(),
            awsAccountId: '282316784512',
            portalId: '1'
        }))
        .then(u => user = u)
        .then(() => DataBroker.create({ user, name: dataBrokerName, group:'aa' }))
        .then(db => dataBroker = db)
        .then(() => Group.create({user, name:'aa', dataBrokers: [dataBroker.id]})));

    describe('CRUD', () => {
        it('create', () => Relationship.create({
            user, dataBroker, source, target, sync, settings
        })
            .then(relationship => {
                should.exist(relationship);
                relationship.user.id.should.equal(user.id);
                relationship.dataBroker.id.should.equal(dataBroker.id);
                relationship.source.protocol.should.equal('nfs');
                relationship.source.nfs.host.should.equal(nfsServer);
                relationship.source.nfs.export.should.equal('/');
                relationship.target.protocol.should.equal('s3');
                relationship.target.s3.bucket.should.equal('bucket');
                relationship.target.s3.prefix.should.equal('folder');
                relationship.sync.autoSync.should.equal(autoSync);
                should.exist(relationship.createdAt);
                should.exist(relationship.updatedAt);
            }));

        it('get', () => Relationship.create({
            user, dataBroker, source, target, sync, settings
        })
            .then(relationship => Relationship.findOne({ _id: relationship.id }).populate('user'))
            .then(relationship => {
                should.exist(relationship);
                relationship.user.id.should.equal(user.id);
                relationship.dataBroker.toString().should.equal(dataBroker.id);
                relationship.source.protocol.should.equal('nfs');
                relationship.source.nfs.host.should.equal(nfsServer);
                relationship.source.nfs.export.should.equal('/');
                relationship.target.protocol.should.equal('s3');
                relationship.target.s3.bucket.should.equal('bucket');
                relationship.target.s3.prefix.should.equal('folder');
                relationship.sync.autoSync.should.equal(autoSync);
                should.exist(relationship.createdAt);
                should.exist(relationship.updatedAt);
            }));

        it('update', () => {
            const updatedNfsServer = 'fs1.netapp.com';
            const updatedSource = {
                protocol: 'nfs',
                nfs: {
                    host: updatedNfsServer, export: 'export1', path: 'folder2'
                }
            };
            return Relationship.create({
                user, dataBroker, source, target, sync, settings
            })
                .then(relationship => {
                    relationship.source = updatedSource;
                    return relationship.save();
                })
                .then(relationship => Relationship.findOne({ _id: relationship.id }).populate('user'))
                .then(relationship => {
                    should.exist(relationship);
                    relationship.user.id.should.equal(user.id);
                    relationship.dataBroker.toString().should.equal(dataBroker.id);
                    relationship.source.protocol.should.equal('nfs');
                    relationship.source.nfs.host.should.equal(updatedNfsServer);
                    relationship.source.nfs.export.should.equal('export1');
                    relationship.source.nfs.path.should.equal('folder2');
                    relationship.target.protocol.should.equal('s3');
                    relationship.target.s3.bucket.should.equal('bucket');
                    relationship.target.s3.prefix.should.equal('folder');
                    relationship.sync.autoSync.should.equal(autoSync);
                    should.exist(relationship.createdAt);
                    should.exist(relationship.updatedAt);
                });
        });

        it('delete', () => Relationship.create({
            user, dataBroker, source, target, sync, settings
        })
            .then(relationship => Relationship.deleteOne({ _id: relationship.id }))
            .then(() => Relationship.findOne({ user }))
            .should.eventually.not.exist);
    });

    describe('Constraints', () => {
        it('no duplicate relationship with same parameters', () => Relationship.create({
            user, dataBroker, source, target, sync, settings
        })
            .then(() => Relationship.create({
                user, nfsServer, dataBroker, source, target, sync, settings
            }))
            .should.eventually.be.rejected);
    });

    describe('JSON representation', () => {
        const sourceString = { protocol: 'nfs', nfs: { host: 'source', export: '/' } };
        const targetString = { protocol: 's3', s3: { bucket: 'bucket', prefix: 'folder' } };

        it('successfully transform', () => Relationship.create({
            user, dataBroker, source, target, sync, settings
        })
            .then(relationship => {
                const json = relationship.toJSON();
                should.exist(json);
                should.exist(json.dataBroker);
                json.source.should.deep.equal(sourceString);
                json.target.should.deep.equal(targetString);
                json.autoSync.should.equal('OFF');
                should.not.exist(json.user);
                should.not.exist(json._id);
                should.not.exist(json.__v);
                should.exist(json.createdAt);
                should.not.exist(json.updatedAt);
            }));
    });
    describe('Custom Methods', () => {
        describe('findTypeAndActiveMessage', () => {
            const innerMessageWithStep1 = { setup: { step: 'setup', status: 'done', timestamp: 1497698146812 } };
            const innerMessageWithStep2 = { setup: { step: 'setup', status: 'failed', timestamp: 1497698146812 } };

            function validateTypeAnActiveMessage(messages, type, activeMessage) {
                const relationship = new Relationship({
                    source,
                    target,
                    sync,
                    user,
                    dataBroker,
                    messages,
                    settings
                });
                const typeAndActiveMessage = relationship.findTypeAndActiveMessage();
                typeAndActiveMessage.type.should.equal(type);
                typeAndActiveMessage.activeMessage.should.equal(activeMessage);
            }

            it('Only copy-relationship messages -should return Initial Copy', () => {
                const messages = { [COPY_RELATIONSHIP_MESSAGE]: innerMessageWithStep1 };
                validateTypeAnActiveMessage(messages, 'Initial Copy', innerMessageWithStep1);
            });

            it('Only sync-relationship messages -should return Sync', () => {
                const messages = { [SYNC_RELATIONSHIP_MESSAGE]: innerMessageWithStep1 };
                validateTypeAnActiveMessage(messages, 'Sync', innerMessageWithStep1);
            });

            it('copy-relationship and sync-relationship messages -should return Sync', () => {
                const messages = {
                    [COPY_RELATIONSHIP_MESSAGE]: innerMessageWithStep1,
                    [SYNC_RELATIONSHIP_MESSAGE]: innerMessageWithStep2
                };
                validateTypeAnActiveMessage(messages, 'Sync', innerMessageWithStep2);
            });

            it('old message and sync-relationship message- should return Sync', () => {
                const messages = { status: {}, [SYNC_RELATIONSHIP_MESSAGE]: innerMessageWithStep1 };
                validateTypeAnActiveMessage(messages, 'Sync', innerMessageWithStep1);
            });
        });

        describe('getRelationshipsForScheduledSync', () => {
            function createRelationshipAndFindToScheduledSync(inDays, inHours, inMinutes, autoSync, nextTime) {
                const sync = {
                    inDays, inHours, inMinutes, autoSync, nextTime
                };
                return new Relationship({
                    source, target, sync, user, dataBroker, settings
                }).save()
                    .then(() => Relationship.getRelationshipsForScheduledSync());
            }


            it('Create relationship with  syncInDays = 0, syncInHours = 0, syncInMinutes = 0 and autoSync = OFF, no relationship should be returning', () => createRelationshipAndFindToScheduledSync(0, 0, 0, 'OFF', Date.now())
                .then(result => {
                    result.length.should.equal(0);
                }));

            it('Create relationship with  syncInDays = 1, syncInHours = 0, syncInMinutes = 0 and autoSync = OFF, no relationship should be returning', () => createRelationshipAndFindToScheduledSync(0, 0, 1, 'ON', Date.now())
                .then(result => {
                    result.length.should.equal(1);
                }));

            it('Create relationship with  syncInDays = 0, syncInHours = 1, syncInMinutes = 0 and autoSync = OFF, no relationship should be returning', () => createRelationshipAndFindToScheduledSync(0, 1, 0, 'ON', Date.now())
                .then(result => {
                    result.length.should.equal(1);
                }));

            it('Create relationship with  syncInDays = 0, syncInHours = 0, syncInMinutes = 1 and autoSync = OFF, no relationship should be returning', () => createRelationshipAndFindToScheduledSync(1, 0, 0, 'ON', Date.now())
                .then(result => {
                    result.length.should.equal(1);
                }));

            it('Create relationship with  syncInDays = 0, syncInHours = 0, syncInMinutes = 1 and autoSync = ON and Date is in a week, no relationship should be returning', () => {
                const nextTime = moment().add(7, 'days').valueOf();
                return createRelationshipAndFindToScheduledSync(0, 0, 1, 'ON', nextTime)
                    .then(result => {
                        result.length.should.equal(0);
                    });
            });

            it('Create relationship with  syncInDays = 0, syncInHours = 0, syncInMinutes = 1 and autoSync = ON and next date is last week,  relationship should return', () => {
                const nextTime = moment().subtract(7, 'days').valueOf();
                return createRelationshipAndFindToScheduledSync(1, 0, 0, 'ON', nextTime)
                    .then(result => {
                        result.length.should.equal(1);
                    });
            });
        });
    });
});
