import sinon from 'sinon';
import Promise from 'bluebird';
import User from '../../src/models/user';
import Timeline from '../../src/models/timeline';
import Message from '../../src/models/message';
import { should } from '../globals';

describe('Timeline Model', () => {
    let user;
    const requestId = 'requestId';
    const status = 'status';
    const summary = 'summary';
    const source = 'source';
    const target = 'target';
    const dataBroker = 'dataBroker';
    const failureMessage = 'failureMessage';
    const sendAsyncSpy = sinon.stub(Message, 'SendAsync');

    beforeEach(() => Promise.all([User.deleteMany(), Timeline.deleteMany()])
        .then(() => User.create({
            name: 'User Name',
            email: 'user@email.com',
            password: 'Netapp234',
            company: 'company',
            trialExpiry: new Date(),
            awsAccountId: '282316784512',
            portalId: '1'
        }))
        .then(u => user = u));

    describe('CRUD', () => {
        it('create', () => Timeline.create({
            user, requestId, status, summary
        })
            .then(timeline => {
                should.exist(timeline);
                timeline.user.id.should.equal(user.id);
                timeline.requestId.should.equal(requestId);
                timeline.status.should.equal(status);
                timeline.summary.should.equal(summary);
                should.exist(timeline.createdAt);
                should.exist(timeline.updatedAt);
                sendAsyncSpy.called.should.equal(true);
                sendAsyncSpy.calledWithExactly(undefined, user.id, requestId, 'timeline', timeline.toJSON(), false);
            }));

        it('create with optional fields', () => Timeline.create({
            user, requestId, status, summary, source, target, dataBroker, failureMessage
        })
            .then(timeline => {
                should.exist(timeline);
                timeline.user.id.should.equal(user.id);
                timeline.requestId.should.equal(requestId);
                timeline.status.should.equal(status);
                timeline.summary.should.equal(summary);
                timeline.summary.should.equal(summary);
                timeline.source.should.equal(source);
                timeline.target.should.equal(target);
                timeline.dataBroker.should.equal(dataBroker);
                should.exist(timeline.createdAt);
                should.exist(timeline.updatedAt);
                sendAsyncSpy.called.should.equal(true);
                sendAsyncSpy.calledWithExactly(undefined, user.id, requestId, 'timeline', timeline.toJSON(), false);
            }));

        it('get', () => Timeline.create({
            user, requestId, status, summary
        })
            .then(timeline => Timeline.findOne({ _id: timeline.id }).populate('user'))
            .then(timeline => {
                should.exist(timeline);
                timeline.user.id.should.equal(user.id);
                timeline.requestId.should.equal(requestId);
                timeline.status.should.equal(status);
                timeline.summary.should.equal(summary);
                should.exist(timeline.createdAt);
                should.exist(timeline.updatedAt);
            }));

        it('update', () => {
            const updateStatus = 'newStatus';
            return Timeline.create({
                user, requestId, status, summary
            })
                .then(timeline => {
                    timeline.status = updateStatus;
                    return timeline.save();
                })
                .then(timeline => Timeline.findOne({ _id: timeline.id }).populate('user'))
                .then(timeline => {
                    should.exist(timeline);
                    timeline.user.id.should.equal(user.id);
                    timeline.status.should.equal(updateStatus);
                    timeline.requestId.should.equal(requestId);
                    timeline.summary.should.equal(summary);
                    should.exist(timeline.createdAt);
                    should.exist(timeline.updatedAt);
                });
        });

        it('add optional fields', () => Timeline.create({
            user, requestId, status, summary
        })
            .then(timeline => {
                timeline.failureMessage = failureMessage;
                return timeline.save();
            })
            .then(timeline => Timeline.findOne({ _id: timeline.id }).populate('user'))
            .then(timeline => {
                should.exist(timeline);
                timeline.user.id.should.equal(user.id);
                timeline.requestId.should.equal(requestId);
                timeline.status.should.equal(status);
                timeline.failureMessage.should.equal(failureMessage);
                should.exist(timeline.createdAt);
                should.exist(timeline.updatedAt);
                sendAsyncSpy.called.should.equal(true);
                sendAsyncSpy.calledWithExactly(undefined, user.id, requestId, 'timeline', timeline.toJSON(), false);
            }));
    });

    describe('JSON representation', () => {
        it('successfully transform', () => Timeline.create({
            user, requestId, status, summary, dataBroker, source, target
        })
            .then(timeline => {
                const json = timeline.toJSON();
                should.exist(json);
                should.exist(json.dataBroker);
                json.source.should.equal(source);
                json.target.should.equal(target);
                should.not.exist(json.user);
                should.not.exist(json._id);
                should.not.exist(json.__v);
                should.exist(json.createdAt);
                should.not.exist(json.updatedAt);
                should.exist(json.id);
            }));
    });
});

