import moment from 'moment';
import User from '../../src/models/user';
import { should } from '../globals';

describe('User Model', () => {
    const name = 'name';
    const email = 'admin@netapp.com';
    const password = 'Netapp234';
    const company = 'netapp';
    const trialExpiry = moment().subtract(1, 'years');
    const awsAccountId = '282316784512';
    const portalId = '1';

    beforeEach(() => User.deleteMany());

    describe('CRUD', () => {
        it('create', () => User.create({
            name, email, password, awsAccountId, trialExpiry, portalId
        })
            .then(user => {
                should.exist(user);
                user.name.should.equal(name);
                user.email.should.equal(email);
                should.exist(user.createdAt);
                should.exist(user.updatedAt);
            }));

        it('create with optional fields', () => User.create({
            name, email, password, awsAccountId, trialExpiry, company, portalId
        })
            .then(user => {
                should.exist(user);
                user.name.should.equal(name);
                user.email.should.equal(email);
                user.company.should.equal(company);
                should.exist(user.createdAt);
                should.exist(user.updatedAt);
            }));

        it('get', () => User.create({
            name, email, password, awsAccountId, trialExpiry, portalId
        })
            .then(user => User.findOne({ _id: user.id }))
            .then(user => {
                should.exist(user);
                user.name.should.equal(name);
                user.email.should.equal(email);
                user.portalId.should.equal(portalId);
                should.exist(user.createdAt);
                should.exist(user.updatedAt);
            }));

        it('update', () => {
            const updatedName = 'newUser';
            return User.create({
                name, password, email, awsAccountId, trialExpiry, portalId
            })
                .then(user => {
                    user.name = updatedName;
                    return user.save();
                })
                .then(user => User.findOne({ _id: user.id }))
                .then(user => {
                    should.exist(user);
                    user.name.should.equal(updatedName);
                    user.email.should.equal(email);
                    should.exist(user.createdAt);
                    should.exist(user.updatedAt);
                });
        });

        it('add optional fields', () => User.create({
            name, password, email, awsAccountId, trialExpiry, portalId
        })
            .then(user => {
                user.company = company;
                return user.save();
            })
            .then(user => User.findOne({ _id: user.id }))
            .then(user => {
                should.exist(user);
                user.name.should.equal(name);
                user.email.should.equal(email);
                user.company.should.equal(company);
                user.portalId.should.equal(portalId);
                should.exist(user.createdAt);
                should.exist(user.updatedAt);
            }));

        it('delete', () => User.create({
            name, password, email, awsAccountId, trialExpiry, portalId
        })
            .then(user => User.deleteOne({ _id: user.id }))
            .then(() => User.findOne({ email }))
            .should.eventually.not.exist);
    });

    describe('Custom Methods', () => {
        it('addTrialDaysForNetappUsersAndSaveIfNeeded- email and company are NetApp', () => User.create({
            name, email, password, company, awsAccountId, trialExpiry, portalId
        })
            .then(user => {
                should.exist(user);
                return user.addTrialDaysForNetappUsersAndSaveIfNeeded()
                    .then(newUser => {
                        const monthFromNow = moment().startOf('day').add(14, 'day');
                        const expectedDay = monthFromNow.weekday();
                        const expectedMonth = monthFromNow.month();
                        const actualDay = newUser.trialExpiry.getDay();
                        const actualMonth = newUser.trialExpiry.getMonth();
                        actualDay.should.equal(expectedDay);
                        actualMonth.should.equal(expectedMonth);
                    });
            }));

        it('addTrialDaysForNetappUsersAndSaveIfNeeded- only company is NetApp', () => {
            const email = 'admin@gmail.com';
            return User.create({
                name, email, password, company, awsAccountId, trialExpiry, portalId
            })
                .then(user => {
                    should.exist(user);
                    return user.addTrialDaysForNetappUsersAndSaveIfNeeded()
                        .then(newUser => {
                            const monthFromNow = moment().startOf('day').add(14, 'day');
                            const expectedDay = monthFromNow.weekday();
                            const expectedMonth = monthFromNow.month();
                            const actualDay = newUser.trialExpiry.getDay();
                            const actualMonth = newUser.trialExpiry.getMonth();
                            actualDay.should.equal(expectedDay);
                            actualMonth.should.equal(expectedMonth);
                        });
                });
        });

        it('addTrialDaysForNetappUsersAndSaveIfNeeded- only the email is related to NetApp', () => {
            const company = 'ABC';
            return User.create({
                name, email, password, company, awsAccountId, trialExpiry, portalId
            })
                .then(user => {
                    should.exist(user);
                    return user.addTrialDaysForNetappUsersAndSaveIfNeeded()
                        .then(newUser => {
                            const monthFromNow = moment().startOf('day').add(14, 'day');
                            const expectedDay = monthFromNow.weekday();
                            const expectedMonth = monthFromNow.month();
                            const actualDay = newUser.trialExpiry.getDay();
                            const actualMonth = newUser.trialExpiry.getMonth();
                            actualDay.should.equal(expectedDay);
                            actualMonth.should.equal(expectedMonth);
                        });
                });
        });

        it('addTrialDaysForNetappUsersAndSaveIfNeeded- both email and company are not related NetApp', () => {
            const company = 'ABC';
            const email = 'admin@gmail.com';
            return User.create({
                name, email, password, company, awsAccountId, trialExpiry, portalId
            })
                .then(user => {
                    should.exist(user);
                    return user.addTrialDaysForNetappUsersAndSaveIfNeeded()
                        .then(newUser => {
                            const expectedDay = newUser.trialExpiry.getDay();
                            const expectedMonth = newUser.trialExpiry.getMonth();
                            const actualDay = user.trialExpiry.getDay();
                            const actualMonth = user.trialExpiry.getMonth();
                            actualDay.should.equal(expectedDay);
                            actualMonth.should.equal(expectedMonth);
                        });
                });
        });

        it('calculateDaysLeftForTrial- expired trial', () => User.create({
            name, email, password, company, awsAccountId, trialExpiry, portalId
        })
            .then(user => {
                should.exist(user);
                user.calculateDaysLeftForTrial().should.equal(-1);
            }));

        it('calculateDaysLeftForTrial- valid trial', () => {
            const trialExpiry = moment().add(3, 'days');
            return User.create({
                name, email, password, company, awsAccountId, trialExpiry, portalId
            })
                .then(user => {
                    should.exist(user);
                    user.calculateDaysLeftForTrial().should.equal(3);
                });
        });

        it('calculateDaysLeftForTrial- 0 days left for the trial', () => {
            const trialExpiry = moment();
            return User.create({
                name, email, password, company, awsAccountId, trialExpiry, portalId
            })
                .then(user => {
                    should.exist(user);
                    user.calculateDaysLeftForTrial().should.equal(0);
                });
        });

        it('useraWithTrialExpired - find billable users', () => {
            const customerIdentifier = 'aaaa';
            return User.create({
                name, email, password, company, awsAccountId, trialExpiry, customerIdentifier, portalId
            })
                .then(netappUser => {
                    should.exist(netappUser);
                    return User.create({
                        name: 'netappUser2',
                        email: 'a@netapp.com',
                        password,
                        company: 'test',
                        awsAccountId,
                        trialExpiry,
                        portalId: '2',
                        subscription: { customerIdentifier }
                    })
                        .then(netappUser2 => {
                            should.exist(netappUser2);
                            return User.create({
                                name: 'notNetapp',
                                email: 'not@test.com',
                                password,
                                company: 'test',
                                awsAccountId,
                                trialExpiry,
                                portalId: '3',
                                subscription: { customerIdentifier }
                            })
                                .then(notNetappUser => {
                                    should.exist(notNetappUser);
                                    return User.BillableUsers()
                                        .then(billableUsers => {
                                            const billableUsersNames = billableUsers.map(user => user.name);
                                            billableUsersNames.should.contains('notNetapp');
                                            billableUsersNames.should.not.contains('name');
                                            billableUsersNames.should.not.contains('netappUser2');
                                        });
                                });
                        });
                });
        });
    });
});

