import Promise from 'bluebird';
import { repeat } from 'lodash';
import { validateRelationship, validateS3Tags, validateSyncSchedule } from '../../src/operations/relationship-operations';
import Relationship from '../../src/models/relationship';
import DataBroker from '../../src/models/data-broker';
import User from '../../src/models/user';
import { expect } from '../globals';
import { BadRequest } from '../../src/utils/errors';

let user,
    dataBroker, asterixDataBroker;
const autoSync = 'OFF';


const dataBrokerName = 'dataBroker'; //TODO need to change tests
const asterixDataBrokerName = 'asterix';
const sync = {
    inDays: 0, inHours: 1, inMinutes: 0, nextTime: new Date(), autoSync
};
const settings = { gracePeriod: 30 };

describe('relationships-operations', () => {
    beforeEach(() => Promise.all([User.deleteMany(), Relationship.deleteMany(), DataBroker.deleteMany()])
        .then(() => User.create({
            name: 'User Name',
            email: 'user@email.com',
            password: 'Netapp234',
            company: 'company',
            trialExpiry: new Date(),
            awsAccountId: '282316784512',
            portalId: '1'
        }))
        .then(u => user = u)
        .then(() => Promise.all([DataBroker.create({ user, name: dataBrokerName, createdAt: 1531402152 }), DataBroker.create({ user, name: asterixDataBrokerName, createdAt:1531142952  })]))
        .then(dataBrokers => [dataBroker, asterixDataBroker] = dataBrokers));

    describe('Relationship validations', () => {
        function validateRelationshipDoesNotThrow(source, target, isAsterix) {
            const dataBrokerForRelationship = isAsterix ? asterixDataBroker : dataBroker;

            const relationship = new Relationship({
                source, target, sync, settings, dataBroker: dataBrokerForRelationship
            });
            expect(validateRelationship.bind(this, relationship)).not.to.throw();
        }

        async function validateRelationshipThrows(expectedError, source, target, isAsterix) {
            const dataBrokerForRelationship = isAsterix ? asterixDataBroker : dataBroker;
            const relationship = new Relationship({
                source, target, sync, settings, dataBroker: dataBrokerForRelationship
            });
            expect(validateRelationship.bind(this, relationship)).to.throw(BadRequest, expectedError);
        }

        describe('validate source & target URLs and type combinations asterix', () => {
            it('validates invalid source type asterix', async() => {
                await validateRelationshipThrows(/Source type can only be one of:/, {
                    protocol: 'xyz',
                    xyz: { host: 'dksjhf', export: 'kslf' }
                }, { protocol: 'xyz', xyz: { host: 'dksjhf', export: 'kslf' } }, true);
            });

            it('validates invalid SGWS source', async() => {
                await validateRelationshipThrows(/Source type can only be one of:/, {
                    protocol: 'sgws',
                    sgws: { host: '1.2.3.4', bucket: 'kslf' }
                }, { protocol: 's3', sgws: { bucket: 'kslf' } }, true);
            });

            it('validates invalid CIFS source', () => {
                validateRelationshipThrows('source or target with protocol cifs must contain host', {
                    protocol: 'cifs',
                    cifs: { share: 'share' }
                }, { protocol: 's3', sgws: { bucket: 'kslf' } }, true);
            });

            it('validates invalid NFS source', () => {
                validateRelationshipThrows('source or target with protocol nfs must contain host', {
                    protocol: 'nfs',
                    nfs: { export: 'export' }
                }, { protocol: 's3', sgws: { bucket: 'kslf' } }, true);
            });

            it('validates invalid S3 source', () => {
                validateRelationshipThrows('source or target with protocol s3 must contain bucket', {
                    protocol: 's3',
                    s3: {}
                }, { protocol: 'nfs', nfs: { host: '1.1.1.1', export: 'export' } }, true);
            });

            it('validates invalid target format', () => {
                validateRelationshipThrows('source or target with protocol s3 must contain bucket', {
                    protocol: 'nfs',
                    nfs: { host: '1.1.1.1', export: 'export' }
                }, { protocol: 's3', s3: {} }, true);
            });

            it('validates invalid target type', () => {
                validateRelationshipThrows(/Target type can only be one of:/, {
                    protocol: 's3',
                    sgws: { bucket: 'kslf' }
                }, { protocol: 'xyz', xyz: { host: 'dksjhf', export: 'kslf' } }, true);
            });

            it('validates limited target types for CIFS source', () => {
                validateRelationshipThrows('Target parameter should be an object', {
                    protocol: 'cifs',
                    cifs: { share: 'share' }
                }, 'xyz://dksjhf/kslf', true);
            });

            it('validates S3 target for CIFS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'cifs',
                    cifs: { host: '1.1.1.1', share: '/' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } }, true);
            });

            it('validates SGWS target for CIFS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'cifs',
                    cifs: { host: '1.1.1.1', share: '/' }
                }, { protocol: 'sgws', sgws: { host: 'hostname', bucket: 'bucket' } }, true);
            });

            it('validates limited target types for NFS source', () => {
                validateRelationshipThrows('Target type can only be one of: s3, sgws', {
                    protocol: 'nfs',
                    nfs: { host: '1.1.1.1', export: 'export' }
                }, { protocol: 'xyz', xyz: { host: 'dksjhf', export: 'kslf' } }, true);
            });

            it('validates S3 target for NFS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'nfs',
                    nfs: { host: '1.1.1.1', export: '/' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } }, true);
            });

            it('validates SGWS target for NFS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'nfs',
                    nfs: { host: '1.1.1.1', export: '/' }
                }, { protocol: 'sgws', sgws: { host: '1.1.1.1', bucket: 'some-bucket' } }, true);
            });

            it('validates limited target types for S3 source', () => {
                validateRelationshipThrows('Target type can only be one of: cifs, nfs', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'xyz', xyz: { host: 'dksjhf', export: 'kslf' } }, true);
            });

            it('validates NFS target for S3 source', () => {
                validateRelationshipDoesNotThrow({ protocol: 's3', s3: { bucket: 'some-bucket' } }, {
                    protocol: 'nfs',
                    nfs: { host: '1.1.1.1', export: '/' }
                }, true);
            });

            it('validates CIFS target for S3 source', () => {
                validateRelationshipDoesNotThrow({ protocol: 's3', s3: { bucket: 'some-bucket' } }, {
                    protocol: 'cifs',
                    cifs: { host: '1.1.1.1', share: '/' }
                }, true);
            });
        });

        describe('validate source & target URLs and type combinations regular dataBroker', () => {
            it('validates invalid source type', () => {
                validateRelationshipThrows(/Source type can only be one of:/, {
                    protocol: 'xyz',
                    xyz: { host: 'dksjhf', export: 'kslf' }
                }, { protocol: 'xyz', xyz: { host: 'dksjhf', export: 'kslf' } });
            });

            it('validates invalid SGWS source', () => {
                validateRelationshipThrows(/Target type can only be one of:/, {
                    protocol: 'sgws',
                    sgws: { host: '1.2.3.4', bucket: 'kslf' }
                }, { protocol: 's3', sgws: { bucket: 'kslf' } });
            });

            it('validates invalid CIFS source', () => {
                validateRelationshipThrows('source or target with protocol cifs must contain host', {
                    protocol: 'cifs',
                    cifs: { share: 'share' }
                }, { protocol: 's3', sgws: { bucket: 'kslf' } });
            });

            it('validates invalid NFS source', () => {
                validateRelationshipThrows('source or target with protocol nfs must contain host', {
                    protocol: 'nfs',
                    nfs: { export: 'export' }
                }, { protocol: 's3', sgws: { bucket: 'kslf' } });
            });

            it('validates invalid S3 source', () => {
                validateRelationshipThrows('source or target with protocol s3 must contain bucket', {
                    protocol: 's3',
                    s3: {}
                }, { protocol: 'nfs', nfs: { host: '1.1.1.1', export: 'export' } });
            });

            it('validates invalid target format', () => {
                validateRelationshipThrows('source or target with protocol s3 must contain bucket', {
                    protocol: 'nfs',
                    nfs: { host: '1.1.1.1', export: 'export' }
                }, { protocol: 's3', s3: {} });
            });

            it('validates invalid target type', () => {
                validateRelationshipThrows(/Target type can only be one of:/, {
                    protocol: 's3',
                    sgws: { bucket: 'kslf' }
                }, { protocol: 'xyz', xyz: { host: 'dksjhf', export: 'kslf' } });
            });

            it('validates limited target types for CIFS source', () => {
                validateRelationshipThrows('Target parameter should be an object', {
                    protocol: 'cifs',
                    cifs: { share: 'share' }
                }, 'xyz://dksjhf/kslf');
            });

            it('validates S3 target for CIFS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'cifs',
                    cifs: { host: '1.1.1.1', share: '/' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });


            it('validates limited target types for NFS source ', () => {
                validateRelationshipThrows('Target type can only be one of: s3, nfs, local, efs, azure, cifs', {
                    protocol: 'nfs',
                    nfs: { host: '1.1.1.1', export: 'export' }
                }, { protocol: 'xyz', xyz: { host: 'dksjhf', export: 'kslf' } });
            });

            it('validates S3 target for NFS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'nfs',
                    nfs: { host: '1.1.1.1', export: '/' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });


            it('validates limited target types for S3 source', () => {
                validateRelationshipThrows('Target type can only be one of: nfs, local, efs, cifs', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'xyz', xyz: { host: 'dksjhf', export: 'kslf' } });
            });

            it('validates NFS target for S3 source', () => {
                validateRelationshipDoesNotThrow({ protocol: 's3', s3: { bucket: 'some-bucket' } }, {
                    protocol: 'nfs',
                    nfs: { host: '1.1.1.1', export: '/' }
                });
            });

            it('validates CIFS target for S3 source', () => {
                validateRelationshipDoesNotThrow({ protocol: 's3', s3: { bucket: 'some-bucket' } }, {
                    protocol: 'cifs',
                    cifs: { host: '1.1.1.1', share: '/' }
                });
            });
        });

        describe('validate URLs IP/Hostname', () => {
            it('validates invalid source ip for CIFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 'cifs',
                    cifs: { host: '1.', share: '/' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates valid source ip for CIFS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'cifs',
                    cifs: { host: '1.1.1.1', share: '/' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid source hostname for CIFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 'cifs',
                    cifs: { host: 'some-host.', share: '/' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates valid source hostname for CIFS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'cifs',
                    cifs: { host: 'host-1.local.example.com', share: '/' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid source ip for NFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 'nfs',
                    nfs: { host: '1.1.', export: '/' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates valid source ip for NFS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'nfs',
                    nfs: { host: '231.233.234.235', export: '/' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid source hostname for NFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 'nfs',
                    nfs: { host: 'host,', export: '/' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates valid source hostname for NFS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'nfs',
                    nfs: { host: 'host-name.local', export: '/' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid target ip for CIFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'cifs', cifs: { host: '1.', share: '/' } });
            });

            it('validates valid target ip for CIFS source', () => {
                validateRelationshipDoesNotThrow({ protocol: 's3', s3: { bucket: 'some-bucket' } }, {
                    protocol: 'cifs',
                    cifs: { host: '1.1.1.1', share: '/' }
                });
            });

            it('validates invalid target hostname for CIFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'cifs', cifs: { host: 'some-host.', share: '/' } });
            });

            it('validates valid target hostname for CIFS source', () => {
                validateRelationshipDoesNotThrow({ protocol: 's3', s3: { bucket: 'some-bucket' } }, {
                    protocol: 'cifs',
                    cifs: { host: 'host-1.local.example.com', share: '/' }
                });
            });

            it('validates invalid target ip for NFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'nfs', nfs: { host: '1.1.', export: '/' } });
            });

            it('validates valid target ip for NFS source', () => {
                validateRelationshipDoesNotThrow({ protocol: 's3', s3: { bucket: 'some-bucket' } }, {
                    protocol: 'nfs',
                    nfs: { host: '231.233.234.235', export: '/' }
                });
            });

            it('validates invalid target hostname for NFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'nfs', nfs: { host: 'host,', export: '/' } });
            });

            it('validates valid target hostname for NFS source', () => {
                validateRelationshipDoesNotThrow({ protocol: 's3', s3: { bucket: 'some-bucket' } }, {
                    protocol: 'nfs',
                    nfs: { host: 'host-name.local', export: '/' }
                });
            });

            it('validates invalid target ip for SGWS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 'cifs',
                    cifs: { host: '1.1.1.1', share: '/' }
                }, { protocol: 'sgws', sgws: { host: '1.1.', port: '8080', bucket: 'bucket' } }, true);
            });

            it('validates valid target ip for SGWS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'cifs',
                    cifs: { host: '1.1.1.1', share: '/' }
                }, { protocol: 'sgws', sgws: { host: '231.233.234.235', port: '8910', bucket: 'bucket' } }, true);
            });

            it('validates invalid target hostname for SGWS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 'nfs',
                    nfs: { host: 'host', export: '/' }
                }, { protocol: 'sgws', sgws: { host: 'host,', port: '8080', bucket: 'bucket' } }, true);
            });

            it('validates valid target hostname for SGWS source', () => {
                validateRelationshipDoesNotThrow({ protocol: 'nfs', nfs: { host: 'host', export: '/' } }, {
                    protocol: 'sgws',
                    sgws: { host: 'host-name.example.com', port: '8080', bucket: 'bucket' }
                }, true);
            });
        });

        describe('validate path components', () => {
            it('validates invalid source CIFS path is \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 'cifs',
                    cifs: { host: 'host', share: 'share', path: '/.' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid target CIFS path is \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'cifs', cifs: { host: 'host', share: 'share', path: '/.' } });
            });

            it('validates invalid source CIFS path ends with \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 'cifs',
                    cifs: { host: 'host', share: 'share', path: 'folder1/folder2/.' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid target CIFS path ends with \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'cifs', cifs: { host: 'host', share: 'share', path: 'folder1/folder2/.' } });
            });

            it('validates invalid source CIFS path contains \'/./\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 'cifs',
                    cifs: { host: 'host', share: 'share', path: 'folder1/./folder2' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid target CIFS path contains \'/./\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'cifs', cifs: { host: 'host', share: 'share', path: 'folder1/./folder2' } });
            });

            it('validates invalid source NFS path is \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 'nfs',
                    nfs: { host: 'host', export: 'share', path: '/.' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid target NFS path is \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'nfs', nfs: { host: 'host', export: 'export', path: '/.' } });
            });

            it('validates invalid source NFS path ends with \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 'nfs',
                    nfs: { host: 'host', export: 'export', path: 'folder1/folder2/.' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid target NFS path ends with \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'nfs', nfs: { host: 'host', export: 'export', path: 'folder1/folder2/.' } });
            });

            it('validates invalid source NFS path contains \'/./\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 'nfs',
                    nfs: { host: 'host', export: 'export', path: 'folder1/./folder2' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid target NFS path contains \'/./\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'nfs', nfs: { host: 'host', export: 'export', path: 'folder1/./folder2' } });
            });

            it('validates invalid source CIFS path is \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 'cifs',
                    cifs: { host: 'host', share: 'share', path: '/..' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid target CIFS path is \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'cifs', cifs: { host: 'host', share: 'share', path: '/..' } });
            });

            it('validates invalid source CIFS path ends with \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 'cifs',
                    cifs: { host: 'host', share: 'share', path: 'folder1/folder2/..' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid target CIFS path ends with \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'cifs', cifs: { host: 'host', share: 'share', path: 'folder1/folder2/..' } });
            });

            it('validates invalid source CIFS path contains \'/../\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 'cifs',
                    cifs: { host: 'host', share: 'share', path: 'folder1/../folder2' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid target CIFS path contains \'/../\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'cifs', cifs: { host: 'host', share: 'share', path: 'folder1/../folder2' } });
            });

            it('validates invalid source NFS path is \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 'nfs',
                    nfs: { host: 'host', export: 'export', path: '/..' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid target NFS path is \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'nfs', nfs: { host: 'host', export: 'export', path: '/..' } });
            });

            it('validates invalid source NFS path ends with \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 'nfs',
                    nfs: { host: 'host', export: 'export', path: 'folder1/folder2/..' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid target NFS path ends with \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'nfs', nfs: { host: 'host', export: 'export', path: 'folder1/folder2/..' } });
            });

            it('validates invalid source NFS path contains \'/../\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 'nfs',
                    nfs: { host: 'host', export: 'export', path: 'folder1/../folder2' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket' } });
            });

            it('validates invalid target NFS path contains \'/../\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket' }
                }, { protocol: 'nfs', nfs: { host: 'host', export: 'export', path: 'folder1/../folder2' } });
            });

            it('validates invalid CIFS source path with folder longer than 255 characters', () => {
                const veryLongFolder = repeat('f', 256);
                validateRelationshipThrows(
                    'contains folders longer than 255 characters',
                    {
                        protocol: 'cifs',
                        cifs: { host: 'host', share: 'share', path: `folder1/${veryLongFolder}/folder3` }
                    }, { protocol: 's3', s3: { bucket: 'some-bucket', prefix: 'path' } }
                );
            });

            it('validates invalid CIFS target path with folder longer than 255 characters', () => {
                const veryLongFolder = repeat('f', 256);
                validateRelationshipThrows(
                    'contains folders longer than 255 characters',
                    { protocol: 's3', s3: { bucket: 'some-bucket', prefix: 'path' } }, {
                        protocol: 'cifs',
                        cifs: { host: 'host', share: 'share', path: `folder1/${veryLongFolder}/folder3` }
                    }
                );
            });

            it('validates invalid NFS source path with folder longer than 255 characters', () => {
                const veryLongFolder = repeat('f', 256);
                validateRelationshipThrows(
                    'contains folders longer than 255 characters',
                    {
                        protocol: 'nfs',
                        nfs: { host: 'host', export: 'export', path: `folder1/${veryLongFolder}/folder3` }
                    }, { protocol: 's3', s3: { bucket: 'some-bucket', prefix: 'path' } }
                );
            });

            it('validates invalid NFS target path with folder longer than 255 characters', () => {
                const veryLongFolder = repeat('f', 256);
                validateRelationshipThrows(
                    'contains folders longer than 255 characters',
                    { protocol: 's3', s3: { bucket: 'some-bucket', prefix: 'path' } }, {
                        protocol: 'nfs',
                        nfs: { host: 'host', export: 'export', path: `folder1/${veryLongFolder}/folder3` }
                    }
                );
            });

            it('validates invalid S3 source path length', () => {
                const folder = '/folderX';
                const path = repeat(folder, (1024 / folder.length) + 2);
                validateRelationshipThrows('S3 prefix maximum characters length can be 1024', {
                    protocol: 's3',
                    s3: { bucket: 'some-bucket', prefix: `${path}` }
                }, { protocol: 'nfs', nfs: { host: 'host', export: '/' } });
            });

            it('validates invalid S3 target path length', () => {
                const folder = '/folderX';
                const path = repeat(folder, (1024 / folder.length) + 2);
                validateRelationshipThrows('S3 prefix maximum characters length can be 1024', {
                    protocol: 'nfs',
                    nfs: { host: 'host', export: '/' }
                }, { protocol: 's3', s3: { bucket: 'some-bucket', prefix: `${path}` } });
            });
        });

        describe('validate S3 tags', () => {
            function validateS3TagsDoesNotThrow(s3Tags) {
                expect(validateS3Tags.bind(this, s3Tags)).not.to.throw();
            }

            function validateS3TagsThrows(expectedError, s3Tags) {
                expect(validateS3Tags.bind(this, s3Tags)).to.throw(BadRequest, expectedError);
            }

            it('should not fail if no tags provided', () => {
                validateS3TagsDoesNotThrow(undefined);
            });

            it('should not fail if empty tags list provided', () => {
                validateS3TagsDoesNotThrow([]);
            });


            it('should fail if provided more than 5 tags', () => {
                const s3Tags = [];
                for (let i = 1; i <= 5 + 1; i++) {
                    s3Tags.push({ key: `key${i}`, value: `value${i}` });
                }
                validateS3TagsThrows('Maximum of 5 S3 tags may be provided', s3Tags);
            });

            it('should fail if one of the tags missing key', () => {
                const key = 'key';
                const value = 'value';
                const s3Tags = [{ key, value }, { key }, { key, value }];
                validateS3TagsThrows('Each Amazon tag should contain both key and value', s3Tags);
            });

            it('should fail if one of the tag\'s key contains not allowed characters', () => {
                const value = 'value';
                const s3Tags = [{ key: 'key', value }, { key: 'key$', value }];
                validateS3TagsThrows('Amazon tag keys may not start with \'aws:\' and may contain letters, numbers, spaces, or the special characters + - . _ : / @', s3Tags);
            });

            it('should fail if one of the tag\'s value contains not allowed characters', () => {
                const s3Tags = [{ key: 'key1', value: 'value1' }, { key: 'key2', value: 'va^lue' }];
                validateS3TagsThrows('Amazon tag values may not start with \'aws:\' and may contain letters, numbers, spaces, or the special characters + - . _ : / @', s3Tags);
            });

            it('should fail if has duplicate key', () => {
                const value = 'value';
                const s3Tags = [{ key: 'key1', value }, { key: 'key2', value }, { key: 'key2', value }, { key: 'key4', value }];
                validateS3TagsThrows('S3 tags keys should be distinct (no duplicated allowed)', s3Tags);
            });
        });


        describe('validate sync schedule', () => {
            function validateSyncIntervalsDoesNotThrow(inDays, inHours, inMinutes, autoSync = 'ON') {
                const sync = {
                    inDays, inHours, inMinutes, autoSync
                };

                expect(validateSyncSchedule.bind(this, sync)).not.to.throw();
            }

            function validateSyncIntervalsThrows(expectedError, inDays, inHours, inMinutes, autoSync = 'ON') {
                const sync = {
                    inDays, inHours, inMinutes, autoSync
                };
                expect(validateSyncSchedule.bind(this, sync)).to.throw(BadRequest, expectedError);
            }

            it('should fail if both intervals ( inHours and inDays) provided', () => {
                validateSyncIntervalsThrows('Only one of \'syncInDays\' or \'syncInHours\' can be provided for sync interval', 10, 10, 0);
            });

            it('should fail if inHours = 0  and inDays = 0 and inMinutes = 0 and autoSync = ON provided', () => {
                validateSyncIntervalsThrows('Please provide one of \\\'syncInHours\\\' or \\\'syncInDays\\\' or \\\'syncInMinutes \\\' for sync interval\' when pause is false', 0, 0, 0);
            });

            it('should fail if inHours = 0  and inDays = 0 and inMinutes = 0 and autoSync = OFF provided', () => {
                validateSyncIntervalsDoesNotThrow(0, 0, 0, 'OFF');
            });

            it('should fail if both intervals ( inMinutes and inDays) provided', () => {
                validateSyncIntervalsThrows('Only one of  \'syncInDays\' or \'syncInMinutes\' can be provided for sync interval', 10, 0, 10);
            });

            it('should fail if both intervals ( inMinutes and inHours) provided', () => {
                validateSyncIntervalsThrows('Only one of \'syncInHours\' or \'syncInMinutes\' can be provided for sync interval', 0, 10, 10);
            });

            it('should fail if all intervals provided', () => {
                validateSyncIntervalsThrows('Only one of \'syncInDays\' or \'syncInHours\' or \'syncInMinutes\' can be provided for sync interval', 10, 10, 10);
            });

            it('should fail on negative inDays', () => {
                validateSyncIntervalsThrows('\'syncInDays\' interval must be between 1 and 365 or 0', -1, 0, 0);
            });

            it('should fail on too high inDays', () => {
                validateSyncIntervalsThrows('\'syncInDays\' interval must be between 1 and 365 or 0', 366, 0, 0);
            });

            it('should fail on negative inHours', () => {
                validateSyncIntervalsThrows('\'syncInHours\' interval must be between 1 and 24 or 0', 0, -1, 0);
            });

            it('should fail on too high inHours', () => {
                validateSyncIntervalsThrows('\'syncInHours\' interval must be between 1 and 24 or 0', 0, 25, 0);
            });

            it('should fail on negative inHours', () => {
                validateSyncIntervalsThrows('\'syncInMinutes\' interval must be between 1 and 60 or 0', 0, 0, -1);
            });

            it('should fail on too high inHours', () => {
                validateSyncIntervalsThrows('\'syncInMinutes\' interval must be between 1 and 60 or 0', 0, 0, 61);
            });

            it('should succeed on lower end valid inDays', () => {
                validateSyncIntervalsDoesNotThrow(1, 0, 0);
            });

            it('should succeed on higher end valid inDays', () => {
                validateSyncIntervalsDoesNotThrow(365, 0, 0);
            });

            it('should succeed on lower end valid inHours', () => {
                validateSyncIntervalsDoesNotThrow(0, 1, 0);
            });

            it('should succeed on higher end valid inHours', () => {
                validateSyncIntervalsDoesNotThrow(0, 24, 0);
            });

            it('should succeed on higher end valid inMinutes', () => {
                validateSyncIntervalsDoesNotThrow(0, 0, 1);
            });

            it('should succeed on higher end valid inMinutes', () => {
                validateSyncIntervalsDoesNotThrow(0, 0, 60);
            });
        });
    });
});
