import fs from 'fs';
import _ from 'lodash';
import Promise from 'bluebird';
import {
    COPY_RELATIONSHIP_MESSAGE,
    SYNC_RELATIONSHIP_MESSAGE,
    ABORT_RELATIONSHIP,
    STARTUP_RELATIONSHIP,
    RUNNING_STATUS,
    MESSAGE_STATUS_DONE,
    MESSAGE_STATUS_FAILED,
    MESSAGE_STATUS_INTERRUPTED,
    MESSAGE_STATUS_ABORTED,
    MESSAGE_STATUS_ABORTING,
    ABORTED_MESSAGE
} from '../../src/utils/consts';
import * as common from '../../src/utils/common';
import User from '../../src/models/user';
import Relationship from '../../src/models/relationship';
import DataBroker from '../../src/models/data-broker';
import { expect } from '../globals';


describe('common', () => {
    let user;
    let dataBroker;
    const host = '1.1.1.1';
    const nfsExport = 'export';
    const path = 'folder1/folder2';
    const bucket = 'bucket';
    const prefix = 'folder';
    const port = '54';
    const accessKey = 'abc';
    const secretKey = 'def';
    const domain = 'a';
    const username = 'admin';
    const password = 'abc';
    const source = { protocol: 'nfs', nfs: { host, export: nfsExport } };
    const target = { protocol: 's3', s3: { bucket, prefix } };
    const autoSync = 'OFF';
    const dataBrokerName = 'dataBrokerName';
    const sync = {
        inDays: 0, inHours: 1, inMinutes: 0, nextTime: new Date(), autoSync
    };
    const settings = { gracePeriod: 30 };

    function createActiveMessage(step, status, timestamp = Date.now()) {
        const activeMesssage = {
            scan: {
                step: 'scan',
                status: 'done',
                timestamp: 1497698161812,
                stats: {
                    itemsScanned: 19,
                    itemsScannedFailed: 0,
                    itemsScannedRetried: 0,
                    itemsMarkedForTransfer: 696,
                    bytesMarkedForTransfer: 55824023,
                    itemsMarkedForDelete: 0,
                    itemsWithinGracePeriod: 0,
                    itemsTransferIgnored: 0
                }
            },
            copy: {
                step: 'copy',
                status: 'done',
                timestamp: 1497698167812,
                stats: {
                    itemsTransferPending: 0,
                    itemsTransferPendingBytes: 0,
                    itemsTransferring: 0,
                    partsTransferringBytes: 0,
                    itemsTransferred: 696,
                    totalTransferredBytes: 55824023,
                    itemsDeleted: 0,
                    itemsTransferFailed: 0,
                    'itemsTransferFailedBytes ': 0,
                    itemsDeleteFailed: 0,
                    readsInTransit: 0,
                    uploadsInTransit: 0,
                    opensInTransit: 0,
                    closesInTransit: 0,
                    transferBps: 2989000
                }
            },
            index: {
                step: 'index',
                status: 'done',
                timestamp: 1497698152812,
                stats: {
                    itemsIndexed: 0,
                    itemsIgnored: 3
                }
            },
            setup: {
                step: 'setup',
                status: 'done',
                timestamp: 1497698146812
            }
        };
        activeMesssage[step].status = status;
        activeMesssage[step].timestamp = timestamp;
        return activeMesssage;
    }
    function createMessagesObjcet(createRelationship, syncRelationship, abortRelationship, startup) {
        const obj = {
            [COPY_RELATIONSHIP_MESSAGE]: createRelationship,
            [SYNC_RELATIONSHIP_MESSAGE]: syncRelationship,
            [ABORT_RELATIONSHIP]: abortRelationship,
            [STARTUP_RELATIONSHIP]: startup
        };
        return _.pickBy(obj, _.identity);
    }

    function createStartupMessage(status, timestamp = Date.now()) {
        return {
            setup: {
                step: 'setup',
                status,
                timestamp
            }
        };
    }

    function createAbortRelationshipMessage(status, timestamp = Date.now()) {
        return {
            abort: {
                step: 'abort',
                status,
                timestamp
            }
        };
    }

    function createOldMessage() {
        return {
            status: {
                lastBucketEventReceivedTime: 1479285431116,
                nfs: {
                    lastUpdateTime: 1479319514136,
                    online: true
                },
                progress: {
                    done: {
                        ignoredDeletes: 0,
                        ignoredObjects: 0,
                        failedDeletes: 0,
                        failedObjects: 0,
                        deletes: 3,
                        bytes: 265165,
                        objects: 7
                    },
                    syncing: {
                        deletes: 0,
                        bytes: 0,
                        objects: 0
                    },
                    pending: {
                        deletes: 0,
                        bytes: 0,
                        objects: 0
                    }
                },
                lastUpdateTime: 1479319514136,
                status: 'ok',
                phase: 'ready'
            }
        };
    }

    beforeEach(async() => {
        await Promise.all([User.deleteMany(), Relationship.deleteMany(), DataBroker.deleteMany()]);
        user = await User.create({
            name: 'User Name',
            email: 'user@email.com',
            password: 'Netapp234',
            company: 'company',
            trialExpiry: new Date(),
            awsAccountId: '282316784512',
            portalId: '1'
        });
        dataBroker =  await DataBroker.create({ user, name: dataBrokerName });
    });

    describe('isRelationshipRunning', () => {
        function validateIsRunningRelationship(status, result) {
            common.isRelationshipRunning(status).should.equal(result);
        }

        it('relationship with status FAILED should return false', () => {
            validateIsRunningRelationship({ status: MESSAGE_STATUS_FAILED }, false);
        });

        it('relationship with status done should return false', () => {
            validateIsRunningRelationship({ status: MESSAGE_STATUS_DONE }, false);
        });

        it('relationship with status running should return true', () => {
            validateIsRunningRelationship({ status: RUNNING_STATUS }, true);
        });

        it('relationship with status aborted should return false', () => {
            validateIsRunningRelationship({ status: ABORTED_MESSAGE }, false);
        });

        it('relationship with status aborting should return false', () => {
            validateIsRunningRelationship({ status: MESSAGE_STATUS_ABORTING }, false);
        });
    });

    describe('isStartupRunning', () => {
        function validateIsStartupRunning(messages, result) {
            const relationship = new Relationship({source, target, sync, settings, user, dataBroker, messages});
            common.isStartupRunning(relationship).should.equal(result);
        }

        it('relationship with old messages should return false', () => {
            validateIsStartupRunning(createOldMessage(), false);
        });

        it('relationship with no startup message should return false', () => {
            validateIsStartupRunning(createMessagesObjcet(createActiveMessage('copy', 'done')), false);
        });

        it('relationship with last startup progress message should return true', () => {
            validateIsStartupRunning(createMessagesObjcet(createActiveMessage('copy', 'done', Date.now() - (24 * 60)), undefined, undefined, createStartupMessage('started')), true);
        });

        it('relationship with last startup message done should return false', () => {
            validateIsStartupRunning(createMessagesObjcet(createActiveMessage('copy', 'done', Date.now() - (24 * 60)), undefined, undefined, createStartupMessage('done')), false);
        });

        it('relationship with old startup message should return false', () => {
            validateIsStartupRunning(createMessagesObjcet(createActiveMessage('copy', 'done', Date.now() - (24 * 60)), undefined, undefined, createStartupMessage('done', Date.now() - (20 * 60))), false);
        });
    });

    describe('findMaxTimestampForAction', () => {
        const activeMessage = {
            scan: {
                timestamp: 4
            },
            copy: {
                timestamp: 1
            },
            index: {
                timestamp: 3
            },
            setup: {
                timestamp: 1
            }
        };

        it('Should return exist and return 4', () => {
            const maxTimestamp = common.findMaxTimestampForAction(activeMessage, 'timestamp');
            expect(maxTimestamp).to.exist;
            maxTimestamp.should.equal(4);
        });
    });

    describe('findRelationshipsForStartup', () => {
        function createNewMessages(status, phase) {
            return {
                [phase]: {
                    scan: {
                        step: 'scan',
                        status: 'done',
                        timestamp: 4,
                        stats: {
                            itemsScanned: 19,
                            itemsScannedFailed: 0,
                            itemsScannedRetried: 0,
                            itemsMarkedForTransfer: 696,
                            bytesMarkedForTransfer: 55824023,
                            itemsMarkedForDelete: 0,
                            itemsWithinGracePeriod: 0,
                            itemsTransferIgnored: 0
                        }
                    },
                    copy: {
                        step: 'copy',
                        status,
                        timestamp: 3,
                        stats: {
                            itemsTransferPending: 0,
                            itemsTransferPendingBytes: 0,
                            itemsTransferring: 0,
                            partsTransferringBytes: 0,
                            itemsTransferred: 696,
                            totalTransferredBytes: 55824023,
                            itemsDeleted: 0,
                            itemsTransferFailed: 0,
                            'itemsTransferFailedBytes ': 0,
                            itemsDeleteFailed: 0,
                            readsInTransit: 0,
                            uploadsInTransit: 0,
                            opensInTransit: 0,
                            closesInTransit: 0,
                            transferBps: 2989000
                        }
                    },
                    index: {
                        step: 'index',
                        status: 'done',
                        timestamp: 2,
                        stats: {
                            itemsIndexed: 0,
                            itemsIgnored: 3
                        }
                    },
                    setup: {
                        step: 'setup',
                        status: 'done',
                        timestamp: 1
                    }
                }
            };
        }

        async function validateFindRelationshipForStartup(source, target, messages) {
            try {
                const relationship = await Relationship.create({
                    source, target, sync, settings, user, dataBroker: dataBroker, messages
                });
                const relationshipId = relationship.id;
                const relationshipsForStartup = await common.findRelationshipsForStartup(dataBroker);
                relationshipsForStartup.should.deep.equal([{relationshipId}]);
            } catch (e) {
                throw new Error(e);
            }
        }

        async function validateFindRelationshipForStartupNoReturnIds(source, target, messages, isNew) {
            try {
                const relationship = await Relationship.create({
                    source, target, sync, user, settings, dataBroker, messages
                });
                const relationshipForStartup = await common.findRelationshipsForStartup(dataBroker);
                expect(relationshipForStartup).to.be.empty;
                const foundRelationship = await Relationship.findById(relationship.id).exec();
                if (isNew) {
                    foundRelationship.messages['create-relationship-v2'].copy.status.should.equal('failed');
                }
            } catch (e) {
                throw new Error(e);
            }
        }

        it('Upload Relationship  (copy) with status DONE should return its id', () => validateFindRelationshipForStartup(source, target, createNewMessages('done', 'create-relationship-v2')));

        it('Upload Relationship (copy) with status FAILED should return its id', () => validateFindRelationshipForStartup(source, target, createNewMessages('failed', 'create-relationship-v2')));

        it('Upload Relationship (copy) with status RUNNING should return its id', () => validateFindRelationshipForStartup(source, target, createNewMessages('progress', 'create-relationship-v2')));

        it('Upload Relationship (sync) with status DONE should return its id', () => validateFindRelationshipForStartup(source, target, createNewMessages('done', 'sync-relationship')));

        it('Upload Relationship (sync) with status FAILED should return its id', () => validateFindRelationshipForStartup(source, target, createNewMessages('failed', 'sync-relationship')));

        it('Upload Relationship (sync) with status RUNNING should return its id', () => validateFindRelationshipForStartup(source, target, createNewMessages('progress', 'sync-relationship')));

        it('DOWNLOAD Relationship (copy) with status RUNNING should not return its id', () => validateFindRelationshipForStartupNoReturnIds(target, source, createNewMessages('progress', 'create-relationship-v2'), true));

        it('DOWNLOAD Relationship (copy) with status DONE should return its id', () => validateFindRelationshipForStartup(target, source, createNewMessages('done', 'create-relationship-v2'), true));

        it('DOWNLOAD Relationship (copy) with status FAILED should return its id', () => validateFindRelationshipForStartup(target, source, createNewMessages('failed', 'create-relationship-v2'), true));

        it('DOWNLOAD Relationship (sync) with status PROGRESS should return its id', () => validateFindRelationshipForStartup(target, source, createNewMessages('progress', 'sync-relationship'), true));

        it('DOWNLOAD Relationship (sync) with status FAILED should return its id', () => validateFindRelationshipForStartup(target, source, createNewMessages('failed', 'sync-relationship'), true));

        it('DOWNLOAD Relationship (sync) with status DONE should return its id', () => validateFindRelationshipForStartup(target, source, createNewMessages('done', 'sync-relationship'), true));

        it('No relationship in the db should return undefined', () => common.findRelationshipsForStartup(dataBroker)
            .then(relationship => relationship.should.deep.equal([])));
    });

    describe('calculateTimelineStatusForRelationship', () => {
        function createStatusObj(status, failureMessage) {
            return {
                status,
                failureMessage
            };
        }

        function createCurrMessage(status, step, timestamp = Date.now()) {
            return {
                step,
                status,
                timestamp
            };
        }


        function validateTimeLineStatus(currMessage, expectedStatusObj) {
            common.calculateTimelineStatusForRelationship(currMessage).should.deep.equal(expectedStatusObj);
        }

        it('relationship with currMessage with status failed return status FAILED and a message', () => {
            validateTimeLineStatus(createCurrMessage('failed', 'index'), createStatusObj(MESSAGE_STATUS_FAILED));
        });

        it('relationship with currMessage with step copy and status done should return status DONE', () => {
            validateTimeLineStatus(createCurrMessage('done', 'copy'), createStatusObj(MESSAGE_STATUS_DONE));
        });

        it('relationship with currMessage with step abort and status done should return status ABORTED', () => {
            validateTimeLineStatus(createCurrMessage('done', 'abort'), createStatusObj(MESSAGE_STATUS_ABORTED, ABORTED_MESSAGE));
        });

        it('relationship with currMessage with step abort and status started should return status ABORTING', () => {
            validateTimeLineStatus(createCurrMessage('started', 'abort'), { status: MESSAGE_STATUS_ABORTING });
        });
    });

    describe('createStringFromEndPointObject', () => {
        const host = '1.1.1.1';
        const nfsExport = 'export';
        const path = 'folder1/folder2';
        const bucket = 'bucket';
        const port = '80';

        function validateStringCreatedFromEndPoint(endPoint, expectedString) {
            common.createStringFromEndPointObject(endPoint).should.equal(expectedString);
        }
        function buildEndPoint(protocol, details) {
            return { protocol, [protocol]: details };
        }
        function buildFileSystemDetailObject(protocol, host, share, path) {
            return (protocol === 'nfs') ? { host, export: share, path } : { host, share, path };
        }

        function buildStorageDetailObject(protocol, bucket, prefix, host, port) {
            return (protocol === 's3') ? { bucket, prefix } : {
                bucket, prefix, host, port
            };
        }
        it('nfs endPoint object with no path', () => {
            validateStringCreatedFromEndPoint(buildEndPoint('nfs', buildFileSystemDetailObject('nfs', host, nfsExport)), `nfs://${host}/${nfsExport}`);
        });

        it('nfs endPoint object with path', () => {
            validateStringCreatedFromEndPoint(buildEndPoint('nfs', buildFileSystemDetailObject('nfs', host, nfsExport, path)), `nfs://${host}/${nfsExport}/${path}`);
        });

        it('cifs endPoint object with no path', () => {
            validateStringCreatedFromEndPoint(buildEndPoint('cifs', buildFileSystemDetailObject('cifs', host, nfsExport)), `cifs://${host}/${nfsExport}`);
        });

        it('cifs endPoint object with path', () => {
            validateStringCreatedFromEndPoint(buildEndPoint('cifs', buildFileSystemDetailObject('cifs', host, nfsExport, path)), `cifs://${host}/${nfsExport}/${path}`);
        });

        it('s3 endPoint object with no prefix', () => {
            validateStringCreatedFromEndPoint(buildEndPoint('s3', buildStorageDetailObject('s3', bucket)), `s3://${bucket}`);
        });

        it('s3 endPoint object with prefix', () => {
            validateStringCreatedFromEndPoint(buildEndPoint('s3', buildStorageDetailObject('s3', bucket, path)), `s3://${bucket}/${path}`);
        });

        it('sgws endPoint object with no prefix and no port', () => {
            validateStringCreatedFromEndPoint(buildEndPoint('sgws', buildStorageDetailObject('sgws', bucket, undefined, host)), `sgws://${host}/${bucket}`);
        });

        it('sgws endPoint object with no prefix and port', () => {
            validateStringCreatedFromEndPoint(buildEndPoint('sgws', buildStorageDetailObject('sgws', bucket, undefined, host, port)), `sgws://${host}:${port}/${bucket}`);
        });
        it('sgws endPoint object with  prefix and port', () => {
            validateStringCreatedFromEndPoint(buildEndPoint('sgws', buildStorageDetailObject('sgws', bucket, path, host, port)), `sgws://${host}:${port}/${bucket}/${path}`);
        });
    });

    describe('createObjectForAsterix', () => {
        const credentials = {
            cifs: {
                domain,
                username,
                password
            },
            sgws: {
                accessKey,
                secretKey
            }
        };

        const credentialsWithNoDomain = {
            cifs: {
                domain,
                username,
                password
            },
            sgws: {
                accessKey,
                secretKey
            }
        };


        function createEndPointObject(protocol, details) {
            return { protocol, [protocol]: details };
        }

        function validateObjectForAsterix(mongoObject, credentials, asterixObject) {
            common.createObjectForAsterix(mongoObject, credentials).should.deep.equal(asterixObject);
        }

        it('nfs object', () => {
            validateObjectForAsterix(createEndPointObject('nfs', { host, export: nfsExport, path }), credentials, { nfs: { host, export: nfsExport, path } });
        });

        it('s3 object', () => {
            validateObjectForAsterix(createEndPointObject('s3', { bucket, prefix }), credentials, { s3: { bucket, prefix } });
        });

        it('sgws object with port', () => {
            validateObjectForAsterix(
                createEndPointObject('sgws', {
                    bucket, prefix, host, port
                }), credentials,
                {
                    sgws: {
                        bucket, host, prefix, port: Number.parseInt(port), accessKey, secretKey
                    }
                }
            );
        });

        it('sgws object with no port', () => {
            validateObjectForAsterix(
                createEndPointObject('sgws', { bucket, prefix, host }), credentials,
                {
                    sgws: {
                        bucket, host, prefix, accessKey, secretKey
                    }
                }
            );
        });

        it('cifs object with domain', () => {
            validateObjectForAsterix(
                createEndPointObject('cifs', { path, share: nfsExport, host }), credentials,
                {
                    cifs: {
                        host, share: nfsExport, path, username, password, domain
                    }
                }
            );
        });

        it('cifs object with  no domain', () => {
            validateObjectForAsterix(
                createEndPointObject('cifs', { path, share: nfsExport, host }), credentialsWithNoDomain,
                {
                    cifs: {
                        host, share: nfsExport, path, username, password, domain
                    }
                }
            );
        });
    });


    describe('updateSubscription', () => {
        const path = require('path');

        beforeEach(async() => {
            await Promise.all([User.deleteMany(), Relationship.deleteMany(), DataBroker.deleteMany()]);
            const users = JSON.parse(fs.readFileSync(path.resolve('./test/resources', 'users.json'), 'utf8'));
            return User.insertMany(users);
        });

        after(() => User.deleteMany());

        async function validateUsersUpdated(subscription, users, expectedSubscription) {
            try {
                await common.updateSubscription(subscription, users);
                const usersIds = users.map(user => ({_id: user._id}));
                const usersFromFind = await User.find({$or: usersIds}).exec();
                usersFromFind.forEach(user => {
                    if (!(user.subscription.type === expectedSubscription.type && user.subscription.status === expectedSubscription.status)) {
                        throw new Error(`subscription ${user.subscription} is not as expected ${JSON.stringify(expectedSubscription)}`);
                    }
                });
            } catch (e) {
                throw new Error(e);
            }
        }

        it('update subscription with empty subscription- everything should be the same', async() => {
            const users = await User.find({ 'subscription.status': 'unsubscribed-init' }).exec();
            validateUsersUpdated(undefined, users, {
                updatedAt: '1487146566938',
                status: 'unsubscribed-init',
                type: 'PAYGO'
            });
        });

        it('update subscription with PAYGO subscription', async() => {
            const users = await User.find({ 'subscription.status': 'unsubscribed-init' }).exec();
            validateUsersUpdated({
                updatedAt: Date.now(),
                status: 'subscribe-success',
                type: 'PAYGO',
                customerIdentifier: 'AD'
            }, users, {
                updatedAt: Date.now(),
                status: 'subscribe-success',
                type: 'PAYGO',
                customerIdentifier: 'AD'
            });
        });

        it('update subscription with CONTRACT subscription', async() => {
            const users = await User.find({ 'subscription.type': 'PAYGO' }).exec();
            validateUsersUpdated({
                updatedAt: Date.now(),
                status: 'subscribe-success',
                type: 'CONTRACT',
                customerIdentifier: 'AD'
            }, users, {
                updatedAt: Date.now(),
                status: 'subscribe-success',
                type: 'CONTRACT',
                customerIdentifier: 'AD'
            });
        });

        it('update CONTRACT subscription with PAYGO subscription should not be updated', async() => {
            const users = await User.find({ 'subscription.type': 'CONTRACT' }).exec();
            validateUsersUpdated({
                updatedAt: Date.now(),
                status: 'subscribe-success',
                type: 'PAYGO',
                customerIdentifier: 'AD'
            }, users, {
                updatedAt: Date.now(),
                status: 'subscribe-success',
                type: 'CONTRACT',
                customerIdentifier: 'AD'
            });
        });
    });

    describe('calculateProgress', () => {
        function createCopyMessage(status, totalTransferredBytes, itemsMarkedForTransferBytes, itemsMarkedForTransfer) {
            return {
                step: 'copy',
                status,
                timestamp: 3,
                stats: {
                    itemsTransferPending: 0,
                    itemsTransferPendingBytes: 0,
                    itemsTransferring: 0,
                    partsTransferringBytes: 0,
                    itemsTransferred: 696,
                    totalTransferredBytes,
                    itemsMarkedForTransferBytes,
                    itemsMarkedForTransfer,
                    itemsDeleted: 0,
                    itemsTransferFailed: 0,
                    itemsTransferFailedBytes: 0,
                    itemsDeleteFailed: 0,
                    readsInTransit: 0,
                    uploadsInTransit: 0,
                    opensInTransit: 0,
                    closesInTransit: 0,
                    transferBps: 2989000
                }
            };
        }

        function createNewMessages(copyObject) {
            return {
                scan: {
                    step: 'scan',
                    status: 'done',
                    timestamp: 4,
                    stats: {
                        itemsScanned: 19,
                        itemsScannedFailed: 0,
                        itemsScannedRetried: 0,
                        itemsMarkedForDelete: 0,
                        itemsWithinGracePeriod: 0,
                        itemsTransferIgnored: 0
                    }
                },
                copy: copyObject,
                index: {
                    step: 'index',
                    status: 'done',
                    timestamp: 2,
                    stats: {
                        itemsIndexed: 0,
                        itemsIgnored: 3
                    }
                },
                setup: {
                    step: 'setup',
                    status: 'done',
                    timestamp: 1
                }
            };
        }

        function validateProgress(messages, status, expectedProgress) {
            const progress = common.calculateProgress(status, messages);
            progress.should.equal(expectedProgress);
        }

        function validateUndefinedProgress(messages, status) {
            const progress = common.calculateProgress(status, messages);
            expect(progress).to.be.undefined;
        }

        it('Calculate progress when there is no itemsMarkedForTransferBytes in copy message and status is Running', () => {
            validateUndefinedProgress(createNewMessages(createCopyMessage('progress', 55824023)), RUNNING_STATUS);
        });

        it('Calculate progress when there is no copy messages and status is Running', () => {
            validateUndefinedProgress(createNewMessages(undefined), RUNNING_STATUS, 0);
        });

        it('Calculate progress when itemsMarkedForTransferBytes is 0 and the status is DONE', () => {
            validateProgress(createNewMessages(createCopyMessage('done', 0, 0, 0)), MESSAGE_STATUS_DONE, 100);
        });

        it(`Calculate progress when itemsMarkedForTransferBytes is 0 and the status is ${MESSAGE_STATUS_FAILED}`, () => {
            validateProgress(createNewMessages(createCopyMessage('failed', 0, 0, 0)), MESSAGE_STATUS_FAILED, 0);
        });

        it(`Calculate progress when itemsMarkedForTransferBytes is 0 and the status is ${RUNNING_STATUS}`, () => {
            validateProgress(createNewMessages(createCopyMessage('progress', 2, 0, 0)), RUNNING_STATUS, 0);
        });

        it(`Calculate progress when itemsMarkedForTransferBytes is 100, totalTransferredBytes is 200 and the status is ${RUNNING_STATUS}`, () => {
            validateProgress(createNewMessages(createCopyMessage('progress', 200, 100, 2)), RUNNING_STATUS, 99);
        });

        it(`Calculate progress when itemsMarkedForTransferBytes is 100, totalTransferredBytes is 100 and the status is ${MESSAGE_STATUS_FAILED}`, () => {
            validateProgress(createNewMessages(createCopyMessage('failed', 100, 100, 2)), MESSAGE_STATUS_FAILED, 0);
        });

        it(`Calculate progress when itemsMarkedForTransferBytes is 100, totalTransferredBytes is 100 and the status is ${MESSAGE_STATUS_DONE}`, () => {
            validateProgress(createNewMessages(createCopyMessage('done', 100, 100, 1000)), MESSAGE_STATUS_DONE, 100);
        });

        it(`Calculate progress when itemsMarkedForTransferBytes is 100, totalTransferredBytes is 200 and the status is ${MESSAGE_STATUS_DONE}`, () => {
            validateProgress(createNewMessages(createCopyMessage('done', 200, 100, 2)), MESSAGE_STATUS_DONE, 100);
        });

        it(`Calculate progress when itemsMarkedForTransferBytes is 100, totalTransferredBytes is 50 and the status is ${RUNNING_STATUS}`, () => {
            validateProgress(createNewMessages(createCopyMessage('progress', 50, 100, 2)), RUNNING_STATUS, 50);
        });

        it(`Calculate progress when itemsMarkedForTransferBytes is 100, totalTransferredBytes is 50 and the status is ${MESSAGE_STATUS_FAILED}`, () => {
            validateProgress(createNewMessages(createCopyMessage('progress', 50, 100, 3)), MESSAGE_STATUS_FAILED, 0);
        });
    });

    describe('calculateStatus', () => {
        function validateStatus(messages, status) {
            const actualStatus = common.calculateStatus(messages);
            actualStatus.status.should.equal(status);
        }

        it('create-relationship-v2 with setup failed should return FAILED', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('setup', 'failed')), MESSAGE_STATUS_FAILED);
        });

        it('create-relationship-v2 with index failed should return FAILED', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('index', 'failed')), MESSAGE_STATUS_FAILED);
        });

        it('create-relationship-v2 with scan failed should return FAILED', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('scan', 'failed')), MESSAGE_STATUS_FAILED);
        });

        it('create-relationship-v2 with copy failed should return FAILED', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'failed')), MESSAGE_STATUS_FAILED);
        });

        it('create-relationship-v2 with copy done should return DONE', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'done')), MESSAGE_STATUS_DONE);
        });

        it('create-relationship-v2 with no failed step or copy done should return RUNNING', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'progress')), RUNNING_STATUS);
        });

        it('old messages should return DONE', () => {
            validateStatus(createOldMessage(), MESSAGE_STATUS_DONE);
        });

        it('create-relationship-v2 done and sync-relationship with no failed step or copy done should return RUNNING', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'done'), createActiveMessage('copy', 'progress')), RUNNING_STATUS);
        });

        it('create-relationship-v2 done and sync-relationship with failed step should return FAILED', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'done'), createActiveMessage('scan', 'failed')), MESSAGE_STATUS_FAILED);
        });

        it('create-relationship-v2 done and sync-relationship with copy done step should return DONE', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'done'), createActiveMessage('copy', 'done')), MESSAGE_STATUS_DONE);
        });

        it('create-relationship-v2 done and sync-relationship with copy progress and abort-relationship with copy done should return ABORTED', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'done'), createActiveMessage('copy', 'progress'), createAbortRelationshipMessage('done')), MESSAGE_STATUS_ABORTED);
        });

        it('create-relationship-v2 done and sync-relationship with copy progress and abort-relationship with status started and abort-relationship with status started should return ABORTING', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'done'), createActiveMessage('copy', 'progress'), createAbortRelationshipMessage('started')), MESSAGE_STATUS_ABORTING);
        });

        it('create-relationship-v2 done and sync-relationship with copy progress and abort-relationship with status started and startup with status done should return ABORTED', () => {
            validateStatus(createMessagesObjcet(
                createActiveMessage('copy', 'done', Date.now() - (2000 * 60 * 60)),
                createActiveMessage('copy', 'progress', Date.now() - (1500 * 60 * 60)), createAbortRelationshipMessage('started', Date.now() - (1000 * 60 * 60)),
                createStartupMessage('started', Date.now())
            ), MESSAGE_STATUS_ABORTED);
        });

        it('create-relationship-v2 done and sync-relationship with copy done and startup with status done should return DONE', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'done'), createActiveMessage('copy', 'done'), undefined, createStartupMessage('done')), MESSAGE_STATUS_DONE);
        });

        it('create-relationship-v2 done and sync-relationship with copy progress and startup with status done should return INTERRUPTED', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'done'), createActiveMessage('copy', 'progress'), undefined, createStartupMessage('done')), MESSAGE_STATUS_INTERRUPTED);
        });

        it('create-relationship-v2 done and sync-relationship with copy failed and startup with status done should return INTERRUPTED', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'done'), createActiveMessage('copy', 'failed'), undefined, createStartupMessage('done')), MESSAGE_STATUS_FAILED);
        });

        it('create-relationship-v2 done and old startup with done and sync-relationship with copy progress should return RUNNING', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'done'), createActiveMessage('copy', 'progress'), undefined, createStartupMessage('done', 1297698167812)), RUNNING_STATUS);
        });

        it('create-relationship-v2 failed and startup message done, should return  FAILED', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'failed', Date.now() - (1000 * 60 * 60)), undefined, undefined, createStartupMessage('done')), MESSAGE_STATUS_FAILED);
        });

        it('create-relationship-v2 progress and startup message failed, should return  FAILED', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'progress', Date.now() - (1000 * 60 * 60)), undefined, undefined, createStartupMessage('failed')), MESSAGE_STATUS_FAILED);
        });

        it('create-relationship-v2 done and startup message done, should return  DONE', () => {
            validateStatus(createMessagesObjcet(createActiveMessage('copy', 'done', Date.now() - (1000 * 60 * 60)), undefined, undefined, createStartupMessage('done')), MESSAGE_STATUS_DONE);
        });
    });


    describe('calculateProgressNewDataBroker', () => {
        function validateProgress(status, bytesToTransfer, bytesTransferred, dirsToTransfer, dirsTransferred, expectedProgress) {
            const progress = common.calculateProgressNewDataBroker(status, bytesToTransfer, bytesTransferred, dirsToTransfer, dirsTransferred);
            progress.should.equal(expectedProgress);
        }

        function validateUndefinedProgress(status, bytesToTransfer, bytesTransferred, dirsToTransfer, dirsTransferred) {
            const progress = common.calculateProgressNewDataBroker(status, bytesToTransfer, bytesTransferred, dirsToTransfer, dirsTransferred);
            expect(progress).to.be.undefined;
        }

        it('Calculate progress when there are no counters status is Running', () => {
            validateUndefinedProgress(RUNNING_STATUS);
        });


        it('Calculate progress when bytesToTransfer = 0 and  dirsToTransfer = 0  and the status is DONE', () => {
            validateProgress(MESSAGE_STATUS_DONE, 0, 0, 0, 0, 100);
        });

        it(`Calculate progress when bytesToTransfer = 0 and  dirsToTransfer = 0 and the status is ${MESSAGE_STATUS_FAILED}`, () => {
            validateProgress(MESSAGE_STATUS_FAILED, 0, 0, 0, 0, 0);
        });

        it(`Calculate progress when bytesToTransfer = 0 and  dirsToTransfer = 0  and the status is ${RUNNING_STATUS}`, () => {
            validateProgress(RUNNING_STATUS, 0, 0, 0, 0, 0);
        });

        it(`Calculate progress when bytesToTransfer + dirsToTransfer = 100, bytesTransferred + dirsToTransfer =  200 and the status is ${RUNNING_STATUS}`, () => {
            validateProgress(RUNNING_STATUS, 50, 50, 100, 100, 99);
        });

        it(`Calculate progress when bytesToTransfer + dirsToTransfer = 100, bytesTransferred + dirsToTransfer =  100, and the status is ${MESSAGE_STATUS_FAILED}`, () => {
            validateProgress(MESSAGE_STATUS_FAILED, 50, 50, 50, 50, 0);
        });

        it(`Calculate progress  when bytesToTransfer + dirsToTransfer = 100, bytesTransferred + dirsToTransfer =  100 and the status is ${MESSAGE_STATUS_DONE}`, () => {
            validateProgress(MESSAGE_STATUS_DONE, 100, 0, 100, 0, 100);
        });

        it(`Calculate progress when bytesToTransfer + dirsToTransfer = 100, bytesTransferred + dirsToTransfer =  200 and the status is ${MESSAGE_STATUS_DONE}`, () => {
            validateProgress(MESSAGE_STATUS_DONE, 0, 100, 0, 200, 100);
        });

        it(`Calculate progress when bytesToTransfer + dirsToTransfer = 100, bytesTransferred + dirsToTransfer =  50 and the status is ${RUNNING_STATUS}`, () => {
            validateProgress(RUNNING_STATUS, 100, 0, 50, 0, 0);
        });

        it(`Calculate progress when bytesToTransfer + dirsToTransfer = 100, bytesTransferred + dirsToTransfer =  50 and the status is ${MESSAGE_STATUS_FAILED}`, () => {
            validateProgress(MESSAGE_STATUS_FAILED, 100, 0, 100, 0, 0);
        });
    });
});
